//---------------------------------------------------------------------------
// ChessBoard Component for Delphi2 and Delphi3
// Version 1.11 - July 25th 1997
// Author: Daniel Terhell
//---------------------------------------------------------------------------

unit ChessBrd;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, Clipbrd;

const
    NoPiece:Integer =-1;
    Black  :Integer = 0;
    White  :Integer = 1;

type
  Intp= ^Integer;
  StandardSet=(Size32, Size40, Size64, Size80);
  DisplayCoords=(West, North, East, South);
  CanStillCastle=(WhiteKingSide, WhiteQueenSide, BlackKingSide, BlackQueenSide);
  CastleSet=set of CanStillCastle;
  CoordSet= set of DisplayCoords;
  Square=(None,A8,B8,C8,D8,E8,F8,G8,H8,A7,B7,C7,D7,E7,F7,G7,H7,
               A6,B6,C6,D6,E6,F6,G6,H6,A5,B5,C5,D5,E5,F5,G5,H5,
               A4,B4,C4,D4,E4,F4,G4,H4,A3,B3,C3,D3,E3,F3,G3,H3,
               A2,B2,C2,D2,E2,F2,G2,H2,A1,B1,C1,D1,E1,F1,G1,H1);

  TMoveEvent     =procedure(Sender:TObject; oldSq, newSq: Square) of object;
  TCaptureEvent  =procedure(Sender:TObject; oldSq, newSq: Square; CapturedPiece: Char) of object;
  TOneSquareEvent=procedure(Sender:TObject; square: Square) of object;
  TPromotionEvent=procedure(Sender:TObject; oldSq, newSq: Square;var NewPiece: Char) of object;

    MoveInfo = record
        position: String;
        Castling: CastleSet;
        OldSquare,NewSquare, EnPassant: Square;
        end;

  TChessBrd = class(TGraphicControl)

    private
    temp: MoveInfo;
    OldCursor: TCursor;

    MoveList: Array[0..256,0..2]of MoveInfo;

    _resizeMinSize,_resizeMaxSize: Integer;
    ResizeState, _resizable: Boolean;
    PromoteTo: Char;
    PieceIndex: Array[0..2,0..6] of Integer;
    Boardx,Boardy,BoardSize, PieceSize, _sizeofsquare, _currentMove: Integer;

    _standardSize: StandardSet;

    _legalmove,_check,_mate,_stalemate,_castle :TMoveEvent;
    _onlytwokingsleft: TNotifyEvent;
    _capture: TCaptureEvent;
    _illegalmove: TOneSquareEvent;
    _promotion: TPromotionEvent;
    _enPassant: Square;

    _position: String;

    list: TImageList;
    _squareLight, _squareDark, _borderBitmap, _custompieceset, Default: TBitmap;
    _lineStyle: TPen;
    _coordfont: TFont;

    _castlingallowed: CastleSet;
    _displayCoords: CoordSet;

    _sizeOfBorder: Integer;
    SquareClick1, SquareClick2: Square;
    _whiteOnTop,_whiteToMove,_boardlines: Boolean;
    _squareColorLight, _squareColorDark, _bordercolor: TColor;


    function CheckLegalBishopMove (oldsq, newsq: Square):Boolean;
    function CheckLegalKingMove   (oldsq, newsq: Square):Boolean;
    function CheckLegalKnightMove (oldsq, newsq: Square):Boolean;
    function CheckLegalPawnMove   (oldsq, newsq: Square):Boolean;
    function CheckLegalRookMove   (oldsq, newsq: Square):Boolean;
    function CheckLegalQueenMove  (oldsq, newsq: Square):Boolean;
    function ExistSquareDark      : Boolean;
    function ExistBorderbitmap    : Boolean;
    function ExistSquareLight     : Boolean;
    function ExistCustomPieceSet  : Boolean;
    function ExistDefaultPieceSet : Boolean;
    function MoveIsLegal (oldsq, newsq: Square):Boolean;
    function PieceToInt (piece: Char): Integer;
    procedure DoPromotion (sq: Square);
    procedure DrawBorder;
    procedure DrawBoard;
    procedure DrawPieces;
    procedure DrawPiece (sq: Square; piece: Char);
    procedure InitializeBitmaps;
    procedure OrganizeBitmaps;

    //--Boring Write Methods--------------------------------------

    procedure Set_BorderBitmap (bmp: TBitmap);
    procedure Set_BorderColor (c: TColor);
    procedure Set_SizeOfBorder (border: Integer);
    procedure Set_CoordFont (f: TFont);
    procedure Set_CurrentMove (moveno: Integer);
    procedure Set_CustomPieceSet (bmp: TBitmap);
    procedure Set_DarkSquare(bmp: TBitmap);
    procedure Set_EnPassant(sq: Square);
    procedure Set_LightSquare(bmp: TBitmap);
    procedure Set_LineStyle (pen: TPen);
    procedure Set_ResizeMaxSize (size: Integer);
    procedure Set_ResizeMinSize (size: Integer);
    procedure Set_Position (pos: String);
    procedure Set_DisplayCoords (cset: CoordSet);
    procedure Set_BoardLines (show: Boolean);
    procedure Set_SizeOfSquare (size: Integer);
    procedure Set_SquareColorDark (c: TColor);
    procedure Set_SquareColorLight (c: TColor);
    procedure Set_StandardSize (stset: StandardSet);
    procedure Set_WhiteOnTop (wabove: Boolean);
    procedure Set_WhiteToMove (wmove: Boolean);

  protected

   procedure Capture (Sender: TObject;oldSq,newSq: Square;CapturedPiece:Char);
   procedure Castle (Sender: TObject;oldSq,newSq: Square);
   procedure Check (Sender: TObject; oldSq, newSq: Square);
   procedure Click; override;
   procedure DragCanceled;override;
   procedure DragDrop(Source: TObject;X,Y: Integer);override;
   procedure DragOver(Source: TObject;X,Y: Integer; State: TDragState;var Accept: Boolean );override;
   procedure EndDrag(drop:Boolean);
   procedure IllegalMove (Sender:TObject; sq: Square);
   procedure LegalMove (Sender:TObject;oldSq,newSq: Square);
   procedure Mate (Sender:TObject;     oldSq,newSq: Square);
   procedure MouseDown(Button:TMouseButton; Shift:TShiftState;X,Y: Integer); override;
   procedure MouseMove(Shift:TShiftState; X,Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);override;
   procedure OnlyTwoKingsLeft (Sender:TObject);
   procedure Paint; override;
   procedure Promotion (Sender: TObject;oldSq,newSq: Square; var NewPiece: Char);
   procedure StaleMate (Sender: TObject;oldSq,newSq: Square);

  public

    FirstMove, LastMove: Integer;
    FirstTurn, LastTurn: Boolean;

    constructor Create(AOwner: TComponent);override;
    destructor destroy;override;

    function BlackInCheckAfter(oldsq, newsq: Square): Boolean;
    function ClearSquare(sq: Square): Boolean;
    function ColorOfPiece (piece: Char): Integer;
    function ColorOfPieceOnSquare (sq: Square): Integer;
    function ColorOfSquare (sq: Square): Integer;
    procedure DrawChessPiece (canvas: TCanvas; x,y: Integer; piece: Char);
    function GetMove (moveno: Integer; whiteMoves: Boolean): MoveInfo;
    function GetMoveList: TStringList;
    function GotoMove (moveno: Integer; whiteMoves:Boolean): Boolean;
    function LegalMoveAvailable: Boolean;
    function MouseToSquare (x, y: Integer): Square;
    function Move (oldsq, newsq: Square): Boolean;
    function MoveBackward: Boolean;
    function MoveForward: Boolean;
    procedure NewGame;
    function SetUpPosition (pos: MoveInfo; moveno: Integer; whiteMoves:Boolean): Boolean;
    procedure UpdateChessBoard (oldpos: String);
    function WhiteInCheckAfter(oldsq, newsq: Square):Boolean;
    function WindowToSquare (x, y: Integer): Square;
    function XPos (sq: Square): Integer;
    function YPos (sq: Square): Integer;

  published
    property BoardLines:Boolean read _boardLines write Set_BoardLines;
    property BorderBitmap:TBitmap read _borderBitmap write Set_BorderBitmap;
    property BorderColor:TColor read _borderColor  write Set_BorderColor;
    property CastlingAllowed:CastleSet  read _castlingAllowed write _castlingAllowed;
    property CoordFont:TFont  read _coordFont write Set_CoordFont;
    property CurrentMove:Integer  read _currentMove write Set_CurrentMove;
    property CustomPieceSet: TBitmap  read _customPieceset write Set_CustomPieceSet;
    property DisplayCoords:CoordSet  read _displayCoords write Set_DisplayCoords;
    property EnPassant:Square read _enPassant write Set_EnPassant;
    property LineStyle:TPen  read _lineStyle write Set_LineStyle;
    property Position:String read _position write Set_Position;
    property Resizable:Boolean  read _resizable write _resizable default FALSE;
    property ResizeMinSize:Integer  read _resizeMinSize write Set_ResizeMinSize;
    property ResizeMaxSize:Integer  read _resizeMaxSize write Set_ResizeMaxSize;
    property SizeOfBorder:Integer  read _sizeofBorder write Set_SizeOfBorder;
    property SizeOfSquare:Integer  read _sizeofsquare write Set_SizeOfSquare;
    property SquareColorDark:TColor  read _squareColorDark write Set_SquareColorDark;
    property SquareColorLight:TColor  read _squareColorLight write Set_SquareColorLight;
    property SquareDark:TBitmap  read _squareDark write Set_DarkSquare;
    property SquareLight:TBitmap  read _squareLight write Set_LightSquare;
    property StandardSize:StandardSet read _standardSize write Set_StandardSize default Size40;
    property WhiteOnTop:Boolean  read _whiteOnTop write Set_WhiteOnTop;
    property WhiteToMove:Boolean  read _whiteToMove write Set_WhiteToMove default TRUE;

    property OnCapture: TCaptureEvent  read _capture write _capture;
    property OnCheck: TMoveEvent read _check write _check;
    property OnCastle: TMoveEvent  read _castle write _castle;
    property OnLegalMove: TMoveEvent  read _legalmove write _legalmove;
    property OnMate: TMoveEvent  read _mate write _mate;
    property OnStaleMate: TMoveEvent  read _stalemate write _stalemate;
    property OnOnlyTwoKingsLeft: TNotifyEvent  read _onlytwokingsleft write _onlytwokingsleft;
    property OnIllegalMove: TOneSquareEvent  read _illegalmove write _illegalmove;
    property OnPromotion: TPromotionEvent read _promotion write _promotion;

    property Visible;
    property Enabled;

    property DragCursor;
    property DragMode;
    property OnClick;
    property OnDblClick ;
    property OnDragDrop ;
    property OnDragOver ;
    property OnEndDrag ;
    property OnMouseDown ;
    property OnMouseMove ;
    property OnMouseUp;
    property OnStartDrag;

  end;

  ChessBrdError = class(Exception);

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Chess', [TChessBrd]);
end;

constructor TChessBrd.Create(AOwner: TComponent);
var
    p,q: Intp;
begin
    inherited Create(AOwner);

    list:=TImageList.CreateSize(40,40);
    Default:=TBitmap.Create;

    _lineStyle:=TPen.Create;
    _coordfont:=TFont.Create;
    _squareLight:=TBitmap.Create;
    _squareDark:=TBitmap.Create;
    _borderBitmap:=TBitmap.Create;
    _customPieceSet:=TBitmap.Create;

    _standardSize:=Size40;
    _resizeMinSize:=100;
    _resizeMaxSize:=1000;

    FirstMove:=1;
    LastMove:=1;
    LastTurn:=TRUE;

    PromoteTo:='q';

    Include (_displayCoords,West);
    Include (_displayCoords,East);
    Include (_displayCoords,North);
    Include (_displayCoords,South);

    InitializeBitmaps;
    OrganizeBitmaps;

    _coordfont.Color:=clWhite;
    _coordfont.Name:='Arial';
    _coordfont.Size:=7;
    _coordfont.Pitch:=fpDefault;

    _sizeofsquare:=40;
    _sizeOfBorder:=24;
    _bordercolor:=clOlive;
    _squareColorDark:=clGray;
    _squareColorLight:=clSilver;

    Boardx:=_sizeOfBorder;
    Boardy:=_sizeOfBorder;

    p:=@Width;
    p^:=8*_sizeofsquare+2*_sizeOfBorder;
    q:=@Height;
    q^:=Width;

    NewGame;
end;



destructor TChessBrd.destroy;
begin
    inherited destroy;

    list.Destroy;
    _lineStyle.Destroy;
    _coordfont.Destroy;
    _squareLight.Destroy;
    _squareDark.Destroy;
    _borderBitmap.Destroy;
    _customPieceSet.Destroy;
    Default.Destroy;

end;


procedure TChessBrd.MouseDown(Button: TMouseButton;
          Shift: TShiftState;X,Y: Integer);
var
    sq: Square;
const
     space: Char=' ';
begin

    inherited MouseDown(Button,Shift,X,Y);

    sq:=WindowToSquare(X,Y);
    if (sq>=A8)AND(sq<=H1)AND
        (_position[Integer(sq)]<>' ') then
         begin
         SquareClick1:=sq;
         BeginDrag(FALSE);
         end;
end;


procedure TChessBrd.Paint;
var
    size: Integer;
begin
    size:=(Width+Height) shr 1;
    _sizeofsquare:=(size-2*_sizeOfBorder) shr 3;
    Width:=8*_sizeofsquare+2*_sizeOfBorder;
    Height:=Width;
    DrawBorder;
    DrawBoard;
    DrawPieces;
end;



procedure TChessBrd.DragOver(Source:TObject;
     X,Y: Integer; State: TDragState; var Accept: Boolean );
var
    sq: Square;
    mid: Integer;
begin
    inherited DragOver(Source, X, Y, State, Accept);

    sq:=WindowToSquare(X,Y);
    mid:=_sizeofsquare shr 1;

    if (list.Dragging=FALSE)then
    begin
        ClearSquare(SquareClick1);
        list.SetDragImage(PieceToInt(_position[Integer(SquareClick1)]), 0,0);
        list.DragLock  (Parent.Handle,X+Self.Left-mid,Y+Self.Top-mid);
        list.BeginDrag (Parent.Handle,0,0);
    end;

    list.DragMove(X+Self.Left-mid,Y+Self.Top-mid);

    list.ShowDragImage();

    if (Source=Self)then
    begin
        if MoveIsLegal(SquareClick1,sq)or (SquareClick1=sq) then
        begin
            Accept:=TRUE;
        end;
    end;
end;

procedure TChessBrd.DragDrop(Source:TObject ;X,Y: Integer);
begin
    inherited DragDrop(Source,X,Y);
    if (list.Dragging)then
    begin
        list.HideDragImage;
        list.EndDrag;
        list.DragUnlock;
    end;

    SquareClick2:=WindowToSquare(X,Y);
    if (SquareClick1<>SquareClick2)then
        Move (SquareClick1,SquareClick2)
    else DrawPiece (SquareClick1,_position[Integer(SquareClick1)]);
end;

procedure TChessBrd.DragCanceled;
begin
    list.HideDragImage;
    list.EndDrag;
    list.DragUnlock;
    DrawPiece (SquareClick1,_position[Integer(SquareClick1)]);
    if @_illegalmove<>nil then _illegalmove(Self, SquareClick1);
end;

procedure TChessBrd.MouseMove(Shift: TShiftState;X,Y: Integer);
var
    w: Integer;
begin
    inherited MouseMove(Shift,X,Y);

    if (Cursor<>crSizeNWSE)and(Cursor<>crSizeWE) and (Cursor<>crSizeNS)then
        OldCursor:=Cursor;

    if (_resizable)and(X>=(Width-10))and(X<=Width)and(Y>=(Height-10))
         and (Y<=Height) then
        begin
        Cursor:=crSizeNWSE;
        if (ssLeft in Shift)then
            ResizeState:=TRUE;
        end
    else if (_resizable) and(ResizeState=FALSE)then
        Cursor:=OldCursor;

    if (_resizable)and(ResizeState) then
    begin
        if (X>Y) then w:=X
          else w:=Y;
        if (w<_resizeMinSize) then w:=_resizeMinSize;
        if (w>_resizeMaxSize) then w:=_resizeMaxSize;
        if (w<>Width) then
            begin
            Width:=w;
            end;
    end;
end;

procedure TChessBrd.MouseUp(Button:TMouseButton;Shift:TShiftState;
          X,Y: Integer);
begin
    inherited MouseUp(Button,Shift,X,Y);
    if (ResizeState) then ResizeState:=FALSE;
end;

procedure TChessBrd.Click;
begin
    inherited Click();
end;

procedure TChessBrd.EndDrag(drop: Boolean);
begin
    inherited EndDrag(drop);
end;

procedure TChessBrd.LegalMove (Sender:TObject; oldSq,newSq:Square);
begin
end;

procedure TChessBrd.IlLegalMove (Sender:TObject; sq:Square);
begin
end;

procedure TChessBrd.Check (Sender:TObject; oldSq, newSq:Square);
begin
end;

procedure TChessBrd.Mate (Sender:TObject; oldSq, newSq:Square);
begin
end;

procedure TChessBrd.Castle (Sender:TObject; oldSq, newSq:Square);
begin
end;

procedure TChessBrd.StaleMate (Sender:TObject; oldSq, newSq:Square);
begin
end;

procedure TChessBrd.Capture (Sender:TObject; oldSq, newSq:Square; CapturedPiece: Char);
begin
end;

procedure TChessBrd.OnlyTwoKingsLeft (Sender:TObject);
begin
end;

procedure TChessBrd.Promotion (Sender:TObject; oldSq,newSq: Square;
     var NewPiece: Char);
var
    i: String;
    r: Integer;
begin
      i:='NBRQnbrq';

      for r:=1 to 8 do
      begin
          if (NewPiece=i[r])then
              Break;
          if (r=8) then NewPiece:='q';
      end;

    PromoteTo:=NewPiece;
end;

procedure TChessBrd.DoPromotion (sq: Square);
begin
    if YPos(sq)=8 then
        begin
        _position[Integer(sq)]:=UpCase(PromoteTo);
        end
    else if (YPos(sq)=1) then
        begin
        _position[Integer(sq)]:=Char(Integer(UpCase(PromoteTo))+32);
        end;

    ClearSquare(sq);
    DrawPiece(sq,_position[Integer(sq)]);
end;

//-----------------------------------------------------------------------
// Boring Write Methods
//-----------------------------------------------------------------------

procedure TChessBrd.Set_Position(pos:String);
var
    PieceStr: String;
    q,n,len: Integer;
begin
    PieceStr:=' PNBRQKpnbrqk';
    len:=Length(pos);
    if (len>64) then len:=64;

    _position:='                                                                ';

    for q:=1 to len do
        begin
        if Integer(pos[q])>0 then
            begin
            for n:=1 to 13 do
                begin
                if PieceStr[n]=pos[q] then
                    begin
                    _position[q]:=pos[q];
                    Break;
                    end;
                end;
            end
        else Break;
        end;

    if (LowerCase(pos)='init')or(LowerCase(pos)='newgame') then
    begin
        _position:='rnbqkbnrpppppppp                                PPPPPPPPRNBQKBNR';
    end;

    Paint;
end;

procedure TChessBrd.Set_ResizeMinSize (size: Integer);
begin
    if (size<_resizeMaxSize) then _resizeMinsize:=size
        else _resizeMinSize:=_resizeMaxSize;
end;

procedure TChessBrd.Set_ResizeMaxSize (size: Integer);
begin
    if (size > _resizeMinsize) then
        _resizeMaxSize:=size
    else _resizeMaxSize:=_resizeMinSize;
end;

procedure  TChessBrd.Set_SizeOfSquare (size: Integer);
begin
    _sizeofsquare:=size;
    Width:=8*_sizeofsquare+2*_sizeOfBorder;
end;

procedure  TChessBrd.Set_CurrentMove (moveno: Integer);
begin
    // Read-Only
end;

procedure  TChessBrd.Set_EnPassant (sq: Square);
begin
    if ((sq>=A6)and(sq<=H6))or((sq>=A3)and(sq<=H3)) then
        _enPassant:=sq;
end;

procedure  TChessBrd.Set_BoardLines (show: Boolean);
begin
    _boardLines:=show;
    Paint;
end;

procedure  TChessBrd.Set_DisplayCoords (cset: CoordSet);
begin
    _displayCoords:=cset;
    Paint;
end;

procedure  TChessBrd.Set_CoordFont(f: TFont);
begin
    _coordfont.Assign(f);
    Paint;
end;

procedure  TChessBrd.Set_StandardSize(stset: StandardSet);
begin
    _standardSize:=stset;

    InitializeBitmaps;
    OrganizeBitmaps;
    Paint;
end;

procedure  TChessBrd.Set_CustomPieceSet(bmp: TBitmap);
begin
    if  (bmp=nil)or((bmp<>nil)and (bmp.Width*2=bmp.Height*3)) then
    begin
       _customPieceSet.Assign(bmp);
       OrganizeBitmaps;
       Paint;
    end;
end;

procedure  TChessBrd.Set_LineStyle (pen: TPen);
begin
    _lineStyle.Assign(pen);
    Paint;
end;

procedure  TChessBrd.Set_SizeOfBorder (border: Integer);
begin
    _sizeOfBorder:=border;
    Boardx:=_sizeOfBorder;
    Boardy:=_sizeOfBorder;

    Width:=8*_sizeofsquare+2*_sizeOfBorder;
end;

procedure  TChessBrd.Set_WhiteOnTop (wabove: Boolean);
begin
    _whiteOnTop:=wabove;
    Paint;
end;

procedure  TChessBrd.Set_WhiteToMove (wmove: Boolean);
begin
    _whiteToMove:=wmove;
end;

procedure  TChessBrd.Set_DarkSquare(bmp: TBitmap);
begin
    _squareDark.Assign(bmp);
    Paint;
end;

procedure  TChessBrd.Set_LightSquare(bmp: TBitmap);
begin
    _squareLight.Assign(bmp);
    Paint;
end;

procedure  TChessBrd.Set_BorderBitmap(bmp: TBitmap);
begin
    _borderBitmap.Assign(bmp);
    Paint;
end;

procedure  TChessBrd.Set_SquareColorDark(c: TColor);
begin
    _squareColorDark:=c;
    Paint;
end;

procedure  TChessBrd.Set_SquareColorLight(c: TColor);
begin
    _squareColorLight:=c;
    Paint;
end;

procedure  TChessBrd.Set_BorderColor(c: TColor);
begin
    _borderColor:=c;
    Paint;
end;

//---------------------------------------------------------------------
//-------PUBLIC Graphic Routines-----------------------------------------------
//---------------------------------------------------------------------

procedure TChessBrd.DrawChessPiece (canvas: TCanvas; x,y: Integer; piece: Char);
var
    v,i,j: Integer;
    c: Char;
begin
    i:=-1;
    c:=piece;

    if (c>='A')and(c<='Z')then
        c:=Char(Integer(c)+32);
    case c of
         'p': i:=0;
         'n': i:=1;
         'b': i:=2;
         'r': i:=3;
         'q': i:=4;
         'k': i:=5;
         end;
    if (piece>='A')and(piece<='Z')
        then j:=0 else j:=1;

    if (i>=0) then v:=PieceIndex[j,i]
        else v:=-1;

    list.Draw (canvas,x,y,v);
end;

procedure  TChessBrd.UpdateChessBoard (oldpos: String);
var
    sq: Square;
begin
    for sq:=A8 to H1 do
        begin
            if (oldpos[Integer(sq)]<>_position[Integer(sq)])then
            begin
                ClearSquare(sq);
                if (_position[Integer(sq)]<>' ')then
                    DrawPiece(sq,_position[Integer(sq)]);
            end;
        end;
end;

procedure  TChessBrd.DrawPiece (sq: Square; piece: Char);
var
    x,y,adj: Integer;
begin
    x:=(Integer(sq)-1) mod 8;
    y:=(Integer(sq)-1) shr 3;

    if (_whiteOnTop)then
        begin
        x:=7-x;
        y:=7-y;
        end;

    adj:=Trunc ((_sizeofsquare-PieceSize)/2);
    list.Draw(Canvas, Boardx+_sizeofsquare*x+adj, Boardy+_sizeofsquare*y+adj, PieceToInt(piece));
end;

//Clears the int according to the Screen coords
//Returns TRUE if a int could be found at (x,y)
function  TChessBrd.ClearSquare(sq: Square): Boolean;
var
    r: TRect;
begin
    if (sq<A8)or(sq>H1)then
        begin
        Result:=FALSE;
        Exit;
        end;

    if (_whiteOnTop)then
        sq:=Square(65-Integer(sq));

    r.Left:=Boardx+((Integer(sq)-1) and 7)*_sizeofsquare;
    r.Right:=r.Left+_sizeofsquare;
    r.Top:=Boardy+((Integer(sq)-1) shr 3)*_sizeofsquare;
    r.Bottom:=r.Top+_sizeofsquare;

    if (ColorOfSquare(sq)=White)then
        begin
        if (ExistSquareLight)then
            begin
            Canvas.Brush.Bitmap:=_squareLight;
            if (_boardLines) then
                Canvas.Rectangle(r.Left,r.Top,r.Right,r.Bottom)
            else Canvas.FillRect(r);
            end
        else
            begin
            Canvas.Brush.Style:=bsSolid;
            Canvas.Brush.Color:=SquareColorLight;
            if (_boardLines)then
                Canvas.Rectangle(r.Left,r.Top,r.Right,r.Bottom)
            else Canvas.FillRect(r);
            end;
        end
    else
        begin
        if (ExistSquareDark)then
            begin
            Canvas.Brush.Bitmap:=_squareDark;
            if (_boardLines)then
                Canvas.Rectangle(r.Left,r.Top,r.Right,r.Bottom)            else Canvas.FillRect(r);
            end
        else
            begin
            Canvas.Brush.Style:=bsSolid;
            Canvas.Brush.Color:=SquareColorDark;
            if (_boardLines)then
                Canvas.Rectangle(r.Left,r.Top,r.Right,r.Bottom)
            else Canvas.FillRect(r);
            end;
        end;
    Result:=TRUE;
end;

//Converts from Window Coordinates to square
//returns 0 if invalid else
// 1-64 as a int number
function TChessBrd.WindowToSquare (x,y: Integer): Square;
var
    xn,yn,xv,yv: Integer;
begin

    xv:=x-(Boardx);
    yv:=y-(Boardy);

    if ((xv<0)or(xv>=8*_sizeofsquare)or(yv<0)or(yv>=8*_sizeofsquare))then
       begin
       Result:=None;
       Exit;
       end;

    xn:=Trunc(xv/_sizeofsquare);
    yn:=Trunc(yv/_sizeofsquare);

    if (xn>7) then xn:=7;
    if (yn>7) then yn:=7;

    if (_whiteOnTop)then
    begin
        xn:=7-xn;
        yn:=7-yn;
    end;

    Result:=Square(8*yn+xn+1);
end;

//----------------------------PRIVATE Graphic Routines

function TChessBrd.ExistCustomPieceSet: Boolean;
begin
    Result:=(_customPieceSet<>nil)and(_customPieceSet.Width>0)and(_customPieceSet.Height>0);
end;

function TChessBrd.ExistDefaultPieceSet: Boolean;
begin
    Result:=(Default<>nil)and(Default.Width>0) and  (Default.Height>0);
end;

function TChessBrd.ExistBorderBitmap: Boolean;
begin
    Result:=(_borderBitmap<>nil)and(_borderBitmap.Width>0)and(_borderBitmap.Height>0);
end;

function TChessBrd.ExistSquareLight: Boolean;
begin
    Result:=(_squareLight<>nil)and(_squareLight.Width>0)and(_squareLight.Height>0);
end;

function TChessBrd.ExistSquareDark: Boolean;
begin
    Result:=(_squareDark<>nil)and(_squareDark.Width>0)and(_squareDark.Height>0);
end;

procedure TChessBrd.OrganizeBitmaps;
var
    Bmp: TBitmap;
    tmp, tmpmask: TBitmap;
    r,n: Integer;
    src,dest: TRect;
begin
    if (ExistCustomPieceSet) then
        Bmp:=_customPieceSet
    else Bmp:=Default;

    if (Bmp.Width*2 = Bmp.Height*3) then PieceSize:=Bmp.Height shr 2
        else PieceSize:=0;

    list.Clear;
    list.Masked:=TRUE;
    list.DrawingStyle:=dsTransparent;
    list.Width:=PieceSize;
    list.Height:=PieceSize;

    tmp:=TBitmap.Create;
    tmpmask:=TBitmap.Create;

    tmp.Width:=PieceSize;
    tmp.Height:=PieceSize;
    tmpmask.Width:=PieceSize;
    tmpmask.Height:=PieceSize;

    for n:=0 to 1 do
    for r:=0 to 5 do
    begin
        src.Left:=r*PieceSize;
        src.Top:=n*(PieceSize*2);
        src.Right:=src.Left+PieceSize;
        src.Bottom:=n*(PieceSize*2)+PieceSize;
        dest.Left:=0;
        dest.Top:=0;
        dest.Right:=PieceSize;
        dest.Bottom:=PieceSize;
        tmp.Canvas.CopyRect(dest,Bmp.Canvas,src);

        src.Left:=r*PieceSize;
        src.Top:=n*(PieceSize*2)+PieceSize;
        src.Right:=src.Left+PieceSize;
        src.Bottom:=n*(PieceSize*2)+(PieceSize*2);
        tmpmask.Canvas.CopyRect(dest,Bmp.Canvas,src);

        PieceIndex[n,r]:=list.Add (tmp,tmpmask);
    end;

    tmpmask.Free;
    tmp.Free;
end;

procedure  TChessBrd.DrawBorder;
var
    dest: TRect;
    abc: String;
    r, v, boffset, soffset: Integer;
begin
    Canvas.CopyMode:=cmSrcCopy;
    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=_bordercolor;

    if (ExistBorderBitmap)then
       Canvas.Brush.Bitmap:=_borderBitmap;

    dest.Left:=0;
    dest.Top:=0;
    dest.Right:=_sizeOfBorder;
    dest.Bottom:=Height;
    Canvas.FillRect (dest);

    dest.Left:=0;
    dest.Top:=0;
    dest.Right:=Width;
    dest.Bottom:=_sizeOfBorder;
    Canvas.FillRect (dest);

    dest.Left:=Width;
    dest.Top:=Height;
    dest.Right:=0;
    dest.Bottom:=Height-_sizeOfBorder;
    Canvas.FillRect (dest);
    dest.Left:=Width;
    dest.Top:=Height;
    dest.Right:=Width-_sizeOfBorder;
    dest.Bottom:=0;
    Canvas.FillRect (dest);

    Canvas.Font:=_coordfont;

    if (WhiteOnTop=FALSE) then abc:='ABCDEFGH'
        else abc:='HGFEDCBA';

    boffset:=((_sizeOfBorder-_coordfont.Size)shr 1)+1;
    soffset:=(Integer(_sizeofsquare-_coordfont.Size) shr 1);

    for r:=0 to 7 do
    begin
        if (WhiteOnTop) then
            v:=r+1 else
        v:=8-r;

        if (West in _displayCoords)then
             Canvas.TextOut(boffset,_sizeOfBorder+r*_sizeofsquare+soffset,IntToStr(v));
        if (East in _displayCoords)then
             Canvas.TextOut(Width-_sizeOfBorder+boffset,_sizeOfBorder+r*_sizeofsquare+soffset,IntToStr(v));
        if (North in _displayCoords)then
             Canvas.TextOut(_sizeOfBorder+r*_sizeofsquare+soffset,boffset,abc[r+1]);
        if (South in _displayCoords)then
    Canvas.TextOut(_sizeOfBorder+r*_sizeofsquare+soffset,Height-_sizeOfBorder+boffset,abc[r+1]);
    end;

    Canvas.Pen:=_lineStyle;

    Canvas.MoveTo(0,0);
    Canvas.LineTo(Width-1,0);
    Canvas.LineTo(Width-1,Width-0);
    Canvas.LineTo(0,Width-1);
    Canvas.LineTo(0,0);

    Canvas.MoveTo(_sizeOfBorder-1,_sizeOfBorder-1);
    Canvas.LineTo(Width-_sizeOfBorder,_sizeOfBorder-1);
    Canvas.LineTo(Width-_sizeOfBorder,Width-_sizeOfBorder);
    Canvas.LineTo(_sizeOfBorder-1,Width-_sizeOfBorder);
    Canvas.LineTo(_sizeOfBorder-1,_sizeOfBorder-1);
end;

procedure  TChessBrd.DrawPieces;
var
    sq: Square;
    len: Integer;
begin
    if (Length(_position)<64) then len:=Length(_position)
        else len:=64;
    for sq:=A8 to Square(len) do
    begin
        if (_position[Integer(sq)]<>' ') then
            DrawPiece(sq,_position[Integer(sq)]);
    end;
end;

procedure  TChessBrd.DrawBoard;
var
    dest: TRect;
    n,x,y: Integer;
begin
    Canvas.CopyMode:=cmSrcCopy;
    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=SquareColorLight;

    if (ExistSquareLight)then
       Canvas.Brush.Bitmap:=_squareLight;

    for n:=0 to 1 do
    for x:=n to 7 do
    for y:=n to 7 do
    begin
        if (x mod 2 =n)and(y mod 2=n)then
            begin
            dest.Left:=Boardx+x*_sizeofsquare;
            dest.Top:=Boardy+y*_sizeofsquare;
            dest.Right:=Boardx+(x+1)*_sizeofsquare;
            dest.Bottom:=Boardy+(y+1)*_sizeofsquare;
            end;
        if (_boardLines)then
            Canvas.Rectangle(dest.Left,dest.Top,dest.Right,dest.Bottom)
        else Canvas.FillRect(dest);
    end;

    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=SquareColorDark;

    if (ExistSquareDark)then
    begin
        Canvas.Brush.Bitmap:=_squareDark;
    end;

    for n:=0 to 1 do
    for x:=(1-n) to 7  do
    for y:=n to 7  do
    begin
         if (x mod 2 <>n) and (y mod 2 =n) then
            begin
            dest.Left:=Boardx+x*_sizeofsquare;
            dest.Top:=Boardy+y*_sizeofsquare;
            dest.Right:=Boardx+(x+1)*_sizeofsquare;
            dest.Bottom:=Boardy+(y+1)*_sizeofsquare;
            if (_boardLines)then
              Canvas.Rectangle(dest.Left,dest.Top,dest.Right,dest.Bottom)            else Canvas.FillRect(dest);
            end;
     end;
end;

//-----------------------------------------------------------------------
//------PUBLIC Real Chess here-------------------------------------------
//-----------------------------------------------------------------------

function TChessBrd.GetMoveList:TStringList;
var
    c: Char;
    m: TStringList;
    s: String;
    r,side: Integer;
    Label BreakOut;
begin
    m:=TStringList.Create;

    for r:=FirstMove to LastMove do
    for side:=0 to 1 do
    begin
        if (r=LastMove)and(LastTurn=Boolean(1-side)) then goto BreakOut;

        if (side=0) then  s:=IntToStr(r)+'. '
        else s:='              ';

        if (r=FirstMove)and(side=0)and(FirstTurn=FALSE)then
            s:=s+' ...   '
        else
        begin
            c:=Char(Integer('A')+XPos(MoveList[r][side].OldSquare)-1);
            s:=s+c;
            c:=char(Integer('0')+YPos(MoveList[r][side].OldSquare));
            s:=s+c;
        end;

        s:=s+'-';

        c:=char(Integer('A')+XPos(MoveList[r][side].NewSquare)-1);
        s:=s+c;
        c:=char(Integer('0')+YPos(MoveList[r][side].NewSquare));
        s:=s+c;
        m.Add(s);

    end;

    BreakOut:
    Result:=m;
end;

function TChessBrd.GetMove (moveno:Integer;whiteMoves: Boolean): MoveInfo;
begin
    temp.OldSquare:=None;
    temp.NewSquare:=None;
    temp.position:='init';
    temp.EnPassant:=None;

    if (moveno>LastMove)or((moveno=LastMove)and(LastTurn)and(_whiteToMove=FALSE)) then
        begin
        Result:=temp;
        Exit;
        end;

    if (moveno<FirstMove)or((moveno=FirstMove)and(FirstTurn=FALSE)and(_whiteToMove))then
        begin
        Result:=temp;
        Exit;
        end;

    Result:=MoveList[moveno][1-Integer(_whiteToMove)];
end;

//Squares are numbered from 1 - 64 (a8,b8...h1)
function TChessBrd.ColorOfSquare (sq:Square): Integer;
begin
    Result:=Integer (((Integer(sq)-1) and 1)=(((Integer(sq)-1) shr 3) and 1));
end;

function  TChessBrd.PieceToInt (piece: Char): Integer;
var
    i: String;
    r: Integer;
begin
    i:='PNBRQKpnbrqk';

    for r:=1 to 12 do
        if (piece=i[r]) then
            begin
            Result:=(PieceIndex[Trunc((r-1)/6)][(r-1) mod 6]);
            Exit;
            end;
    Result:=-1;
end;


//returns color of the PIECE on a square
function TChessBrd.ColorOfPieceOnSquare (sq: Square): Integer;
begin
    Result:=NoPiece;
      if ((_position[Integer(sq)]>='b')and(_position[Integer(sq)]<='r'))then Result:=Black
 else if ((_position[Integer(sq)]>='B')and(_position[Integer(sq)]<='R'))then Result:=White;
end;

function TChessBrd.ColorOfPiece (piece:Char): Integer;
var
    i: String;
    r: Integer;
begin
    i:='NBRQnbrq';

    Result:=NoPiece;

    for r:=1 to 8 do
    begin
         if (piece=i[r]) then Break;
         if (r=8) then Exit;
    end;

    Result:=Integer ((piece>='B')and(piece<='R'));
end;


// Returns XPos of a int according to ChessBoard coordinates (A-H)
function TChessBrd.XPos (sq: Square): Integer;
begin
    Result:=(1+(Integer(sq)-1) mod 8);
end;

// Returns YPos of a int according to ChessBoard coordinates (1-8)
function TChessBrd.YPos (sq: Square): Integer;
begin
    Result:=8-((Integer(sq)-1) shr 3);
end;

function TChessBrd.MouseToSquare (x,y: Integer): Square;
begin
    if (x>=1)and(x<=8)and(y>=1)and (y<=8) then
        Result:=Square(8*(8-y)+x)
    else Result:=None;
end;

function TChessBrd.SetUpposition (pos: MoveInfo; moveno: Integer; whiteMoves:Boolean): Boolean;
var
    i: String;
    r,n: Integer;
    Label BreakOut;
begin
    i:=' pnbrqkPNBRQK';

    if  (moveno>200)or
        ((pos.EnPassant<>None)and(YPos(pos.EnPassant)<>3)and(YPos(pos.EnPassant)<>6))then
           begin
           Result:=FALSE;
           Exit;
           end;

    for r:=1 to 64 do
    begin
        for n:=1 to 13 do
        begin
            if (pos.position[r]=i[n]) then goto BreakOut;
        end;
        Result:=FALSE;
        Exit;

    BreakOut:
    end;

    _position:=pos.position;
    _whiteToMove:=_whiteToMove;
    FirstTurn:=_whiteToMove;
    _currentMove:=moveno;
    FirstMove:=moveno;

    _castlingallowed:=pos.Castling;
    _enPassant:=pos.EnPassant;

    Paint;
    Result:=TRUE;
end;

procedure  TChessBrd.NewGame;
begin
    _position:='rnbqkbnrpppppppp                                PPPPPPPPRNBQKBNR';

    _whiteToMove:=TRUE;
    _currentMove:=1;
    FirstMove:=1;
    FirstTurn:=TRUE;
    LastMove:=1;
    LastTurn:=TRUE;

    _enPassant:=None;

    Include (_castlingallowed,WhiteKingSide);
    Include (_castlingallowed,WhiteQueenSide);
    Include (_castlingallowed,BlackKingSide);
    Include (_castlingallowed,BlackQueenSide);
end;

function TChessBrd.Move (oldsq, newsq: Square): Boolean;
var
    ch,stillmove: Boolean;
    oldpiece: Char;
    sq: Square;
    r: Integer;
    count: Integer;
    Label Out;
begin
    if (MoveIsLegal(oldsq,newsq)=FALSE)then
        begin
        Result:=FALSE;
        Exit;
        end;

    ch:=(WhiteInCheckAfter(oldsq,newsq)or BlackInCheckAfter(oldsq,newsq));

    MoveList[_currentMove][1-Integer(_whiteToMove)].position:=_position;
    MoveList[_currentMove][1-Integer(_whiteToMove)].Castling:=_castlingallowed;
    MoveList[_currentMove][1-Integer(_whiteToMove)].EnPassant:=_enPassant;
    MoveList[_currentMove][1-Integer(_whiteToMove)].OldSquare:=Square(oldsq);
    MoveList[_currentMove][1-Integer(_whiteToMove)].NewSquare:=Square(newsq);
    LastMove:=_currentMove;
    LastTurn:=_whiteToMove;

    if (_whiteToMove=FALSE) then _currentMove:=_currentMove+1;
    _whiteToMove:=not _whiteToMove;

    oldpiece:=_position[Integer(newsq)];
    _position[Integer(newsq)]:=_position[Integer(oldsq)];
    _position[Integer(oldsq)]:=' ';
    ClearSquare(oldsq);
    ClearSquare(newsq);
    DrawPiece(newsq,_position[Integer(newsq)]);

    //Call OnLegalMove event handler
    if (@_legalmove<>nil) then
        _legalmove(Self,oldsq,newsq);

    //Eventually call OnCapture event handler (also in En Passant)
    if (oldpiece<>' ')and(@_capture<>nil)then
        _capture(Self,oldsq,newsq,oldpiece);

    //More to do if last move was a promotion
    if ((((_position[Integer(newsq)])='p')or(_position[Integer(newsq)]='P'))and
        ((YPos(newsq)=8) or (YPos(newsq)=1)))then
    begin
        if (@_promotion<>nil)then
             _promotion(Self,oldsq,newsq,PromoteTo);

        DoPromotion(newsq);
    end;

    //More to do if last move was En Passant capture
    if (_enPassant=newsq)then
    begin
        if (YPos(newsq)=6)and (_position[Integer(newsq)]='P') then
        begin
            sq:=MouseToSquare(XPos(newsq),5);
            if (@_capture<>nil)then
                _capture(Self,oldsq,newsq,_position[Integer(sq)]);
            _position[Integer(sq)]:=' ';
            ClearSquare(sq);
        end
        else if (YPos(newsq)=3)and (_position[Integer(newsq)]='p')then
        begin
            sq:=MouseToSquare(XPos(newsq),4);
            if (@_capture<>nil)then
                _capture(Self,oldsq,newsq,_position[Integer(sq)]);
            _position[Integer(sq)]:=' ';
            ClearSquare(sq);
        end;
    end;

    //More to do if last move allows En Passant continuation
    if (((_position[Integer(newsq)]='P')and(YPos(oldsq)=2)and(YPos(newsq)=4))or
        ((_position[Integer(newsq)]='p')and(YPos(oldsq)=7)and(YPos(newsq)=5)))then
    begin
        _enPassant:=Square ((Integer(oldsq)+Integer(newsq)) shr 1);
    end

    else _enPassant:=None;

    //More to do if last move was castling
    if (_position[Integer(newsq)]='K')and(oldsq=E1) then
    begin
        if (Square(newsq)=G1)then
        begin
            _position[Integer(F1)]:='R';
            _position[Integer(H1)]:=' ';
            ClearSquare(H1);
            DrawPiece(F1,'R');
            if (@_castle<>nil)then
                _castle(Self,oldsq,newsq);
        end
        else if (Square(newsq)=C1)then
        begin
            _position[Integer(D1)]:='R';
            _position[Integer(A1)]:=' ';
            ClearSquare(A1);
            DrawPiece(D1,'R');
            if (@_castle<>nil)then
                _castle(Self,oldsq,newsq);
        end;
    end
    else if (_position[Integer(newsq)]='k')and(Square(oldsq)=E8)then
    begin
        if (Square(newsq)=G8)then
        begin
            _position[Integer(F8)]:='r';
            _position[Integer(H8)]:=' ';
            ClearSquare(H8);
            DrawPiece(F8,'r');
            if (@_castle<>nil)then
                _castle(Self,oldsq,newsq);
        end
        else if (Square(newsq)=C8)then
        begin
            _position[Integer(D8)]:='r';
            _position[Integer(A8)]:=' ';
            ClearSquare(A8);
            DrawPiece(D8,'r');
            if (@_castle<>nil)then
                _castle(Self,oldsq,newsq);
        end;
    end;

    //Eventually Remove Castling Allowance

    case Square(oldsq) of
        //if a rook moved
        A1: begin
            Exclude(_castlingallowed,WhiteQueenSide);
            goto Out;
            end;
        H1: begin
            Exclude (_castlingallowed,WhiteKingSide);
            goto Out;
            end;
        A8: begin
            Exclude (_castlingallowed,BlackQueenSide);
            goto Out;
            end;
        H8: begin
            Exclude (_castlingallowed,BlackKingSide);
            goto Out;
            end;

        //or if a king moved
        E1: begin
            Exclude (_castlingallowed,WhiteQueenSide);
            Exclude (_castlingallowed,WhiteKingSide);
            goto Out;
            end;
        E8: begin
            Exclude (_castlingallowed,BlackQueenSide);
            Exclude (_castlingallowed,BlackKingSide);
            goto Out;
            end;
    end;

    Out:

    //More to do if there's mate or stalemate
        stillmove:=LegalMoveAvailable();
    if (ch)then
    begin
        if (@_check<>nil)then
            _check(Self,oldsq,newsq);
        if (stillmove=FALSE)and(@_mate<>nil)then
                _mate(Self,oldsq,newsq);
    end
    else
    begin
        if (stillmove=FALSE)and(@_stalemate<>nil)then
            _stalemate(Self,oldsq,newsq);
    end;

    //More to do if there's only two kings left
    count:=0;
    for r:=1 to 64 do
         if(_position[r]<>' ')and(_position[r]<>'k')and(_position[r]<>'K')then
            count:=count+1;
    if (count=0)and(@_onlytwokingsleft<>nil)then
        _onlytwokingsleft(Self);

    MoveList[_currentMove][1-Integer(_whiteToMove)].position:=_position;
    MoveList[_currentMove][1-Integer(_whiteToMove)].Castling:=_castlingallowed;
    MoveList[_currentMove][1-Integer(_whiteToMove)].enPassant:=_enPassant;
    LastMove:=_currentMove;
    LastTurn:=_whiteToMove;

    Result:=TRUE;
end;

function TChessBrd.MoveBackward: Boolean;
begin
     if (_whiteToMove)then
        Result:=(GotoMove  (_currentMove-1, FALSE))
     else Result:=(GotoMove (_currentMove, TRUE));
end;

function TChessBrd.MoveForward: Boolean;
begin
    if (_whiteToMove)then
        Result:=GotoMove  (_currentMove, FALSE)
    else Result:=GotoMove (_currentMove+1, TRUE);
end;


function TChessBrd.GotoMove (moveno: Integer; whiteMoves: Boolean): Boolean;
var
    oldpos: String;
begin

    Result:=FALSE;

    if (moveno>LastMove)or((moveno=LastMove)and(LastTurn)and(whiteMoves=FALSE))then
        begin
        Exit;
        end;

    if (moveno<FirstMove)or((moveno=FirstMove)and(FirstTurn=FALSE)and(whiteMoves))then
        begin
        Exit;
        end;

    _currentMove:=moveno;
    _whiteToMove:=whiteMoves;

    oldpos   :=_position;
    _position :=MoveList[_currentMove][1-Integer(_whiteToMove)].position;
    _enPassant:=MoveList[_currentMove][1-Integer(_whiteToMove)].EnPassant;
    _castlingallowed :=MoveList[_currentMove][1-Integer(_whiteToMove)].Castling;

    UpdateChessBoard(oldpos);

    Result:=TRUE;
end;

function  TChessBrd.LegalMoveAvailable: Boolean;
var
    r,n: Square;
begin
    Result:=TRUE;

    if (_whiteToMove)then
        begin
        for r:=A8 to H1 do
        if (ColorOfPieceOnSquare(Square(r))=White)then
            begin
            for n:=A8 to H1 do
                 if (MoveIsLegal(r,n)and (WhiteInCheckAfter(r,n)=FALSE))then
                     Exit;
            end;
        end
    else
        begin
        for r:=A8 to H1 do
        if (ColorOfPieceOnSquare(r)=Black)then
            begin
            for n:=A8 to H1 do
                 if (MoveIsLegal(r,n)and (BlackInCheckAfter(r,n)=FALSE))then
                    Exit;
            end;
        end;

    Result:=FALSE;
end;

//-------------------------------------------------------------
//-------------------PRIVATE Stuff here------------------------
//-------------------------------------------------------------

function TChessBrd.MoveIsLegal (oldsq, newsq: Square): Boolean;
var
    piece: Char;
    Label Out;
begin
    Result:=FALSE;

    piece:=Char(Integer(UpCase(_position[Integer(oldsq)]))+32);

    if (oldsq<A8)or(oldsq>H1)or(newsq<A8)or(newsq>H1)then
        Exit;

    //Turn to the right color ?
    if (_whiteToMove)then
        begin
        if (ColorOfPieceOnSquare(oldsq)<>White) then Exit;
        end
    else
        begin
        if (ColorOfPieceOnSquare(oldsq)<>Black) then Exit;
        end;
    // Old int does't contain a piece ?
    if (ColorOfPieceOnSquare(oldsq)=NoPiece) then
         Exit;

    // Can't take piece of own color
    if (ColorOfPieceOnSquare(oldsq)=ColorOfPieceOnSquare(newsq)) then
           Exit;

    Result:=TRUE;

    case piece of
        'p':  begin
              Result:=CheckLegalPawnMove(oldsq,newsq);
              goto Out;
              end;
        'n':  begin
              Result:=CheckLegalKnightMove(oldsq,newsq);
              goto Out;
              end;
        'b':  begin
              Result:=CheckLegalBishopMove(oldsq,newsq);
              goto Out;
              end;
        'r':  begin
              Result:=CheckLegalRookMove(oldsq,newsq);
              goto Out;
              end;
        'q':  begin
              Result:=CheckLegalQueenMove(oldsq,newsq);
              goto Out;
              end;
        'k':  begin
              Result:=CheckLegalKingMove(oldsq,newsq);
              goto Out;
              end;
    end;

    Out:

    if (Result)then
        begin
        if (ColorOfPieceOnSquare(oldsq)=Black)and
            (BlackInCheckAfter(oldsq,newsq)=TRUE)then
             Result:=FALSE
        else if (ColorOfPieceOnSquare(oldsq)=White)and
            (WhiteInCheckAfter(oldsq,newsq)=TRUE)then
             Result:=FALSE;
        end;
end;

//Checks whether black is in check after the specified move
function TChessBrd.BlackInCheckAfter(oldsq, newsq: Square): Boolean;
var
    v: Char;
    pos: String;
    x,y,r,res: Integer;
    kingsq: Square;
begin
    Result:=TRUE;

    pos:=_position;

    v:=pos[Integer(oldsq)];
    pos[Integer(oldsq)]:=' ';
    pos[Integer(newsq)]:=v;

    for r:=1 to 65 do
     begin
     res:=r;
     if (pos[r]='k') then
         Break;
     end;

    kingsq:=Square(res);
    if (res>64) then
        Exit;

    //pawn check
    for x:=-1 to 1 do
     if (x<>0)and(pos[Integer(MouseToSquare(XPos(kingsq)+x,YPos(kingsq)-1))]='P') then
       Exit;

    //knight check
    for y:=-1 to 1 do
     if (abs(y)=1) then
      for x:=-2 to 2 do
       if (abs(x)=2)and(pos[Integer(MouseToSquare(XPos(kingsq)+x,YPos(kingsq)+y))]='N')then
         Exit;

    for y:=-2 to 2 do
     if (abs(y)=2) then
      for x:=-1 to 1 do
       if (x<>0)and(pos[Integer(MouseToSquare(XPos(kingsq)+x,YPos(kingsq)+y))]='N')then
            Exit;

    //bishop (and queen) check
    for y:=-1 to 1 do
     if (y<>0) then
      for x:=-1 to 1 do
       if (x<>0) then
        for r:=1 to 7 do
         begin
          v:=pos[Integer(MouseToSquare(XPos(kingsq)+x*r,YPos(kingsq)+y*r))];
          if (v='Q')or(v='B')or((v='K')and(r=1))then
             Exit;
          if (v<>' ') then
             Break;
         end;

    //rook (and queen) check
    for x:=-1 to 1 do
     if (x<>0) then
        for r:=1 to 7 do
        begin
             v:=pos[Integer(MouseToSquare(XPos(kingsq)+x*r,YPos(kingsq)))];
             if (v='Q')or(v='R')or((v='K')and(r=1))then
             begin
                Exit;
             end;
             if (v<>' ')then
               Break;
        end;

    for y:=-1  to 1 do
      begin
      if (y<>0) then
        for r:=1  to 7 do
        begin
             v:=pos[Integer(MouseToSquare(XPos(kingsq),YPos(kingsq)+r*y))];
             if (v='Q')or(v='R')or ((v='K')and(r=1))then
                 begin
                 Result:=TRUE;
                 Exit;
                 end;
             if (v<>' ')then
                 Break;
        end;
      end;

Result:=FALSE;
end;

//Checks whether white is in check after the specified move
function TChessBrd.WhiteInCheckAfter(oldsq, newsq: Square): Boolean;
var
    x,y,r,res: Integer;
    kingsq: Square;
    v: Char;
    pos: String;
begin
    Result:=TRUE;

    pos:=_position;

    //Can move to same Square
    v:=pos[Integer(oldsq)];
    pos[Integer(oldsq)]:=' ';
    pos[Integer(newsq)]:=v;

    for r:=1 to 65 do
       begin
       res:=r;
       if (pos[r]='K') then
          Break;
       end;
    kingsq:=Square(res);
    if (res>64) then
        begin
        Result:=FALSE;
        Exit;
        end;

    //pawn check
    for x:=-1 to 1 do
        if (x<>0)and(pos[Integer(MouseToSquare(XPos(kingsq)+x,YPos(kingsq)+1))]='p') then
            Exit;

    //knight check
    for y:=-1 to 1 do
     if(y<>0) then
      for x:=-2 to 2 do
        if (abs(x)=2)and(pos[Integer(MouseToSquare(XPos(kingsq)+x,YPos(kingsq)+y))]='n')then
            Exit;

    for y:=-2 to 2 do
     if (abs(y)=2)then
      for x:=-1 to 1 do
        if (x<>0)and (pos[Integer(MouseToSquare(XPos(kingsq)+x,YPos(kingsq)+y))]='n')then
          Exit;


    //bishop (and queen) check
    for y:=-1  to 1 do
      if (y<>0) then
        for x:=-1 to 1 do
            if (x<>0) then
             for r:=1 to 7 do
             begin
                  v:=pos[Integer(MouseToSquare(XPos(kingsq)+x*r,YPos(kingsq)+y*r))];
                  if (v='q')or(v='b')or ((v='k')and(r=1))then
                     Exit;
                  if (v<>' ') then
                    Break;
             end;

    //rook (and queen) check
    for x:=-1 to 1 do
     if (x<>0) then
       for r:=1 to 7 do
        begin
             v:=pos[Integer(MouseToSquare(XPos(kingsq)+x*r,YPos(kingsq)))];
             if (v='q')or(v='r')or((v='k')and(r=1))then
                Exit;
             if (v<>' ')then
                 break;
        end;

    for y:=-1 to 1 do
      if (y<>0) then
        for r:=1 to 7 do
          begin
            v:=pos[Integer(MouseToSquare(XPos(kingsq),YPos(kingsq)+r*y))];
            if (v='q')or(v='r')or((v='k')and(r=1))then
              Exit;
            if (v<>' ') then
              Break;
          end;

Result:=FALSE;
end;

//Checks whether a pawn move is legal, assuming the king isn't in check
function TChessBrd.CheckLegalPawnMove(oldsq,newsq: Square): Boolean;
var
    deltax: Integer;
begin
    Result:=TRUE;

    deltax:=abs(XPos(newsq)-XPos(oldsq));

    if (ColorOfPieceOnSquare(oldsq)=White)then
    begin
        if (YPos(newsq)=YPos(oldsq)+1)then
        begin
            if (deltax=0)then
                begin
                if (_position[Integer(newsq)]=' ')then
                    Exit;
                end
            else if (deltax=1)then
                begin
                if (ColorOfPieceOnSquare(newsq)=Black)or
                    (_enPassant=newsq)then
                    Exit;
                end;
        end
        else if ((YPos(oldsq)=2)and (YPos(newsq)=4))then
        begin
            if (deltax=0)and(_position[Integer(newsq)]=' ')
               and  (_position[(Integer(oldsq)+Integer(newsq)) shr 1]=' ') then Exit;
        end;

    end
    else if (ColorOfPieceOnSquare(oldsq)=Black)then
    begin
        if (YPos(newsq)=YPos(oldsq)-1)then
        begin
            if (deltax=0)then
                begin
                if (_position[Integer(newsq)]=' ') then Exit;
                end
            else if (deltax=1)then
                begin
                if (ColorOfPieceOnSquare(newsq)=White)or
                    (_enPassant=newsq)then
                     Exit;
                end;
        end
        else if (YPos(oldsq)=7)and(YPos(newsq)=5)then
        begin
            if (deltax=0)and(_position[Integer(newsq)]=' ')and
                (_position[Integer((Integer(oldsq)+Integer(newsq)) shr 1)]=' ')then
                    Exit;
        end;
    end;

Result:=FALSE;
end;

//Checks whether a knight move is legal, assuming the king isn't in check
function TChessBrd.CheckLegalKnightMove(oldsq,newsq: Square): Boolean;
var
    deltax,deltay: Integer;
begin
    deltax:=abs(XPos(oldsq)-XPos(newsq));
    deltay:=abs(YPos(oldsq)-YPos(newsq));
    if (((deltax=2)and (deltay=1))or((deltax=1)and (deltay=2)))
        then Result:=TRUE
    else Result:=FALSE;
end;

//Checks whether a bishop move is legal, assuming the king isn't in check
function  TChessBrd.CheckLegalBishopMove(oldsq, newsq: Square): Boolean;
var
    x,y,r,n,m,p: Integer;
    v: Square;
begin
    Result:=TRUE;

    x:=XPos(oldsq);
    y:=YPos(oldsq);
    p:=ColorOfPieceOnSquare(oldsq);

    for n:=-1 to 1 do
      if (n<>0) then
        for m:=-1 to 1 do
          if (m<>0) then
            for r:=1 to 7 do
             begin
              v:=MouseToSquare(x+n*r,y+m*r);
              if ((ColorOfPieceOnSquare(v)=NoPiece)or(ColorOfPieceOnSquare(v)<>p))then
               begin
                if (v=newsq) then Exit;
               end;
              if (ColorOfPieceOnSquare(v)<>NoPiece)then break;
             end;

    Result:=FALSE;
end;

//Checks whether a rook move is legal, assuming the king isn't in check
function TChessBrd.CheckLegalRookMove(oldsq, newsq: Square): Boolean;
var
    x,y,r,n,p: Integer;
    v: Square;
begin
    Result:=TRUE;

    x:=XPos(oldsq);
    y:=YPos(oldsq);
    p:=ColorOfPieceOnSquare(oldsq);

    for n:=-1 to 1 do
     if (n<>0) then
        for r:=1 to 7 do
         begin
            v:=MouseToSquare(x+n*r,y);
            if ((ColorOfPieceOnSquare(v)=NoPiece)or(ColorOfPieceOnSquare(v)<>p))then
            begin
                if (v=newsq) then Exit;
            end;
            if (ColorOfPieceOnSquare(v)<>NoPiece) then Break;
        end;

    for n:=-1 to 1 do
     if (n<>0) then
        for r:=1 to 7 do
        begin
            v:=MouseToSquare(x,y+n*r);
            if ((ColorOfPieceOnSquare(v)=NoPiece)or(ColorOfPieceOnSquare(v)<>p))then
            begin
                if (v=newsq) then Exit;
            end;
            if (ColorOfPieceOnSquare(v)<>NoPiece) then Break;
        end;

    Result:=FALSE;
end;

//Checks whether a queen move is legal, assuming the king isn't in check
function TChessBrd.CheckLegalQueenMove(oldsq, newsq: Square): Boolean;
begin
    Result:=CheckLegalBishopMove(oldsq,newsq);
    if (Result=FALSE) then Result:=CheckLegalRookMove(oldsq,newsq);
end;

//Checks whether a king move is legal, assuming the king isn't in check
function TChessBrd.CheckLegalKingMove(oldsq,newsq: Square): Boolean;
var
    deltax,deltay: Integer;
begin
    Result:=TRUE;

    deltax:=abs(XPos(oldsq)-XPos(newsq));
    deltay:=abs(YPos(oldsq)-YPos(newsq));

    if ((deltax<=1)and (deltay<=1))then
        Exit;

    if (Square(oldsq)=E1)and(Square(newsq)=G1)then
    begin
        if (WhiteKingSide in _castlingallowed) and  (_position[Integer(H1)]='R') and
            (_position[Integer(G1)]=' ') and  (_position[Integer(F1)]=' ') and
            (not WhiteInCheckAfter(E1,E1)) and
            (not WhiteInCheckAfter(E1,F1)) and
            (not WhiteInCheckAfter(E1,G1)) then
                Exit;
    end;
    if (oldsq=E1)and(newsq=C1) then
    begin
        if (WhiteQueenSide in _castlingallowed) and(_position[Integer(A1)]='R')and
           (_position[Integer(B1)]=' ')and(_position[Integer(C1)]=' ')and (_position[Integer(D1)]=' ') and
            (not WhiteInCheckAfter(E1,E1)) and
            (not WhiteInCheckAfter(E1,D1)) and
            (not WhiteInCheckAfter(E1,C1)) then
                Exit;
    end;
    if (oldsq=E8)and(newsq=G8)then
    begin
        if (BlackKingSide in _castlingallowed)and(_position[Integer(H8)]='r')and
            (_position[Integer(G8)]=' ')and(_position[Integer(F8)]=' ') and
            (not BlackInCheckAfter(E8,E8)) and
            (not BlackInCheckAfter(E8,F8))and
            (not BlackInCheckAfter(E8,G8)) then
                Exit;
    end;
    if (oldsq=E8)and(newsq=C8)then
    begin
        if (BlackQueenSide in _castlingallowed) and (_position[Integer(A8)]='r') and
           (_position[Integer(B8)]=' ')and(_position[Integer(C8)]=' ')and
           (_position[Integer(D8)]=' ')and
            (not BlackInCheckAfter(E8,E8)) and
            (not BlackInCheckAfter(E8,D8)) and
            (not BlackInCheckAfter(E8,C8)) then
                Exit;
    end;

    Result:=FALSE;
end;

//procedure TChessBrd.GetPieceBitmap(piece: Char, bool GetWhitePiece, bool GetMask);
//begin
//    int index=0;
//
//    case
//
//end;



procedure  TChessBrd.InitializeBitmaps;
const
set32: Array[0..3101] of Integer = (
$30764D42,$00000000,$00760000,$00280000,$00C00000,$00800000,$00010000,
$00000004,$30000000,$00000000,$00000000,$00100000,$00100000,$00000000,
$00000000,$80000080,$80000000,$00800080,$00800000,$80800080,$C0C00000,
$808000C0,$00000080,$FF0000FF,$FF000000,$00FF00FF,$00FF0000,$FFFF00FF,
$FFFF0000,$FFFF00FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$88FFFFFF,$88888888,
$88888888,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,$88888888,$FFFFFF8F,$FFFFFFFF,
$FFFFFFFF,$88888888,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$FFFFFF08,$88888880,$88FF8F88,
$80888888,$FFFFFFFF,$888888FF,$88888888,$8F888888,$FFFFFFFF,$8088F8FF,
$00000000,$FFFF8808,$FFFFFFFF,$88F8FFFF,$00000000,$FFFF8808,$FFFFFFFF,
$8888FFFF,$88888888,$FF8F8888,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,
$FFFFFF08,$00000080,$00F80800,$00000000,$FFFFFF8F,$000000F8,$00000000,
$08000000,$FFFFFFFF,$0000F8FF,$00000000,$FF8F0000,$FFFFFFFF,$0080FFFF,
$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,
$80FFFFFF,$00000000,$00000000,$F8FFFF08,$00000000,$00800000,$00000000,
$FFFFFF08,$000000F8,$00000000,$08000000,$FFFFFFFF,$000080FF,$00000000,
$FF080000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,
$00000000,$FF080000,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$F8FFFF08,
$00000000,$00000000,$00000000,$FFFFFF08,$000000F8,$00000000,$08000000,
$FFFFFFFF,$000080FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$F8FFFFFF,
$00000000,$00000000,$FFFFFF08,$88888888,$00000080,$88888888,$FFFFFF8F,
$000000F8,$00000000,$08000000,$FFFFFFFF,$0000F8FF,$00000000,$FF8F0000,
$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$FFFFFFFF,$F88FF8FF,$00000000,$00000000,$FFFFFF08,$F8FFFFFF,
$00000080,$FFFFFFFF,$FFFFFFFF,$000088FF,$00000000,$8F080000,$FFFFFFFF,
$0000F8FF,$00000000,$FF8F0000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,
$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$FF088088,$00000080,
$00000000,$FFFFFF08,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$0000F8FF,
$00000000,$FF080000,$FFFFFFFF,$0000F8FF,$00000000,$FF8F0000,$FFFFFFFF,
$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,
$F8FFFFFF,$8F000000,$000000F8,$00000000,$FFFFFF08,$80FFFFFF,$00000000,
$FFFF8F00,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$000080FF,
$00000000,$FF080000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,
$0080FFFF,$00000000,$FF8F0000,$F8FFFFFF,$8F000000,$000080FF,$00000000,
$FFFFFF08,$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$0080FFFF,$00000000,
$FF8F0000,$FFFFFFFF,$000080FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,
$00000000,$FF080000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$80FFFFFF,
$08000000,$0000808F,$00000000,$FFFFFF8F,$F8FFFFFF,$00000000,$FFFFFF08,
$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$FFFFFFFF,$000000F8,$00000000,
$8F000000,$FFFFFFFF,$000080FF,$00000000,$8F000000,$FFFFFFFF,$00F8FFFF,
$00000000,$FFFF0800,$80FFFFFF,$00000000,$00008808,$00000000,$FFFFFF8F,
$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,
$FFFFFFFF,$000000F8,$00000000,$8F000000,$FFFFFFFF,$000000F8,$00000000,
$08000000,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$F8FFFFFF,$00000000,
$00000800,$00000000,$FFFFFF8F,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,
$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$000000F8,$00000000,$8F000000,
$FFFFFFFF,$00000080,$00000000,$00000000,$FFFFFF8F,$F8FFFFFF,$00000080,
$FFFFFF88,$F8FFFFFF,$00000000,$00000000,$00000000,$FFFFFF8F,$00F8FFFF,
$00000000,$FFFF0800,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,
$00000080,$00000000,$08000000,$FFFFFFFF,$00000080,$00000000,$00000000,
$FFFFFF8F,$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$00000080,$00000000,
$08000000,$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$FFFFFFFF,$80FFFFFF,
$00000000,$FFFF8F00,$FFFFFFFF,$00000080,$00000000,$08000000,$F8FFFFFF,
$00000000,$00000000,$00000000,$FFFFFF08,$80FFFFFF,$00000000,$FFFF8F00,
$FFFFFFFF,$00000080,$00000000,$08000000,$FFFFFFFF,$0080FFFF,$00000000,
$FF8F0000,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$00800080,
$00000000,$08000800,$F8FFFFFF,$00000000,$00000000,$00000000,$FFFFFF08,
$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$000000F8,$00000000,$8F000000,
$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$80FFFFFF,$00000000,
$FFFF8F00,$F8FFFFFF,$00800000,$80000008,$00000800,$F8FFFF8F,$00000000,
$00000000,$00000000,$FFFFFF08,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,
$000000F8,$00000000,$8F000000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,
$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$F8FFFFFF,$00800800,$80000008,
$00800800,$FFFFFF8F,$00000080,$00000000,$00000000,$FFFFFF8F,$80FFFFFF,
$00000000,$FFFF8F00,$FFFFFFFF,$000080FF,$00000000,$FF080000,$FFFFFFFF,
$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,
$F8FFFFFF,$00808F00,$88000088,$00F80800,$FFFFFF8F,$00000080,$00000000,
$00000000,$FFFFFF8F,$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$000080FF,
$00000000,$FF080000,$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$FFFFFFFF,
$00F8FFFF,$00000000,$FFFF0800,$80FFFFFF,$00808F00,$88000088,$00F80800,
$FFFFFF08,$000080F8,$08000088,$88000080,$FFFFFFFF,$FFFFFFFF,$00000080,
$FFFFFF8F,$FFFFFFFF,$000080FF,$00000000,$FF8F0000,$FFFFFFFF,$00F8FFFF,
$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$80FFFFFF,
$0080FF08,$88000088,$80FF0800,$FFFFFF08,$8888F8FF,$08000088,$FF8888F8,
$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,
$FFFF0800,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$0000F8FF,
$00000000,$FF080000,$00F8FFFF,$08808F00,$FF0880FF,$00F80880,$FFFF8F00,
$FFFFFFFF,$8F0080FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00000080,$FFFFFF8F,
$FFFFFFFF,$0000F8FF,$00000000,$FFFF8F00,$FFFFFFFF,$F8FFFFFF,$00000080,
$FFFFFF88,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$00F8FFFF,$08808F00,
$FF0880FF,$00F80880,$FFFF8F00,$FFFFFFFF,$FF08F8FF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$00000080,$FFFFFF8F,$FFFFFFFF,$0080FFFF,$00000000,$FFFFFF08,
$FFFFFFFF,$FFFFFFFF,$8F0080FF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$80FFFFFF,$0880FF08,$FF0880FF,$80FF0880,$FFFFFF08,$FFFFFFFF,
$FF08F8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,
$0080FFFF,$08000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,
$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$F8FFFFFF,$0000F88F,$F808808F,
$F88F0000,$FFFFFF8F,$FFFFFFFF,$8F0888FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$8F0080FF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$8F880080,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$08000088,$FF080080,
$FFFFFFFF,$0000F8FF,$88000088,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$080000F8,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF88F8FF,$FFFFFFFF,$FFFFFFFF,$8800F8FF,
$FF8F0088,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,
$0000F8FF,$08000088,$FF080080,$FFFFFFFF,$0880FFFF,$8F0000F8,$FFFF0880,
$FFFFFFFF,$FFFFFFFF,$8F0888FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FF88FFFF,$FFFF08F8,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$8F0080FF,$FFFFFFFF,$FFFFFFFF,$8F88FFFF,$8F8888FF,$FF8F88FF,$FFFFFFFF,
$8FF8FFFF,$FF0880FF,$FFFF8FF8,$FFFFFFFF,$FFFFFFFF,$FF08F8FF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFF8FFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF88F8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF8FF8FF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FF8FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,
$00000000,$00000000,$FFFFFF0F,$FFFFFFF0,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,$00000000,$FFFFFF8F,
$FFFFFFFF,$FFFFFFFF,$00000000,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$00000000,$00000000,$FFFFFF0F,$000000F0,
$00FF0F00,$00000000,$FFFFFFFF,$000000FF,$00000000,$0F000000,$FFFFFFFF,
$0080FFFF,$00000000,$FFFF0800,$FFFFFFFF,$00F0FFFF,$00000000,$FFFF0000,
$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,$F0FFFFFF,$00000000,
$F00F0000,$FFFFFF0F,$00000000,$00F00000,$00000000,$FFFFFF0F,$000000FF,
$00000000,$0F000000,$FFFFFFFF,$0000F0FF,$00000000,$FF0F0000,$FFFFFFFF,
$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,
$FFFFFFFF,$F8FFFFFF,$00000000,$F00F0000,$FFFFFF0F,$00000000,$00000000,
$00000000,$FFFFFF0F,$000000FF,$00000000,$0F000000,$FFFFFFFF,$FF0FF8FF,
$00000000,$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,$000000F0,$FF0FFFFF,$FFFFFFFF,
$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,$FFFFFFFF,$00000000,$F00F0000,
$FFFFFF0F,$FFFFFFFF,$000F0FF0,$FFFFFFFF,$FFFFFFFF,$FFFF00FF,$FFFFFFFF,
$0FF0FFFF,$FFFFFFFF,$0F00FFFF,$FFFFFFFF,$FFFF00F0,$FFFFFFFF,$00F0FFFF,
$FFFFFFFF,$FFFF00F0,$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,
$FFFFFFFF,$00000080,$F00F0000,$FFFFFF0F,$FFFFFFFF,$000000F0,$FFFFFFFF,
$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,$F00FFFFF,$00000000,
$FFFFF00F,$FFFFFFFF,$FFF0FFFF,$00000000,$FFFFF00F,$FFFFFFFF,$0000FFFF,
$00000000,$FF0F0000,$FFFFFFFF,$FF0FF0FF,$000000F8,$F00F0000,$FFFFFF0F,
$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,
$FFFFFFFF,$FF00FFFF,$FFFFFFFF,$FFFF00FF,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,
$FFFF00FF,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$FF080000,
$000000FF,$F00F0000,$FFFFFF0F,$F0FFFFFF,$000000F0,$FFFFFFF0,$FFFFFFFF,
$FF00FFFF,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,$0000F0FF,$00000000,$FF0F0000,
$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,$00F0FFFF,$00000000,
$FFFF0000,$FFFFFFFF,$FF000F00,$0000F0FF,$F00F0000,$FFFFFF8F,$FFFFFFFF,
$FFFFFF0F,$FFFFFF0F,$FFFFFFFF,$00F0FFFF,$00000000,$FFFF0000,$FFFFFFFF,
$FF0FF0FF,$FFFFFFFF,$FF08F0FF,$FFFFFFFF,$FF0FFFFF,$00000000,$FF0FFF0F,
$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$F0FFFFFF,$0F000000,$0000F8FF,
$800F0000,$FFFFFFFF,$FFFFFFFF,$00000000,$FFFFFF0F,$FFFFFFFF,$00FFFFFF,
$FFFFFFFF,$FFFF0FF0,$FFFFFFFF,$000080FF,$00000000,$FF000000,$FFFFFFFF,
$0F0FF0FF,$FF0FFFFF,$FF000FFF,$FFFFFFFF,$00FFFFFF,$00000000,$FFFF0F00,
$F8FFFFFF,$0000000F,$0000FF0F,$00FF0000,$FFFFFFFF,$FFFFFFFF,$00000000,
$FFFFFF0F,$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,$000000FF,
$00000000,$FF000000,$FFFFFFFF,$00F000FF,$000F0F00,$0FF00000,$FFFFFFFF,
$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$00000000,$00000F00,$00FF0000,
$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFF00,$FFFFFFFF,$F0FFFFFF,$00000000,
$FFFFFF00,$FFFFFFFF,$000000FF,$00000000,$FF000000,$FFFFFFFF,$00000FF8,
$000F0F00,$080F0000,$FFFFFFFF,$FFFFFFFF,$000000F0,$FFFFFFFF,$FFFFFFFF,
$00000080,$00000000,$08FF0000,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,
$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,$000000F8,$00000000,
$8F000000,$FFFFFFFF,$0000F0F0,$000F0F00,$F0000000,$FFFFFFFF,$FFFFFFFF,
$00000000,$FFFFFF0F,$FFFFFFFF,$000000F0,$00000000,$0FFF0000,$FFFFFFFF,
$00FFFFFF,$00000000,$FFFF0F00,$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF00,
$FFFFFFFF,$000000F0,$00000000,$0F000000,$FFFFFFFF,$0000000F,$F0F0F000,
$0F000000,$FFFFFF0F,$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$000000F8,
$00F00000,$8FF00000,$FFFFFFFF,$00F8FFFF,$00F00000,$FFFF0800,$FFFFFFFF,
$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,$00F000F0,$00000000,$0F000F00,
$FFFFFFFF,$0000000F,$F0F0F000,$0F000000,$FFFFFF0F,$F0FFFFFF,$00000000,
$FFFFFF00,$FFFFFFFF,$000000FF,$000F0000,$FFF00F00,$FFFFFFFF,$00F0FFFF,
$00F00000,$FFFF0000,$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,
$00F00880,$F000000F,$08800F00,$FFFFFFFF,$0000008F,$0FF0000F,$0F000000,
$FFFFFF8F,$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,$000080FF,$000F0000,
$FF080F00,$FFFFFFFF,$00F0FFFF,$F0FFFF00,$FFFF0000,$FFFFFFFF,$F0FFFFFF,
$00000000,$FFFFFF00,$FFFFFFFF,$00F08F00,$F800008F,$00F80F00,$FFFFFFFF,
$0F00F0F0,$000F0FF0,$F00000FF,$FFFFFFFF,$F8FFFFFF,$00000000,$FFFFFF08,
$FFFFFFFF,$000FF0FF,$00000000,$FF0FFF00,$FFFFFFFF,$00F8FFFF,$00F00000,
$FFFF0800,$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF00,$FFFFFFFF,$00F0FF00,
$FF0000FF,$00FF0F00,$FFFFFFFF,$F0FF0FF8,$000F0F00,$08FFFF00,$FFFFFFFF,
$FFFFFFFF,$00000000,$FFFFFF0F,$FFFFFFFF,$F00FF0FF,$00000000,$FF8FF00F,
$FFFFFFFF,$00FFFFFF,$00F00000,$FFFF0F00,$FFFFFFFF,$00FFFFFF,$FFFFFFFF,
$FFFF0FF0,$F8FFFFFF,$00F0FF08,$FF0000FF,$80FF0F00,$FFFFFF8F,$0800F8FF,
$0F0F0FFF,$FF0800F8,$FFFFFFFF,$FFFFFFFF,$000000F0,$FFFFFFFF,$FFFFFFFF,
$F000F8FF,$00000000,$FFFF00FF,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,
$FFFFFFFF,$00F0FFFF,$00000000,$FFFF0000,$F0FFFFFF,$08F0FF0F,$FF0880FF,
$F0FF0F80,$FFFFFF0F,$FFFFFFFF,$0FFF0FFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0F0000FF,$FFFFFFFF,$FFFFFFFF,$0000FFFF,$0F000000,$FFFF0FF0,$FFFFFFFF,
$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$FF00FFFF,$FFFFFFFF,$FF0FF0FF,
$00FFFFFF,$0FF0FF00,$FF0FF0FF,$00FF0FF0,$FFFFFF00,$FFFFFFFF,$FF00F0FF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$FFFFFFFF,$FFFFFFFF,$0F00FFFF,
$FF0F00F0,$FFFFFF00,$FFFFFFFF,$FFFFFFFF,$000000F0,$FFFFFFFF,$FFFFFFFF,
$0000FFFF,$00000000,$FF0F0000,$80FFFFFF,$0FF0FF08,$FF0FF0FF,$80FF0FF0,
$FFFFFF08,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$080000F8,
$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,
$FF00F0FF,$FFFFFFFF,$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,$F0FFFFFF,
$0FF0FF0F,$FF0FF0FF,$F0FF0FF0,$FFFFFF0F,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$0F0000FF,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$8F000000,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0F0000FF,$FFFFFFFF,$FFFFFFFF,$0000FFFF,
$00000000,$FF0F0000,$FFFFFFFF,$0000FFFF,$FF0FF0FF,$FFFF0000,$FFFFFFFF,
$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF08F8FF,$FFFFFFFF,
$FFFFFFFF,$00F0FFFF,$FFFF00F0,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0F0000FF,
$FFFFFFFF,$FFFFFFFF,$0000FFFF,$0F0000FF,$FF0F00F0,$FFFFFFFF,$0880FFFF,
$FF0000FF,$FFFF0880,$FFFFFFFF,$FFFFFFFF,$0F0000FF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF00FFFF,$FFFF00FF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$0F0000FF,$FFFFFFFF,$FFFFFFFF,$0000FFFF,$0F0000FF,
$FF0F00F0,$FFFFFFFF,$0FF0FFFF,$FF0880FF,$FFFF0FF0,$FFFFFFFF,$FFFFFFFF,
$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFF0FFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF00F0FF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF0FF0FF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$88FFFFFF,$88888888,$88888888,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,
$88888888,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$88888888,$FFFFFF8F,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,
$FFFFFF08,$88888880,$88FF8F88,$80888888,$FFFFFFFF,$888888FF,$88888888,
$8F888888,$FFFFFFFF,$8088F8FF,$00000000,$FFFF8808,$FFFFFFFF,$88F8FFFF,
$00000000,$FFFF8808,$FFFFFFFF,$8888FFFF,$88888888,$FF8F8888,$FFFFFFFF,
$80FFFFFF,$00000000,$00000000,$FFFFFF08,$00000080,$00F80800,$00000000,
$FFFFFF8F,$000000F8,$00000000,$08000000,$FFFFFFFF,$0000F8FF,$00000000,
$FF8F0000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,
$00000000,$FF080000,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$F8FFFF08,
$00000000,$00800000,$00000000,$FFFFFF08,$000000F8,$00000000,$08000000,
$FFFFFFFF,$000080FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$80FFFFFF,
$00000000,$00000000,$F8FFFF08,$00000000,$00000000,$00000000,$FFFFFF08,
$000000F8,$00000000,$08000000,$FFFFFFFF,$000080FF,$00000000,$FF080000,
$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,$FFFFFF08,$88888888,
$00000080,$88888888,$FFFFFF8F,$000000F8,$00000000,$08000000,$FFFFFFFF,
$0000F8FF,$00000000,$FF8F0000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,
$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$F88FF8FF,$00000000,
$00000000,$FFFFFF08,$F8FFFFFF,$00000080,$FFFFFFFF,$FFFFFFFF,$000088FF,
$00000000,$8F080000,$FFFFFFFF,$0000F8FF,$00000000,$FF8F0000,$FFFFFFFF,
$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,
$FFFFFFFF,$FF088088,$00000080,$00000000,$FFFFFF08,$80FFFFFF,$00000000,
$FFFF8F00,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,
$00000000,$FF8F0000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,
$0000F8FF,$00000000,$FF080000,$F8FFFFFF,$8F000000,$000000F8,$00000000,
$FFFFFF08,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$FFFFFFFF,$000080FF,$00000000,$FF080000,$FFFFFFFF,$0000F8FF,
$00000000,$FF080000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$F8FFFFFF,
$8F000000,$000080FF,$00000000,$FFFFFF08,$F8FFFFFF,$00000000,$FFFFFF08,
$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$000080FF,$00000000,
$FF080000,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$FFFFFFFF,$0080FFFF,
$00000000,$FF8F0000,$80FFFFFF,$08000000,$0000808F,$00000000,$FFFFFF8F,
$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,
$FFFFFFFF,$000000F8,$00000000,$8F000000,$FFFFFFFF,$000080FF,$00000000,
$8F000000,$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$80FFFFFF,$00000000,
$00008808,$00000000,$FFFFFF8F,$F8FFFFFF,$00000000,$FFFFFF08,$FFFFFFFF,
$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$000000F8,$00000000,$8F000000,
$FFFFFFFF,$000000F8,$00000000,$08000000,$FFFFFFFF,$80FFFFFF,$00000000,
$FFFF8F00,$F8FFFFFF,$00000000,$00000800,$00000000,$FFFFFF8F,$80FFFFFF,
$00000000,$FFFF8F00,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,
$000000F8,$00000000,$8F000000,$FFFFFFFF,$00000080,$00000000,$00000000,
$FFFFFF8F,$F8FFFFFF,$00000080,$FFFFFF88,$F8FFFFFF,$00000000,$00000000,
$00000000,$FFFFFF8F,$00F8FFFF,$00000000,$FFFF0800,$FFFFFFFF,$80FFFFFF,
$00000000,$FFFF8F00,$FFFFFFFF,$00000080,$00000000,$08000000,$FFFFFFFF,
$00000080,$00000000,$00000000,$FFFFFF8F,$F8FFFFFF,$00000000,$FFFFFF08,
$FFFFFFFF,$00000080,$00000000,$08000000,$FFFFFFFF,$00F8FFFF,$00000000,
$FFFF0800,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$00000080,
$00000000,$08000000,$F8FFFFFF,$00000000,$00000000,$00000000,$FFFFFF08,
$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$00000080,$00000000,$08000000,
$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$80FFFFFF,$00000000,
$FFFF8F00,$FFFFFFFF,$00800080,$00000000,$08000800,$F8FFFFFF,$00000000,
$00000000,$00000000,$FFFFFF08,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,
$000000F8,$00000000,$8F000000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,
$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,$F8FFFFFF,$00800000,$80000008,
$00000800,$F8FFFF8F,$00000000,$00000000,$00000000,$FFFFFF08,$80FFFFFF,
$00000000,$FFFF8F00,$FFFFFFFF,$000000F8,$00000000,$8F000000,$FFFFFFFF,
$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,$80FFFFFF,$00000000,$FFFF8F00,
$F8FFFFFF,$00800800,$80000008,$00800800,$FFFFFF8F,$00000080,$00000000,
$00000000,$FFFFFF8F,$80FFFFFF,$00000000,$FFFF8F00,$FFFFFFFF,$000080FF,
$00000000,$FF080000,$FFFFFFFF,$0080FFFF,$00000000,$FF8F0000,$FFFFFFFF,
$80FFFFFF,$00000000,$FFFF8F00,$F8FFFFFF,$00808F00,$88000088,$00F80800,
$FFFFFF8F,$00000080,$00000000,$00000000,$FFFFFF8F,$F8FFFFFF,$00000000,
$FFFFFF08,$FFFFFFFF,$000080FF,$00000000,$FF080000,$FFFFFFFF,$00F8FFFF,
$00000000,$FFFF0800,$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$80FFFFFF,
$00808F00,$88000088,$00F80800,$FFFFFF08,$000080F8,$08000088,$88000080,
$FFFFFFFF,$FFFFFFFF,$00000080,$FFFFFF8F,$FFFFFFFF,$000080FF,$00000000,
$FF8F0000,$FFFFFFFF,$00F8FFFF,$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,
$00000000,$FF8F0000,$80FFFFFF,$0080FF08,$88000088,$80FF0800,$FFFFFF08,
$8888F8FF,$08000088,$FF8888F8,$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,
$FFFFFFFF,$0000F8FF,$00000000,$FFFF0800,$FFFFFFFF,$80FFFFFF,$00000000,
$FFFF8F00,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$00F8FFFF,$08808F00,
$FF0880FF,$00F80880,$FFFF8F00,$FFFFFFFF,$8F0080FF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$00000080,$FFFFFF8F,$FFFFFFFF,$0000F8FF,$00000000,$FFFF8F00,
$FFFFFFFF,$F8FFFFFF,$00000080,$FFFFFF88,$FFFFFFFF,$0000F8FF,$00000000,
$FF080000,$00F8FFFF,$08808F00,$FF0880FF,$00F80880,$FFFF8F00,$FFFFFFFF,
$FF08F8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00000080,$FFFFFF8F,$FFFFFFFF,
$0080FFFF,$00000000,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$8F0080FF,$FFFFFFFF,
$FFFFFFFF,$0000F8FF,$00000000,$FF080000,$80FFFFFF,$0880FF08,$FF0880FF,
$80FF0880,$FFFFFF08,$FFFFFFFF,$FF08F8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$080000F8,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$08000000,$FFFFFF8F,$FFFFFFFF,
$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,$FF080000,
$F8FFFFFF,$0000F88F,$F808808F,$F88F0000,$FFFFFF8F,$FFFFFFFF,$8F0888FF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$8F0080FF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,
$8F880080,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,
$0000F8FF,$08000088,$FF080080,$FFFFFFFF,$0000F8FF,$88000088,$FF8F0000,
$FFFFFFFF,$FFFFFFFF,$080000F8,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF88F8FF,
$FFFFFFFF,$FFFFFFFF,$8800F8FF,$FF8F0088,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$080000F8,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$08000088,$FF080080,$FFFFFFFF,
$0880FFFF,$8F0000F8,$FFFF0880,$FFFFFFFF,$FFFFFFFF,$8F0888FF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF88FFFF,$FFFF08F8,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$8F0080FF,$FFFFFFFF,$FFFFFFFF,$8F88FFFF,
$8F8888FF,$FF8F88FF,$FFFFFFFF,$8FF8FFFF,$FF0880FF,$FFFF8FF8,$FFFFFFFF,
$FFFFFFFF,$FF08F8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF8FFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF88F8FF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FF8FF8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF8FFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$00000000,$00000000,$FFFFFF0F,$FFFFFFF0,
$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$F8FFFFFF,$00000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$00000080,$FFFFFF8F,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,
$F0FFFFFF,$FFFFFF0F,$000000F0,$00FF0F00,$00000000,$FFFFFFFF,$000000FF,
$00000000,$0F000000,$FFFFFFFF,$0800FFFF,$FFFFFFFF,$FFFF0080,$FFFFFFFF,
$00F0FFFF,$FFFFFF8F,$FFFF0080,$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,
$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFF0F,$0FF000FF,
$FFFFFFFF,$FFFFFF0F,$FFFF0FFF,$FFFFFFFF,$0FFFFFFF,$FFFFFFFF,$FF0FF0FF,
$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,$FF00FFFF,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,
$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$F8FFFFFF,$FFFFFF8F,$F0FFFFFF,
$FFFFFF0F,$00000000,$FF00F00F,$00000000,$FFFFFF0F,$FFFF0FFF,$FFFFFFFF,
$0FFFFFFF,$FFFFFFFF,$0800F0FF,$FFFFFFFF,$FF0F0080,$FFFFFFFF,$0000FFFF,
$FFFFFF8F,$FF0F0080,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFF0F,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$000F0FF0,$FFFFFFFF,
$FFFFFFFF,$000000FF,$00000000,$0F000000,$FFFFFFFF,$F80FFFFF,$00000000,
$FFFFF08F,$FFFFFFFF,$FFF0FFFF,$00000080,$FFFFF08F,$FFFFFFFF,$FF0FFFFF,
$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFF80,$F0FFFFFF,$FFFFFF0F,
$FFFFFFFF,$000000F0,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,
$FFFFFFFF,$8F00FFFF,$FFFFFFFF,$FFFF00F8,$FFFFFFFF,$08F0FFFF,$FFFFFFFF,
$FFFF00F8,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FF0FF0FF,
$FFFF0FF8,$F0FFFFFF,$FFFFFF0F,$F0FFFFFF,$F0FFFF00,$FFFFFF00,$FFFFFFFF,
$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$800FFFFF,$00000000,$FFFFF008,
$FFFFFFFF,$F8F0FFFF,$00000000,$FFFFF008,$FFFFFFFF,$FF80FFFF,$FFFFFFFF,
$FF8FF0FF,$FFFFFFFF,$FF080000,$FFFF88FF,$F0FFFFFF,$FFFFFF0F,$F0FFFFFF,
$F0FFFF00,$FFFFFF00,$FFFFFFFF,$0000FFFF,$00000000,$FF0F0000,$FFFFFFFF,
$FF0FF8FF,$FFFFFFFF,$FF8FF0FF,$FFFFFFFF,$FF00FFFF,$FFFFFFFF,$FF0FF0FF,
$FFFFFFFF,$FFF0FFFF,$FFFFFFFF,$FFFFF0FF,$FFFFFFFF,$FFF0F000,$FFFFF0FF,
$F0FFFFFF,$FFFFFF8F,$FFFFFFFF,$0F00000F,$FFFFFF0F,$FFFFFFFF,$0FF0FFFF,
$FFFFFFFF,$FFFF00FF,$FFFFFFFF,$00F8F0FF,$00000000,$FF0F8F00,$FFFFFFFF,
$0800FFFF,$FFFFFFFF,$FF0F00F8,$FFFFFFFF,$0FF8FFFF,$FFFFFFFF,$FFFF08FF,
$F0FFFFFF,$0FF0FF0F,$FF0FF8FF,$80FFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFF0F,
$FFFFFF0F,$FFFFFFFF,$00FFFFFF,$00000000,$FFFF0F00,$FFFFFFFF,$FFFF80FF,
$FFFFFFFF,$FF08FFFF,$FFFFFFFF,$F80FF0FF,$00000000,$FF00FF08,$FFFFFFFF,
$00FFFFFF,$FFFFFFFF,$FFFF0FF0,$F8FFFFFF,$80FF0F0F,$FF0FFF8F,$00FFFFFF,
$FFFFFFFF,$FFFFFFFF,$F0FFFF00,$FFFFFF0F,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,
$FFFFFFF0,$FFFFFFFF,$FF0800FF,$FFFFFFFF,$FF0080FF,$FFFFFFFF,$FFFF00FF,
$FF0FFFFF,$0FF0FFFF,$FFFFFFFF,$F8FFFFFF,$F0FFFF00,$FFFFFF08,$FFFFFFFF,
$FFFFFF0F,$FF0F0F80,$00FFFFFF,$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF00,
$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,$FFFFFFFF,$00F00FFF,$00000000,
$FFF00F00,$FFFFFFFF,$FFFF0FF8,$FF0FFFFF,$08FFFFFF,$FFFFFFFF,$FFFFFFFF,
$00FF0FF0,$FFFFFFFF,$FFFFFFFF,$FFFFFF00,$FF0F80FF,$08FFFFFF,$FFFFFFFF,
$80FFFFFF,$FFFFFFFF,$FFFF8FF0,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,
$FFFFFFFF,$FF000FF8,$0F8FF8F0,$8FF000FF,$FFFFFFFF,$FFFFFFF0,$FF0FFFFF,
$F0FFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFF00,$FFFFFF0F,$FFFFFFFF,$FFFFFFF0,
$FF0FFFFF,$0FF8FFFF,$FFFFFFFF,$0FFFFFFF,$FFFFFFFF,$FFFF0FFF,$FFFFFFFF,
$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,$FFFFFFFF,$FF00F0F0,$0FFFFFF0,$0F0F00FF,
$FFFFFFFF,$FFFFFF0F,$FFF8F8FF,$FFFFFFFF,$FFFFFF0F,$F8FFFFFF,$FFFFFF0F,
$FFFFFF08,$FFFFFFFF,$FFFF0FF8,$FF00FFFF,$8FF0FFFF,$FFFFFFFF,$8FF0FFFF,
$FF0FFFFF,$FFFF80FF,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,$FFFFFFFF,
$FFF0F0F0,$08FFFF80,$0F0F0FFF,$FFFFFFFF,$FFFFFF0F,$FFF0F0FF,$FFFFFFFF,
$FFFFFF0F,$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,$FFFFFFFF,$FFFF0FFF,$FFF0FFFF,
$FF80FFFF,$FFFFFFFF,$FFF0FFFF,$FF0FFFFF,$FFFFF0FF,$FFFFFFFF,$F0FFFFFF,
$FFFFFFFF,$FFFFFFF0,$FFFFFFFF,$FFF00880,$F0F00F0F,$08800FFF,$FFFFFFFF,
$FFFFFF8F,$0FF808FF,$FFFFFFFF,$FFFFFF8F,$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,
$FFFFFFFF,$FFFF80FF,$FFF8FFFF,$FF08FFFF,$FFFFFFFF,$FFF0FFFF,$0F0000FF,
$FFFFF0FF,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,$FFFFFFF0,$FFFFFFFF,$F0F08F00,
$F8F00F8F,$00F80F0F,$FFFFFFFF,$FFFFFFF0,$08FF0FF8,$F0FFFFFF,$FFFFFFFF,
$F8FFFFFF,$FFFFFF0F,$FFFFFF08,$FFFFFFFF,$FFF0F0FF,$FFFFFFFF,$FF0FF8FF,
$FFFFFFFF,$8FF8FFFF,$FF0FFFFF,$FFFF88FF,$FFFFFFFF,$F0FFFFFF,$FFFFFFFF,
$FFFFFFF0,$FFFFFFFF,$F0F0FF00,$FFF00FFF,$00FF0F0F,$FFFFFFFF,$F0FF0FF8,
$00FF0F00,$08FFFF00,$FFFFFFFF,$FFFFFFFF,$F0FFFF00,$FFFFFF0F,$FFFFFFFF,
$0FF0F0FF,$FFFFFFFF,$FF8FF0FF,$FFFFFFFF,$0FFFFFFF,$FF0FFFFF,$FFFF0FFF,
$FFFFFFFF,$00FFFFFF,$00000000,$FFFF0F00,$F8FFFFFF,$F0F0FF08,$FFF00FFF,
$80FF0F0F,$FFFFFF8F,$0800F8FF,$0FFF0FFF,$FF0800F8,$FFFFFFFF,$FFFFFFFF,
$00FF0FF0,$FFFFFFFF,$FFFFFFFF,$0F0FF8FF,$FFFFFFFF,$FFFF00FF,$FFFFFFFF,
$80FFFFFF,$FFFFFFFF,$FFFF8FF0,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$FFFF00FF,
$F0FFFFFF,$08F0FF0F,$FF0880FF,$F0FF0F80,$FFFFFF0F,$FFFFFFFF,$0FFF0FFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0FFF0FFF,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,
$FFFFFFFF,$FFFF0FF0,$FFFFFFFF,$F8FFFFFF,$F0FFFF00,$FFFFFF08,$FFFFFFFF,
$0000FFFF,$00000000,$FF0F0000,$0FFFFFFF,$0FF0FFF0,$FF0FF0FF,$0FFF0FF0,
$FFFFFFF0,$FFFFFFFF,$FF00F0FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00FF0FF0,
$FFFFFFFF,$FFFFFFFF,$F080FFFF,$FF0FFFFF,$FFFFFF00,$FFFFFFFF,$FFFFFFFF,
$08F000F8,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FF0FFFFF,$8FFFFFFF,
$0FF0FFF8,$FF0FF0FF,$8FFF0FF0,$FFFFFFF8,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$08FF0FF8,$FFFFFFFF,$FFFFFFFF,$F0F0FFFF,$0000FF00,
$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FF00F0FF,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,
$FFFFFFFF,$FF0FFFFF,$F0FFFFFF,$0FF0FF0F,$FF0FF0FF,$F0FF0FF0,$FFFFFF0F,
$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0FF000FF,$FFFFFFFF,
$FFFFFFFF,$F0F0FFFF,$8F00F000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0FF000FF,
$FFFFFFFF,$FFFFFFFF,$F00FFFFF,$00FF0F00,$FF0FFF00,$FFFFFFFF,$F00FFFFF,
$FF0FF0FF,$FFFFF00F,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FF08F8FF,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$FFFFF0F0,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$0FFF0FFF,$FFFFFFFF,$FFFFFFFF,$F00FFFFF,$0FFF0FFF,
$FF0FFFF0,$FFFFFFFF,$F88FFFFF,$FFF00FFF,$FFFFF88F,$FFFFFFFF,$FFFFFFFF,
$0F0000FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FF00FFFF,$FFFF00FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0FF000FF,$FFFFFFFF,
$FFFFFFFF,$0000FFFF,$0F0000FF,$FF0F00F0,$FFFFFFFF,$0FF0FFFF,$FFF88FFF,
$FFFF0FF0,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFF0FFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FF00F0FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF0FFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0000FFFF);

set40: Array [0..4829] of Integer=(
$4B764D42,$00000000,$00760000,$00280000,$00F00000,$00A00000,$00010000,
$00000004,$4B000000,$00000000,$00000000,$00100000,$00100000,$00000000,
$00000000,$80000080,$80000000,$00800080,$00800000,$80800080,$C0C00000,
$808000C0,$00000080,$FF0000FF,$FF000000,$00FF00FF,$00FF0000,$FFFF00FF,
$FFFF0000,$FFFF00FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$888888F8,$88888888,
$88888888,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$888888FF,
$8F888888,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$8888F8FF,$FF8F8888,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$00000080,$00000000,$00000000,$FFFFFF08,$888808F8,$FF888888,$888888FF,
$8F808888,$FFFFFFFF,$8888F8FF,$88888888,$88888888,$FF8F8888,$FFFFFFFF,
$88FFFFFF,$00000088,$08000000,$FFFF8F88,$FFFFFFFF,$F8FFFFFF,$00008088,
$88080000,$FFFFFF8F,$FFFFFFFF,$8888FFFF,$88888888,$88888888,$FFFF8888,
$FFFFFFFF,$FFFFFFFF,$00000080,$00000000,$00000000,$FFFFFF08,$00000080,
$88000000,$00000088,$08000000,$FFFFFFFF,$000080FF,$00000000,$00000000,
$FF080000,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,
$80F8FFFF,$00000000,$00000000,$FFFF8F08,$FFFFFFFF,$0000F8FF,$00000000,
$00000000,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$00000080,$00000000,$00000000,
$FFFFFF08,$00000080,$00000000,$00000000,$08000000,$FFFFFFFF,$000080FF,
$00000000,$00000000,$FF080000,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FF8F0000,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,
$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$000000F8,
$00000000,$00000000,$F8FFFF08,$00000000,$00000000,$00000000,$00000000,
$FFFFFF8F,$000080FF,$00000000,$00000000,$FF080000,$FFFFFFFF,$0080FFFF,
$00000000,$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$00000000,
$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,
$FFFFFFFF,$000000F8,$00000000,$00000000,$FFFFFF08,$00000080,$00000000,
$00000000,$08000000,$FFFFFFFF,$000080FF,$00000000,$00000000,$FF080000,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,
$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0000F8FF,$00000000,$00000000,
$FF8F0000,$FFFFFFFF,$FFFFFFFF,$000080FF,$00000000,$00000000,$FFFFFF08,
$888888F8,$00008888,$88880000,$8F888888,$FFFFFFFF,$000080FF,$00000000,
$00000000,$FF080000,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$0000F8FF,
$00000000,$00000000,$FF8F0000,$FFFFFFFF,$8FF8FFFF,$000080FF,$00000000,
$00000000,$FFFFFF08,$FFFFFFFF,$008088FF,$FF880800,$FFFFFFFF,$FFFFFFFF,
$0088F8FF,$00000000,$00000000,$FF8F8800,$FFFFFFFF,$00F8FFFF,$00000000,
$00000000,$FFFF0800,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,
$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,$088088FF,
$0000F8FF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$000000F8,$88000000,
$FFFFFFFF,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,
$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$00F8FFFF,$00000000,
$00000000,$FFFF8F00,$FFFFFFFF,$0080F8FF,$00000000,$00000000,$FF8F0800,
$FFFFFFFF,$000000F8,$0080FF8F,$00000000,$00000000,$FFFFFF08,$F8FFFFFF,
$00000080,$00000000,$FFFFFF8F,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,
$FFFF8F00,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,
$0080FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FFFF0800,$FFFFFFFF,$00000080,$00F8FF08,$00000000,$00000000,
$FFFFFF08,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,
$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FF8F0000,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,
$0088FFFF,$00000000,$00000000,$FFFF8800,$F8FFFFFF,$00000000,$80FFFF08,
$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,
$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$0080FFFF,
$00000000,$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$00000000,
$FF8F0000,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$F8FFFFFF,
$00000000,$80FF8800,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$000000F8,
$8F000000,$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,$FFFFFF8F,
$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF080000,$FFFFFFFF,$000080FF,
$00000000,$00000000,$FF080000,$FFFFFFFF,$80F8FFFF,$00000000,$00000000,
$FFFF8F08,$F8FFFFFF,$00000000,$F8880000,$00000000,$00000000,$FFFFFF8F,
$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00000080,
$08000000,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF080000,
$FFFFFFFF,$000000F8,$00000000,$00000000,$8F000000,$FFFFFFFF,$88FFFFFF,
$00000000,$00000000,$FFFFFF88,$F8FFFFFF,$00000000,$88000000,$00000000,
$00000000,$FFFFFF8F,$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$000080FF,$00000000,
$00000000,$8F000000,$FFFFFFFF,$00000080,$00000000,$00000000,$08000000,
$FFFFFFFF,$F8FFFFFF,$00000080,$08000000,$FFFFFF8F,$FFFFFFFF,$00000080,
$00000000,$00000080,$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,
$000080FF,$00000000,$00000000,$8F000000,$FFFFFFFF,$00000080,$00000000,
$00000000,$08000000,$FFFFFFFF,$FFFFFFFF,$00008088,$88080000,$FFFFFFFF,
$FFFFFFFF,$00000080,$00000000,$00000000,$08000000,$FFFFFFFF,$F8FFFFFF,
$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,
$FFFFFFFF,$FFFFFFFF,$080080FF,$00000000,$00000000,$8F000008,$F8FFFFFF,
$00000000,$00000000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$000000F8,
$8F000000,$FFFFFFFF,$FFFFFFFF,$000000F8,$00000000,$00000000,$08000000,
$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,
$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$080080FF,$00000000,$00000000,
$8F000008,$F8FFFFFF,$00000000,$00000000,$00000000,$00000000,$FFFFFF8F,
$FFFFFFFF,$00000088,$88000000,$FFFFFFFF,$FFFFFFFF,$000000F8,$00000000,
$00000000,$08000000,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$FFFFFF08,
$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$880080FF,
$00080000,$00000800,$8F008008,$F8FFFFFF,$00000000,$00000000,$00000000,
$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,
$000080FF,$00000000,$00000000,$8F000000,$FFFFFFFF,$80FFFFFF,$00000000,
$00000000,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,
$FFFFFFFF,$880000F8,$00080000,$00000800,$08008008,$F8FFFFFF,$00000000,
$00000000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,
$FFFFFFFF,$FFFFFFFF,$000080FF,$00000000,$00000000,$8F000000,$FFFFFFFF,
$80FFFFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$00000080,
$08000000,$FFFFFFFF,$FFFFFFFF,$F80800F8,$00880000,$00800800,$0800F808,
$F8FFFFFF,$00000000,$00000000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,
$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,$00000000,
$FF080000,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,
$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$F80800F8,$00880000,
$00800800,$0800F808,$FFFFFFFF,$00000080,$00000000,$00000000,$08000000,
$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$0000F8FF,
$00000000,$00000000,$FF080000,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,
$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,
$F88F00F8,$00880000,$00800800,$0880FF08,$FFFFFFFF,$00000080,$00000000,
$00000000,$08000000,$FFFFFFFF,$FFFFFFFF,$00000088,$88000000,$FFFFFFFF,
$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,$F8FFFFFF,
$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,
$FFFFFF8F,$FFFFFFFF,$F88F00F8,$00F80800,$00F80800,$0880FF08,$FFFFFFFF,
$000000F8,$00000000,$00000000,$8F000000,$FFFFFFFF,$FFFFFFFF,$000000F8,
$8F000000,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FF8F0000,
$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,
$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$F8FF0880,$80FF0800,$00F88F00,
$00F8FF08,$FFFFFF8F,$000088FF,$00008800,$00880000,$FF880000,$FFFFFFFF,
$FFFFFFFF,$000080F8,$8F080000,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$F8FFFFFF,$F8FF0800,
$80FF0800,$00F88F00,$00F8FF08,$FFFFFF08,$8888FFFF,$0000F888,$888F0000,
$FFFF8888,$FFFFFFFF,$FFFFFFFF,$008088FF,$FF880800,$FFFFFFFF,$FFFFFFFF,
$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$000080FF,
$FF080000,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,
$80FFFFFF,$F88F0000,$80FF8F00,$80FF8F00,$0080FF08,$FFFF8F00,$FFFFFFFF,
$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FF8F0000,
$FFFFFFFF,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,
$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FFFF0800,$80FFFFFF,$F88F0000,$80FF8F00,$80FF8F00,$0080FF08,
$FFFF8F00,$FFFFFFFF,$00F8FFFF,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00000000,$08000000,
$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$0088FFFF,$FFFF8800,$FFFFFFFF,$FFFFFFFF,
$0080FFFF,$00000000,$00000000,$FFFF0800,$80FFFFFF,$808F0000,$80FF8F00,
$80FF8F00,$00808F00,$FFFF8F00,$FFFFFFFF,$80FFFFFF,$FFFFFF08,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,
$00000000,$8F080000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,
$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$F8FFFFFF,
$00F80800,$80FF0800,$00F88F00,$00F80800,$FFFFFF08,$FFFFFFFF,$80FFFFFF,
$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,
$FFFFFFFF,$80FFFFFF,$00000000,$FF8F8808,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FFFF0800,$FFFFFFFF,$00F88F88,$00F80800,$00F80800,$88FF0800,$FFFFFF8F,
$FFFFFFFF,$80F8FFFF,$FFFF8808,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080F8FF,
$FF8F0800,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00800800,$FFFFFF08,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$0080FFFF,
$00808800,$00880800,$FFFF0800,$FFFFFFFF,$00F8FFFF,$00F80800,$00F80800,
$FFFF0800,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$0088FFFF,$FFFF8800,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,$00808F08,
$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,
$FFFFFFFF,$0080FFFF,$00808F00,$00F80800,$FFFF0800,$FFFFFFFF,$80FFFFFF,
$00F88F00,$80FF0800,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$88F8FFFF,$FFFF8F88,$FFFFFFFF,$FFFFFFFF,
$00F8FFFF,$08F8FF8F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,
$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$88F8FFFF,$88F8FF88,$88F88F88,$FFFF8F88,
$FFFFFFFF,$F8FFFFFF,$80FFFF88,$F8FF8F00,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,
$80F8FFFF,$FFFF8808,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$88F8FFFF,$8FFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$88FFFFFF,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,$FFFFFF88,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$88FFFFFF,$FFFFFF88,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$00000000,$00000000,$FFFFFF0F,
$FFFF0FFF,$FFFFFFFF,$FFFFFFFF,$FFF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000FF,$0F000000,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$00008077,$77080000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$00000000,
$00000000,$FFFFFF0F,$000000F0,$FF070000,$000070FF,$0F000000,$FFFFFFFF,
$0000F0FF,$00000000,$00000000,$FF0F0000,$FFFFFFFF,$00FFFFFF,$00000000,
$00000000,$FFFF0F00,$FFFFFFFF,$80F7FFFF,$00000000,$00000000,$FFFF7F08,
$FFFFFFFF,$0000FFFF,$00000000,$00000000,$FFFF0000,$FFFFFFFF,$FFFFFFFF,
$000000F0,$00000000,$F00F0000,$FFFFFF0F,$00000070,$87000000,$00000078,
$07000000,$FFFFFFFF,$0000F0FF,$00000000,$00000000,$FF0F0000,$FFFFFFFF,
$00F0FFFF,$00000000,$00000000,$FFFF0000,$FFFFFFFF,$0070FFFF,$00000000,
$00000000,$FFFF0700,$FFFFFFFF,$0000FFFF,$00000000,$00000000,$FFFF0000,
$FFFFFFFF,$FFFFFFFF,$000000FF,$00000000,$F00F0000,$FFFFFF0F,$00000080,
$08000000,$00000080,$08000000,$FFFFFFFF,$0000F0FF,$00000000,$00000000,
$FF0F0000,$FFFFFFFF,$0FF0FFFF,$00000000,$00000000,$FFFF000F,$FFFFFFFF,
$0700F7FF,$00000080,$08000000,$FF7F0070,$FFFFFFFF,$0000FFFF,$00000000,
$00000000,$FFFF0000,$FFFFFFFF,$FFFFFFFF,$000000FF,$00000000,$F00F0000,
$FFFFFF0F,$000000F0,$00000000,$00000000,$0F000000,$FFFFFFFF,$0000F0FF,
$00000000,$00000000,$FF0F0000,$FFFFFFFF,$00FFFFFF,$000000FF,$0F000000,
$FFFF0FF0,$FFFFFFFF,$0070FFFF,$008077FF,$FF770800,$FFFF0700,$FFFFFFFF,
$0000FFFF,$00000000,$00000000,$FFFF0000,$FFFFFFFF,$FFFFFFFF,$0000F0FF,
$00000000,$F00F0000,$FFFFFF0F,$FFFFFFFF,$0070FFFF,$FFFF0700,$FFFFFFFF,
$FFFFFFFF,$0F00F0FF,$FFFFFFFF,$FFFFFFFF,$FF0F00F0,$FFFFFFFF,$00FFFFFF,
$FFFFFF00,$F0FFFFFF,$FFFF0F00,$FFFFFFFF,$00F7FFFF,$FFFF8700,$0078FFFF,
$FFFF7F00,$FFFFFFFF,$0000FFFF,$00000000,$00000000,$FFFF0000,$FFFFFFFF,
$FFFFFFFF,$0000F0FF,$00000000,$F00F0000,$FFFFFF0F,$FFFFFFFF,$80F7FFFF,
$FFFF7F08,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$00000000,$00000000,$FFFFFF00,
$FFFFFFFF,$00FFFFFF,$00000000,$00000000,$FFFF0F00,$FFFFFFFF,$00FFFFFF,
$000080F7,$7F080000,$FFFFFF00,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FFFF0800,$FFFFFFFF,$0FF0FFFF,$0000FFFF,$00000000,$F00F0000,$FFFFFF0F,
$FFFFFFFF,$000000F7,$7F000000,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$00000000,
$00000000,$FFFFFF00,$FFFFFFFF,$00FFFFFF,$000000FF,$0F000000,$FFFF0FF0,
$FFFFFFFF,$00F8FFFF,$FFFF7F08,$80F7FFFF,$FFFF8F00,$FFFFFFFF,$0070FFFF,
$00000000,$00000000,$FFFF0700,$FFFFFFFF,$000000FF,$00F0FFFF,$00000000,
$F00F0000,$FFFFFF0F,$FFFFFFFF,$00000000,$00000000,$FFFFFFFF,$FFFFFFFF,
$00FFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFF00,$FFFFFFFF,$00FFFFFF,$FFFFFF00,
$F0FFFFFF,$FFFF0F00,$FFFFFFFF,$00F0FFFF,$00000000,$00000000,$FFFF0F00,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$000F00F0,
$00FFFF0F,$00000000,$F00F0000,$FFFFFF0F,$FFFFFFFF,$0000FF70,$07FF0000,
$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$00000000,$00000000,$FFFFFF00,$FFFFFFFF,
$0FF0FFFF,$00000000,$00000000,$FFFF000F,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FFFF0800,$FFFFFFFF,$00F7FFFF,$00000000,$00000000,$FFFF7F00,
$FFFFFFFF,$000F0000,$F0FFFF0F,$00000000,$00FF0000,$FFFFFFFF,$FFFFFFFF,
$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$70FFFFFF,$00000000,$00000000,
$FFFFFF07,$FFFFFFFF,$0FF0FFFF,$000000FF,$0F000000,$FFFF00FF,$FFFFFFFF,
$0F00F7FF,$000088F7,$7F880000,$FF7F00F0,$FFFFFFFF,$80FFFFFF,$00000000,
$00000000,$FFFFFF08,$FFFFFFFF,$0000F000,$F0FFFF00,$00000000,$00FF0000,
$FFFFFFFF,$FFFFFFFF,$000000F7,$7F000000,$FFFFFFFF,$FFFFFFFF,$F7FFFFFF,
$FFFF0F00,$00F0FFFF,$FFFFFF7F,$FFFFFFFF,$0000FFFF,$FFFFFF0F,$FFFFFFFF,
$FF0F0000,$FFFFFFFF,$FF0070FF,$F0FFFF87,$78FFFF0F,$FF0700FF,$FFFFFFFF,
$70FFFFFF,$00000000,$00000000,$FFFFFF07,$FFFFFFFF,$0000FF00,$FFFF0000,
$00000000,$00FF0000,$FFFFFFFF,$FFFFFFFF,$00000FFF,$FFF00000,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$00000070,$07000000,$FFFFFFFF,$FFFFFFFF,$0000FFFF,
$00000000,$00000000,$FF0F0000,$FFFFFFFF,$F00F00F7,$F0000000,$0000000F,
$7F00F00F,$FFFFFFFF,$F7FFFFFF,$00000000,$00000000,$FFFFFF7F,$FFFFFFFF,
$00000F00,$FF000000,$00000000,$00FF0000,$FFFFFFFF,$FFFFFFFF,$FFFF00F7,
$7F00FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,
$FFFFFFFF,$0000F0FF,$00000000,$00000000,$FF000000,$FFFFFFFF,$00FF00F8,
$F0000000,$0000000F,$8F00FF00,$FFFFFFFF,$FFFFFFFF,$00000078,$87000000,
$FFFFFFFF,$FFFFFFFF,$000000F0,$00000000,$000000F0,$00FF0000,$FFFFFFFF,
$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,
$0F000000,$FFFFFFFF,$FFFFFFFF,$0000F0FF,$00000000,$00000000,$FF000000,
$FFFFFFFF,$00F00F70,$F0000000,$0000000F,$07F00F00,$FFFFFFFF,$FFFFFFFF,
$000070FF,$FF070000,$FFFFFFFF,$FFFFFFFF,$000000F0,$00000000,$00000000,
$0FF00F00,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,$FFFFFFFF,$0F00F0FF,$00000000,
$00000000,$FF00000F,$FFFFFFFF,$00007F80,$F0080000,$0000800F,$08F70000,
$FFFFFFFF,$FFFFFFFF,$000000F7,$7F000000,$FFFFFFFF,$FFFFFFFF,$000000FF,
$00000000,$00000000,$0FF00F00,$FFFFFFFF,$FFFFFFFF,$00000000,$00000000,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,$FFFFFFFF,
$0F00F0FF,$00000000,$00000000,$FF00000F,$FFFFFFFF,$0000F800,$800F0000,
$0000F008,$008F0000,$FFFFFFFF,$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,
$FFFFFFFF,$000000FF,$00000000,$0000000F,$0FF00F00,$FFFFFFFF,$F8FFFFFF,
$0F000000,$000000F0,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,
$FFFFFFFF,$FFFFFFFF,$FF00F0FF,$000F0000,$00000F00,$FF00F00F,$FFFFFFFF,
$0000F000,$008F0000,$0000F800,$000F0000,$FFFFFFFF,$FFFFFFFF,$000000F0,
$0F000000,$FFFFFFFF,$FFFFFFFF,$0000F0FF,$00000000,$0000000F,$FF00FF00,
$FFFFFFFF,$F0FFFFFF,$0F000000,$000000F0,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,
$000000F0,$0F000000,$FFFFFFFF,$FFFFFFFF,$FF0000FF,$000F0000,$00000F00,
$0F00F00F,$FFFFFFFF,$0000F000,$0FF80800,$00808FF0,$000F0000,$FFFFFFFF,
$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,$FFFFFFFF,$0000F0FF,$00000000,
$0000F000,$FF00FF00,$FFFFFFFF,$F0FFFFFF,$FF0F0000,$0000F0FF,$FFFFFF0F,
$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,$FFFFFFFF,$FF0F00FF,
$00FF0000,$00F00F00,$0F00FF0F,$FFFFFFFF,$0000F700,$07708F00,$00F80770,
$007F0000,$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,$FFFFFFFF,
$0F00FFFF,$00000000,$0000F000,$FF0FF00F,$FFFFFFFF,$F8FFFFFF,$FF0F0000,
$0000F0FF,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,$FFFFFFFF,
$FFFFFFFF,$FF0F00FF,$00FF0000,$00F00F00,$0F00FF0F,$FFFFFFFF,$00708F80,
$7000F78F,$F87F0007,$08F80700,$FFFFFFFF,$FFFFFFFF,$000000F0,$0F000000,
$FFFFFFFF,$FFFFFFFF,$0F00FFFF,$000000F0,$00000000,$FF0FF00F,$FFFFFFFF,
$F7FFFFFF,$0F000000,$000000F0,$FFFFFF7F,$FFFFFFFF,$FFFFFFFF,$00000070,
$07000000,$FFFFFFFF,$FFFFFFFF,$FFFF00FF,$00FF0000,$00F00F00,$0FF0FF0F,
$FFFFFFFF,$FFFF0870,$F00080FF,$FF08000F,$0780FFFF,$FFFFFFFF,$FFFFFFFF,
$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,$0F00FFFF,$000000FF,$00000000,
$FFFF00FF,$FFFFFFFF,$FFFFFFFF,$0F000080,$080000F0,$FFFFFFFF,$FFFFFFFF,
$F7FFFFFF,$FFFF0F00,$00F0FFFF,$FFFFFF7F,$FFFFFFFF,$FFFF00FF,$00FF0F00,
$00FF0F00,$0FF0FF0F,$FFFFFFFF,$000000F7,$F0000000,$0000000F,$7F000000,
$FFFFFFFF,$FFFFFFFF,$000000F7,$7F000000,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,
$0000F0FF,$0F000000,$FFFF00F0,$FFFFFFFF,$FFFFFFFF,$000000F8,$8F000000,
$FFFFFFFF,$FFFFFFFF,$70FFFFFF,$00000000,$00000000,$FFFFFF07,$FFFFFFFF,
$FFFF0FF0,$F0FF0F00,$00FFFF00,$00FFFF0F,$FFFFFFFF,$0000F7FF,$70007F00,
$00F70007,$FF7F0000,$FFFFFFFF,$FFFFFFFF,$000070FF,$FF070000,$FFFFFFFF,
$FFFFFFFF,$00F0FFFF,$00000000,$0F000000,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,
$000080FF,$FF080000,$FFFFFFFF,$FFFFFFFF,$00F7FFFF,$00000000,$00000000,
$FFFF7F00,$FFFFFFFF,$FFFF0F00,$F0FF0F00,$00FFFF00,$00FFFF0F,$FFFFFF0F,
$FFFFFFFF,$0700FFFF,$FFFF0070,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0070FFFF,
$FFFF0700,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$00000000,$F00F0000,$FFFFFF00,
$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$0FF0FFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$F0FFFFFF,$FFFF0000,$F0FFFF00,$F0FFFF00,
$00F0FF0F,$FFFFFF00,$FFFFFFFF,$0070FFFF,$FFFF0700,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$0000000F,
$00FF0000,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,
$FFFFFFFF,$00F0FFFF,$00000000,$00000000,$FFFF0F00,$F0FFFFFF,$FFFF0000,
$F0FFFF00,$F0FFFF00,$00F0FF0F,$FFFFFF00,$FFFFFFFF,$00F7FFFF,$FFFF7F00,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000FFFF,$FFFF0000,$FFFFFFFF,$FFFFFFFF,
$F0FFFFFF,$0000FF00,$0F00FF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,
$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00000000,$00000000,$FFFF0F00,
$F0FFFFFF,$F0FF0000,$F0FFFF00,$F0FFFF00,$00F0FF00,$FFFFFF00,$FFFFFFFF,
$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000FFFF,$FFFF0000,
$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$00000000,$FF0F0000,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$00F8FFFF,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00000000,
$00000000,$FFFF0F00,$FFFFFFFF,$00FF0F00,$F0FF0F00,$00FFFF00,$00FF0F00,
$FFFFFF0F,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$00000000,$FFFFFF0F,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,
$00F0FFFF,$00000000,$00000000,$FFFF0F00,$FFFFFFFF,$00FFFFFF,$00FF0F00,
$00FF0F00,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$0070FFFF,$FFFF0700,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,
$00F00F00,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$FFFF0F00,
$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00F0FF00,$00FF0F00,$FFFF0F00,$FFFFFFFF,
$00FFFFFF,$00FF0F00,$00FF0F00,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,
$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$80F7FFFF,$FFFF7F08,$FFFFFFFF,
$FFFFFFFF,$00FFFFFF,$00F0FF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$00F8FFFF,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00F0FF00,$00FF0F00,
$FFFF0F00,$FFFFFFFF,$F0FFFFFF,$00FFFF00,$F0FF0F00,$FFFFFF00,$FFFFFFFF,
$FFFFFFFF,$00F0FFFF,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$0FFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF00,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$888888F8,$88888888,$88888888,
$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$888888FF,$8F888888,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$8888F8FF,$FF8F8888,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00000080,
$00000000,$00000000,$FFFFFF08,$888808F8,$FF888888,$888888FF,$8F808888,
$FFFFFFFF,$8888F8FF,$88888888,$88888888,$FF8F8888,$FFFFFFFF,$88FFFFFF,
$00000088,$08000000,$FFFF8F88,$FFFFFFFF,$F8FFFFFF,$00008088,$88080000,
$FFFFFF8F,$FFFFFFFF,$8888FFFF,$88888888,$88888888,$FFFF8888,$FFFFFFFF,
$FFFFFFFF,$00000080,$00000000,$00000000,$FFFFFF08,$00000080,$88000000,
$00000088,$08000000,$FFFFFFFF,$000080FF,$00000000,$00000000,$FF080000,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$80F8FFFF,
$00000000,$00000000,$FFFF8F08,$FFFFFFFF,$0000F8FF,$00000000,$00000000,
$FF8F0000,$FFFFFFFF,$FFFFFFFF,$00000080,$00000000,$00000000,$FFFFFF08,
$00000080,$00000000,$00000000,$08000000,$FFFFFFFF,$000080FF,$00000000,
$00000000,$FF080000,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FF8F0000,
$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0000F8FF,
$00000000,$00000000,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$000000F8,$00000000,
$00000000,$F8FFFF08,$00000000,$00000000,$00000000,$00000000,$FFFFFF8F,
$000080FF,$00000000,$00000000,$FF080000,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,
$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,$FFFFFFFF,
$000000F8,$00000000,$00000000,$FFFFFF08,$00000080,$00000000,$00000000,
$08000000,$FFFFFFFF,$000080FF,$00000000,$00000000,$FF080000,$FFFFFFFF,
$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FFFF0800,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,
$FFFFFFFF,$FFFFFFFF,$000080FF,$00000000,$00000000,$FFFFFF08,$888888F8,
$00008888,$88880000,$8F888888,$FFFFFFFF,$000080FF,$00000000,$00000000,
$FF080000,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,
$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$0000F8FF,$00000000,
$00000000,$FF8F0000,$FFFFFFFF,$8FF8FFFF,$000080FF,$00000000,$00000000,
$FFFFFF08,$FFFFFFFF,$008088FF,$FF880800,$FFFFFFFF,$FFFFFFFF,$0088F8FF,
$00000000,$00000000,$FF8F8800,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,
$FFFF0800,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,
$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,$088088FF,$0000F8FF,
$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$000000F8,$88000000,$FFFFFFFF,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$00F8FFFF,
$00000000,$00000000,$FFFF0800,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,
$FFFF8F00,$FFFFFFFF,$0080F8FF,$00000000,$00000000,$FF8F0800,$FFFFFFFF,
$000000F8,$0080FF8F,$00000000,$00000000,$FFFFFF08,$F8FFFFFF,$00000080,
$00000000,$FFFFFF8F,$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,
$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FFFF0800,$FFFFFFFF,$00000080,$00F8FF08,$00000000,$00000000,$FFFFFF08,
$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,$00000000,
$00000000,$FFFF8F00,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FF8F0000,
$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$FFFFFFFF,$0088FFFF,
$00000000,$00000000,$FFFF8800,$F8FFFFFF,$00000000,$80FFFF08,$00000000,
$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,
$80FFFFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$0080FFFF,$00000000,
$00000000,$FF8F0000,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF8F0000,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFF8F00,$F8FFFFFF,$00000000,
$80FF8800,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$000000F8,$8F000000,
$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,
$0000F8FF,$00000000,$00000000,$FF080000,$FFFFFFFF,$000080FF,$00000000,
$00000000,$FF080000,$FFFFFFFF,$80F8FFFF,$00000000,$00000000,$FFFF8F08,
$F8FFFFFF,$00000000,$F8880000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,
$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,
$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF080000,$FFFFFFFF,
$000000F8,$00000000,$00000000,$8F000000,$FFFFFFFF,$88FFFFFF,$00000000,
$00000000,$FFFFFF88,$F8FFFFFF,$00000000,$88000000,$00000000,$00000000,
$FFFFFF8F,$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$000080FF,$00000000,$00000000,
$8F000000,$FFFFFFFF,$00000080,$00000000,$00000000,$08000000,$FFFFFFFF,
$F8FFFFFF,$00000080,$08000000,$FFFFFF8F,$FFFFFFFF,$00000080,$00000000,
$00000080,$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$000080FF,
$00000000,$00000000,$8F000000,$FFFFFFFF,$00000080,$00000000,$00000000,
$08000000,$FFFFFFFF,$FFFFFFFF,$00008088,$88080000,$FFFFFFFF,$FFFFFFFF,
$00000080,$00000000,$00000000,$08000000,$FFFFFFFF,$F8FFFFFF,$00000000,
$00000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,
$FFFFFFFF,$080080FF,$00000000,$00000000,$8F000008,$F8FFFFFF,$00000000,
$00000000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$000000F8,$8F000000,
$FFFFFFFF,$FFFFFFFF,$000000F8,$00000000,$00000000,$08000000,$FFFFFFFF,
$F8FFFFFF,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$00000080,
$08000000,$FFFFFFFF,$FFFFFFFF,$080080FF,$00000000,$00000000,$8F000008,
$F8FFFFFF,$00000000,$00000000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,
$00000088,$88000000,$FFFFFFFF,$FFFFFFFF,$000000F8,$00000000,$00000000,
$08000000,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,
$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$880080FF,$00080000,
$00000800,$8F008008,$F8FFFFFF,$00000000,$00000000,$00000000,$00000000,
$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$000080FF,
$00000000,$00000000,$8F000000,$FFFFFFFF,$80FFFFFF,$00000000,$00000000,
$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,
$880000F8,$00080000,$00000800,$08008008,$F8FFFFFF,$00000000,$00000000,
$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,
$FFFFFFFF,$000080FF,$00000000,$00000000,$8F000000,$FFFFFFFF,$80FFFFFF,
$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,
$FFFFFFFF,$FFFFFFFF,$F80800F8,$00880000,$00800800,$0800F808,$F8FFFFFF,
$00000000,$00000000,$00000000,$00000000,$FFFFFF8F,$FFFFFFFF,$00000080,
$08000000,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,$00000000,$FF080000,
$FFFFFFFF,$80FFFFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,
$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$F80800F8,$00880000,$00800800,
$0800F808,$FFFFFFFF,$00000080,$00000000,$00000000,$08000000,$FFFFFFFF,
$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$00000000,
$00000000,$FF080000,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,$FFFFFF8F,
$FFFFFFFF,$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$F88F00F8,
$00880000,$00800800,$0880FF08,$FFFFFFFF,$00000080,$00000000,$00000000,
$08000000,$FFFFFFFF,$FFFFFFFF,$00000088,$88000000,$FFFFFFFF,$FFFFFFFF,
$0000F8FF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,$F8FFFFFF,$00000000,
$00000000,$FFFFFF8F,$FFFFFFFF,$F8FFFFFF,$00000000,$00000000,$FFFFFF8F,
$FFFFFFFF,$F88F00F8,$00F80800,$00F80800,$0880FF08,$FFFFFFFF,$000000F8,
$00000000,$00000000,$8F000000,$FFFFFFFF,$FFFFFFFF,$000000F8,$8F000000,
$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FF8F0000,$FFFFFFFF,
$FFFFFFFF,$00000080,$08000000,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00000000,
$00000000,$FFFFFF08,$FFFFFFFF,$F8FF0880,$80FF0800,$00F88F00,$00F8FF08,
$FFFFFF8F,$000088FF,$00008800,$00880000,$FF880000,$FFFFFFFF,$FFFFFFFF,
$000080F8,$8F080000,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FFFF0800,$FFFFFFFF,$FFFFFFFF,$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,
$00F8FFFF,$00000000,$00000000,$FFFF8F00,$F8FFFFFF,$F8FF0800,$80FF0800,
$00F88F00,$00F8FF08,$FFFFFF08,$8888FFFF,$0000F888,$888F0000,$FFFF8888,
$FFFFFFFF,$FFFFFFFF,$008088FF,$FF880800,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,
$00000000,$00000000,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$000080FF,$FF080000,
$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$80FFFFFF,
$F88F0000,$80FF8F00,$80FF8F00,$0080FF08,$FFFF8F00,$FFFFFFFF,$0080FFFF,
$FFFF0800,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,
$FFFFFFFF,$00F8FFFF,$00000000,$00000000,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,
$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,
$FFFF0800,$80FFFFFF,$F88F0000,$80FF8F00,$80FF8F00,$0080FF08,$FFFF8F00,
$FFFFFFFF,$00F8FFFF,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,
$FF8F0000,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00000000,$08000000,$FFFFFF8F,
$FFFFFFFF,$FFFFFFFF,$0088FFFF,$FFFF8800,$FFFFFFFF,$FFFFFFFF,$0080FFFF,
$00000000,$00000000,$FFFF0800,$80FFFFFF,$808F0000,$80FF8F00,$80FF8F00,
$00808F00,$FFFF8F00,$FFFFFFFF,$80FFFFFF,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00000000,
$8F080000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,
$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,$F8FFFFFF,$00F80800,
$80FF0800,$00F88F00,$00F80800,$FFFFFF08,$FFFFFFFF,$80FFFFFF,$FFFFFF08,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FF8F0000,$FFFFFFFF,$FFFFFFFF,
$80FFFFFF,$00000000,$FF8F8808,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,
$FFFF0800,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00000000,$00000000,$FFFF0800,
$FFFFFFFF,$00F88F88,$00F80800,$00F80800,$88FF0800,$FFFFFF8F,$FFFFFFFF,
$80F8FFFF,$FFFF8808,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080F8FF,$FF8F0800,
$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$00800800,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$00808800,
$00880800,$FFFF0800,$FFFFFFFF,$00F8FFFF,$00F80800,$00F80800,$FFFF0800,
$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0088FFFF,$FFFF8800,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,$00808F08,$FFFFFF8F,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,
$0080FFFF,$00808F00,$00F80800,$FFFF0800,$FFFFFFFF,$80FFFFFF,$00F88F00,
$80FF0800,$FFFF8F00,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$88F8FFFF,$FFFF8F88,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,
$08F8FF8F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$00F8FFFF,$FFFF8F00,
$FFFFFFFF,$FFFFFFFF,$88F8FFFF,$88F8FF88,$88F88F88,$FFFF8F88,$FFFFFFFF,
$F8FFFFFF,$80FFFF88,$F8FF8F00,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,$80F8FFFF,
$FFFF8808,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$88F8FFFF,$8FFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$88FFFFFF,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F8FFFFFF,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$80FFFFFF,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$88FFFFFF,$FFFFFF88,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$000000F0,$00000000,$00000000,$FFFFFF0F,$FFFF0FFF,
$FFFFFFFF,$FFFFFFFF,$FFF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000FF,$0F000000,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$00008077,$77080000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$00000000,$00000000,
$FFFFFF0F,$000000F0,$FF070000,$000070FF,$0F000000,$FFFFFFFF,$0000F0FF,
$00000000,$00000000,$FF0F0000,$FFFFFFFF,$00FFFFFF,$00000000,$00000000,
$FFFF0F00,$FFFFFFFF,$80F7FFFF,$00000000,$00000000,$FFFF7F08,$FFFFFFFF,
$0000FFFF,$00000000,$00000000,$FFFF0000,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,
$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$77778770,$87707777,$77770778,$07787777,
$FFFFFFFF,$0000F0FF,$00000000,$00000000,$FF0F0000,$FFFFFFFF,$00F0FFFF,
$FFFFFF00,$F0FFFFFF,$FFFF0000,$FFFFFFFF,$0070FFFF,$FFFF8F00,$00F8FFFF,
$FFFF0700,$FFFFFFFF,$0000FFFF,$00000000,$00000000,$FFFF0000,$FFFFFFFF,
$FFFFFFFF,$FFFF00FF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$77777780,$08F77777,
$77777F80,$08777777,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,$FFFFFFFF,$FF0FF0FF,
$FFFFFFFF,$00F0FFFF,$FFFFFF00,$F0FFFFFF,$FFFF0000,$FFFFFFFF,$0000FFFF,
$FFFF7F88,$88F7FFFF,$FF7F0000,$FFFFFFFF,$FF00FFFF,$FFFFFFFF,$FFFFFFFF,
$FFFF00FF,$FFFFFFFF,$FFFFFFFF,$FFFF00FF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,
$000000F0,$700F0000,$0000F007,$0F000000,$FFFFFFFF,$0000F0FF,$00000000,
$00000000,$FF0F0000,$FFFFFFFF,$00FFFFFF,$00000000,$00000000,$FFFF0F00,
$FFFFFFFF,$0070FFFF,$00000000,$00000000,$FFFF0700,$FFFFFFFF,$7F00FFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF00F7,$FFFFFFFF,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,
$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$F000F7FF,$FF7F000F,$FFFFFFFF,$FFFFFFFF,
$0000F0FF,$00000000,$00000000,$FF0F0000,$FFFFFFFF,$00FFFFFF,$000000FF,
$0F000000,$FFFF0FF0,$FFFFFFFF,$00FFFFFF,$000080FF,$FF080000,$FFFFFF00,
$FFFFFFFF,$8F00FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF00F8,$FFFFFFFF,$FFFFFFFF,
$FF00F0FF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$0078FFFF,$FFFF8700,
$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFF00,$FFFFFFFF,
$00FFFFFF,$FFFFFF00,$F0FFFFFF,$FFFF0F00,$FFFFFFFF,$00FFFFFF,$FFFF7F08,
$80F7FFFF,$FFFFFF00,$FFFFFFFF,$0F80FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF08F0,
$FFFFFFFF,$0FF0FFFF,$0F00FFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,
$000000F7,$7F000000,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFF00,$FFFFFFFF,$00FFFFFF,$000000FF,$0F000000,$FFFF0FF0,$FFFFFFFF,
$00F8FFFF,$00000078,$87000000,$FFFF8F00,$FFFFFFFF,$0770FFFF,$FFFFFFFF,
$FFFFFFFF,$FFFF0770,$FFFFFFFF,$000000FF,$00F0FFFF,$FFFFFFFF,$F0FFFFFF,
$FFFFFF0F,$FFFFFFFF,$FFFF0700,$0070FFFF,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,
$00000000,$00000000,$FFFFFF00,$FFFFFFFF,$00FFFFFF,$FFFFFF00,$F0FFFFFF,
$FFFF0F00,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$FFFFFFFF,
$08F8FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF8F80,$FFFFFFFF,$000000F0,$00FFFF0F,
$FFFFFF0F,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$00000070,$07000000,$FFFFFFFF,
$FFFFFFFF,$00FFFFFF,$00000000,$00000000,$FFFFFF00,$FFFFFFFF,$00F0FFFF,
$00000000,$00000000,$FFFF0000,$FFFFFFFF,$0080FFFF,$FFFF7708,$8077FFFF,
$FFFF0800,$FFFFFFFF,$00F7FFFF,$FFFFFF7F,$F7FFFFFF,$FFFF7F00,$FFFFFFFF,
$F0000F00,$F0FFFF0F,$FFFFFF0F,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,
$0F000000,$FFFFFFFF,$FFFFFFFF,$70FFFFFF,$FFFFFF00,$00FFFFFF,$FFFFFF07,
$FFFFFFFF,$0FF0FFFF,$000000F0,$0F000000,$FFFF00FF,$FFFFFFFF,$0000F7FF,
$00000000,$00000000,$FF7F0000,$FFFFFFFF,$80FFFFFF,$FFFFFF07,$70FFFFFF,
$FFFFFF08,$FFFFFFFF,$FFFFFF00,$F0FFFF00,$FFFFFF00,$00FFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFF07F7,$7F70FFFF,$FFFFFFFF,$FFFFFFFF,$F7FFFFFF,$00000000,
$00000000,$FFFFFF7F,$FFFFFFFF,$0F00FFFF,$FFFFFFFF,$FFFFFFFF,$FF0F00FF,
$FFFFFFFF,$770070FF,$00000078,$87000000,$FF070077,$FFFFFFFF,$70FFFFFF,
$FFFF7F00,$00F7FFFF,$FFFFFF07,$FFFFFFFF,$FF0FF000,$FFFF00F0,$FFFFFF00,
$00FFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFF00FF,$FF00FFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$00000070,$07000000,$FFFFFFFF,$FFFFFFFF,$0000FFFF,$FFFFFF00,
$F0FFFFFF,$FF0F0000,$FFFFFFFF,$FF0700F7,$F0FFFFFF,$FFFFFF0F,$7F0070FF,
$FFFFFFFF,$F7FFFFFF,$FF7F0800,$000078FF,$FFFFFF7F,$FFFFFFFF,$FF0FF000,
$FF0000FF,$FFFFFF00,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F7,$7F000000,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,
$F000F0FF,$00000000,$000000F0,$FF00F000,$FFFFFFFF,$FF7F00F0,$F0FFFFFF,
$FFFFFF0F,$0F00F7FF,$FFFFFFFF,$FFFFFFFF,$7F000078,$870000F7,$FFFFFFFF,
$FFFFFFFF,$FFFF0FF0,$0000FFFF,$FFFF0FF0,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,
$000000F8,$8F000000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,
$FFFFFFFF,$FFFFFFFF,$000FF0FF,$0000000F,$0F000000,$FF000F00,$FFFFFFFF,
$FFFF0780,$70FFFFFF,$FFFFFF07,$0870FFFF,$FFFFFFFF,$FFFFFFFF,$070070FF,
$FF070070,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$00FFFFFF,$FFFF0F00,$0FF0FFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF8F80,$08F8FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$000FF0FF,$0F00FF00,$F00F000F,
$FF000F00,$FFFFFFFF,$FFFF8F00,$80FFFFFF,$FFFFFF08,$00F8FFFF,$FFFFFFFF,
$FFFFFFFF,$7F0000F7,$7F0000F7,$FFFFFFFF,$FFFFFFFF,$FFFF00FF,$FFFFFFFF,
$FFFF0F00,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFF08,$80FFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$0F00F0FF,
$0F00FF00,$F00F00FF,$FF00000F,$FFFFFFFF,$FFFF7F00,$00FFFFFF,$FFFFFF00,
$00F7FFFF,$FFFFFFFF,$FFFFFFFF,$FF7F00F8,$8F00F7FF,$FFFFFFFF,$FFFFFFFF,
$FFFF00FF,$FFFFFFFF,$FFFF0FF0,$0F00FFFF,$FFFFFFFF,$F8FFFFFF,$F0FFFF07,
$70FFFF0F,$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,
$FFFFFFFF,$0F00F0FF,$000FF000,$F0000FF0,$FF00000F,$FFFFFFFF,$FFFFFF00,
$00F8FFFF,$FFFF8F00,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFF08F0,$0F80FFFF,
$FFFFFFFF,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,$FFFF0FF0,$FF00FFFF,$FFFFFFFF,
$F0FFFFFF,$F0FFFF0F,$F0FFFF0F,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,
$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FF0000FF,$000FF000,$F0000FF0,$0F00F00F,
$FFFFFFFF,$FFFFFF00,$07F0FFFF,$FFFF0F70,$00FFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FF0FF0FF,$FFFFFFFF,$FFFF00FF,
$FF00F0FF,$FFFFFFFF,$F0FFFFFF,$00F0FF0F,$F0FF0F00,$FFFFFF0F,$FFFFFFFF,
$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FF0000FF,$000FF000,
$F0000FF0,$0F00F00F,$FFFFFFFF,$FFFF7F00,$8F70FFFF,$FFFF07F8,$00F7FFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$F000FFFF,
$FFFFFFFF,$FFFF00FF,$FF0FF0FF,$FFFFFFFF,$F8FFFFFF,$00F0FF08,$80FF0F00,
$FFFFFF8F,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,
$FF0F00FF,$00FF0000,$00F00FF0,$0F00FF0F,$FFFFFFFF,$FFFF0F80,$7F00F7FF,
$FF7F00F7,$08F0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF08F0,$0F80FFFF,$FFFFFFFF,
$FFFFFFFF,$F000FFFF,$FFFFFF0F,$FFFFFFFF,$FF0F00FF,$FFFFFFFF,$F7FFFFFF,
$F0FFFF00,$00FFFF0F,$FFFFFF7F,$FFFFFFFF,$FFFFFFFF,$FFFF0F70,$07F0FFFF,
$FFFFFFFF,$FFFFFFFF,$FF0F00FF,$00FF0000,$00F00FF0,$0F00FF0F,$FFFFFFFF,
$FF7F00F0,$FF0080F7,$870000FF,$0F00F7FF,$FFFFFFFF,$FFFFFFFF,$FF8F00F8,
$8F00F8FF,$FFFFFFFF,$FFFFFFFF,$F000FFFF,$FFFFFF00,$FFFFFFFF,$FFFF00FF,
$FFFFFFFF,$FFFFFFFF,$F0FF8F80,$08F8FF0F,$FFFFFFFF,$FFFFFFFF,$F7FFFFFF,
$00000000,$00000000,$FFFFFF7F,$FFFFFFFF,$FFFF00FF,$00FF0000,$00F00FF0,
$0FF0FF0F,$FFFFFFFF,$000000F7,$FF000000,$000000FF,$7F000000,$FFFFFFFF,
$FFFFFFFF,$7F0000F7,$FF0000FF,$FFFFFFFF,$FFFFFFFF,$0FF0FFFF,$FFFF0F00,
$FFFFFFFF,$FFFF00F0,$FFFFFFFF,$FFFFFFFF,$FFFF08F8,$8F80FFFF,$FFFFFFFF,
$FFFFFFFF,$70FFFFFF,$00000000,$00000000,$FFFFFF07,$FFFFFFFF,$FFFF00F0,
$F0FF0F00,$00FFFF00,$00F0FF0F,$FFFFFFFF,$0000F8FF,$7F00F700,$007F00F7,
$FF8F0000,$FFFFFFFF,$FFFFFFFF,$0F0070FF,$FF0F00F0,$FFFFFFFF,$FFFFFFFF,
$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$FF8F80FF,
$FF08F8FF,$FFFFFFFF,$FFFFFFFF,$00F7FFFF,$FFFFFF0F,$F0FFFFFF,$FFFF7F00,
$FFFFFFFF,$FFFF0F00,$F0FF0F00,$00FFFF00,$00FFFF0F,$FFFFFF0F,$FFFFFFFF,
$0700FFFF,$FFFF0070,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0F70FFFF,$FFFF07F0,
$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$FFFF0FF0,$F0FFFFFF,$FFFFFF00,$FFFFFFFF,
$FFFFFFFF,$7F00F8FF,$FF8F00F7,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00000000,
$00000000,$FFFF0F00,$F0FFFFFF,$FFFF0000,$F0FF0F00,$00FFFF00,$00F0FF0F,
$FFFFFF00,$FFFFFFFF,$0070FFFF,$FFFF0700,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$7F80FFFF,$FFFF08F7,$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$FFFF0F00,$00F0FF00,
$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$0080FFFF,$FFFF0800,$FFFFFFFF,$FFFFFFFF,
$00F0FFFF,$00000000,$00000000,$FFFF0F00,$F0FFFFFF,$FFFF000F,$F0FFFF00,
$F0FFFF00,$0FF0FF0F,$FFFFFF00,$FFFFFFFF,$00F7FFFF,$FFFF7F00,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FF00FFFF,$FFFF00FF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,
$FF0F000F,$0F000000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$80FFFFFF,$FFFFFF08,
$FFFFFFFF,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF0FF0,$F0FFFFFF,
$F0FF0000,$F0FFFF00,$F0FFFF00,$00F0FF00,$FFFFFF00,$FFFFFFFF,$70FFFFFF,
$FFFFFF07,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$7F00FFFF,$FFFF00F7,$FFFFFFFF,
$FFFFFFFF,$F0FFFFFF,$FF00F00F,$FF0F0000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$08F8FFFF,$FFFF8F80,$FFFFFFFF,$FFFFFFFF,$0FF0FFFF,$0F000000,$000000F0,
$FFFF0FF0,$FFFFFFFF,$00FF0F00,$F0FF0F00,$00FFFF00,$00FF0F00,$FFFFFF0F,
$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0780FFFF,
$FFFF0870,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$F0000000,$FFFFFF0F,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$8FF0FFFF,$FFFF0FF8,$FFFFFFFF,$FFFFFFFF,$0FF0FFFF,
$0F000000,$000000F0,$FFFF0FF0,$FFFFFFFF,$00FFFFFF,$00FF0FF0,$00FF0F00,
$FFFF0FF0,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$0070FFFF,$FFFF0700,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$00F00F00,
$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$8FF0FFFF,$FFFF0FF8,$FFFFFFFF,
$FFFFFFFF,$00F0FFFF,$00F0FF00,$00FF0F00,$FFFF0F00,$FFFFFFFF,$00FFFFFF,
$00FF0F00,$00FF0FF0,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$FFFF0F00,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$80F7FFFF,$FFFF7F08,$FFFFFFFF,$FFFFFFFF,
$00FFFFFF,$00F0FF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$08F8FFFF,
$FFFF8F80,$FFFFFFFF,$FFFFFFFF,$00F0FFFF,$00F0FF00,$00FF0F00,$FFFF0F00,
$FFFFFFFF,$F0FFFFFF,$00FFFF00,$F0FF0F00,$FFFFFF00,$FFFFFFFF,$FFFFFFFF,
$00F0FFFF,$FFFF0F00,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$00FFFFFF,$0FFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$80FFFFFF,$FFFFFF08,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF00,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$F0FFFFFF,$FFFFFF0F,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000FFFF);

set64: Array [0..3087] of Integer=(
$303E4D42,$00000000,$003E0000,$00280000,$01800000,$01000000,$00010000,
$00000001,$30000000,$00000000,$00000000,$00020000,$00020000,$00000000,
$FFFF0000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$3FFFFFFF,$FFFFFFFF,$FFFFFFFC,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FF03C0FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,
$03FF3F00,$07FC3FE0,$FFFFFFC0,$FFFFFFFF,$FFFFFFFF,$FF0080FF,$FFFFFFFF,
$1F0000F8,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$00FE3F00,$00E00700,
$E0FF7F00,$00000000,$FFFFFF03,$030000E0,$FFFFFFFF,$030000C0,$FEFFFFFF,
$00000000,$FFFFFF3F,$000000F0,$00FE3F00,$00C00300,$E0FF7F00,$00000000,
$FEFFFF03,$00000000,$FEFFFF3F,$00000000,$FCFFFF7F,$00000000,$FFFFFF1F,
$000000F0,$00FC3F00,$00800100,$E0FF3F00,$00000000,$F8FFFF03,$00000000,
$F8FFFF0F,$00000000,$FCFFFF1F,$00000000,$FFFFFF1F,$000000F0,$00FC3F00,
$00000000,$E0FF3F00,$00000000,$FCFFFF03,$00000000,$FCFFFF1F,$00000000,
$FCFFFF3F,$00000000,$FFFFFF1F,$000000F8,$00FE3F00,$00000000,$E0FF7F00,
$00000000,$FEFFFF03,$00000000,$FCFFFF3F,$00000000,$FCFFFF3F,$00000000,
$FFFFFF1F,$000000F8,$80FF3F00,$00000000,$E0FFFF01,$00000000,$FEFFFF03,
$00000000,$FEFFFF3F,$00000000,$FCFFFF7F,$00000000,$FFFFFF1F,$000000FC,
$F8FF3F00,$FC01803F,$E0FFFF1F,$00000000,$FFFFFF03,$00000000,$FFFFFF7F,
$00000000,$FCFFFFFF,$00000000,$FFFFFF1F,$000000FC,$FFFF3F00,$FF07E0FF,
$FFFFFFFF,$00000000,$FFFFFF7F,$00000000,$FFFFFF7F,$00000000,$FCFFFFFF,
$00000000,$FFFFFF1F,$000000FE,$FFFF3F00,$FF07E0FF,$FFFFFFFF,$00000000,
$FFFFFF7F,$00000080,$FFFFFFFF,$00000000,$FCFFFFFF,$00000000,$FFFFFF1F,
$000000FF,$FFFF3F00,$FF0000FF,$FFFFFFFF,$00000000,$FFFFFF7F,$00000000,
$FFFFFF7F,$00000000,$FEFFFFFF,$00000000,$FCFFFF3F,$0000807F,$FFFF3F00,
$1F0000F8,$FFFFFFFF,$00000000,$FFFFFF7F,$00000000,$FEFFFF7F,$00000000,
$FEFFFF7F,$00000000,$C0FFFF3F,$0000803F,$FFFF3F00,$0F0000F0,$FFFFFFFF,
$00000000,$FFFFFF7F,$00000000,$FEFFFF7F,$00000000,$FEFFFF7F,$00000000,
$00FFFF3F,$0000C01F,$FFFF3F00,$0F0000F0,$FFFFFFFF,$00000000,$FFFFFF7F,
$00000000,$FEFFFF7F,$00000000,$FFFFFF7F,$00000000,$00FEFF7F,$0000E01F,
$FFFF3F00,$1F0000F8,$FFFFFFFF,$00000080,$FEFFFFFF,$00000000,$FCFFFF3F,
$00000000,$FFFFFF3F,$00000000,$00FEFF7F,$0000F00F,$FFFF7F00,$1F0000F8,
$FFFFFFFF,$010000C0,$FEFFFFFF,$00000000,$F8FFFF3F,$00000000,$FFFFFF1F,
$00000080,$00FCFFFF,$0000F807,$FFFF7F00,$1F0000F8,$FFFFFFFF,$030000E0,
$FCFFFFFF,$00000000,$F0FFFF1F,$00000000,$FFFFFF0F,$00000080,$00FCFFFF,
$0000FC03,$FFFF7F00,$3F0000FC,$FFFFFFFF,$070000F0,$F8FFFFFF,$00000000,
$E0FFFF0F,$00000000,$FFFFFF07,$010000C0,$00FCFFFF,$0000FE01,$FFFF7F00,
$3F0000FC,$FFFFFFFF,$0F0000F8,$F8FFFFFF,$00000000,$C0FFFF0F,$00000000,
$FFFFFF03,$030000E0,$00FCFFFF,$00007E00,$FFFF7F00,$7F0000FE,$FFFFFFFF,
$0F0000F8,$F0FFFFFF,$00000000,$80FFFF07,$00000000,$FFFFFF01,$070000F0,
$00FCFFFF,$00001F00,$FFFF7F00,$3F0000FC,$FFFFFFFF,$0F0000F8,$F0FFFFFF,
$00000000,$00FFFF07,$00000000,$FFFFFF00,$0F0000F8,$00FCFFFF,$00000700,
$FFFFFF00,$1F0000F8,$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$00FFFF07,
$00000000,$FFFFFF00,$3F0000FE,$00FEFFFF,$00800100,$FFFFFF00,$0F0000F0,
$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$00FEFF07,$00000000,$FFFF7F00,
$FF0080FF,$00FEFFFF,$00800000,$FFFFFF00,$0F0000F0,$FFFFFFFF,$0F0000F8,
$F0FFFFFF,$00000000,$00FEFF07,$00000000,$FFFF7F00,$7F0000FF,$00FFFFFF,
$00000000,$FFFFFF00,$070000E0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,$02000020,
$00FEFF03,$00000000,$FFFF7F00,$3F0000FE,$00FFFFFF,$00000000,$FFFFFF01,
$070000E0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,$02000020,$00FEFF03,$00000000,
$FFFF7F00,$1F0000FC,$80FFFFFF,$00000000,$FFFFFF01,$030000C0,$FFFFFFFF,
$0F0000F8,$E0FFFFFF,$03041060,$00FEFF03,$00000000,$FFFF7F00,$1F0000FC,
$C0FFFFFF,$00000000,$FFFFFF01,$030000C0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,
$03041060,$00FEFF03,$00000000,$FFFF7F00,$0F0000F8,$C0FFFFFF,$00000000,
$FFFFFF03,$030000C0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,$030410E0,$00FEFF83,
$00000000,$FFFF7F00,$0F0000F8,$E0FFFFFF,$00000000,$FFFFFF03,$030000C0,
$FFFFFFFF,$0F0000F8,$C0FFFFFF,$030630E0,$00FFFF81,$00000000,$FFFFFF00,
$0F0000F8,$E0FFFFFF,$00000000,$FFFFFF07,$030000C0,$FFFFFFFF,$0F0000F8,
$C1FFFFFF,$030E38E0,$00FFFFC1,$00000000,$FFFFFF00,$0F0000F8,$F0FFFFFF,
$00000000,$FFFFFF07,$030000C0,$FFFFFFFF,$0F0000F8,$C3FFFFFF,$030E38E0,
$80FFFFC1,$00000000,$FFFFFF01,$0F0000F8,$F0FFFFFF,$00000000,$FFFFFF0F,
$030000C0,$FFFFFFFF,$0F0000F8,$C3FFFFFF,$030F78E0,$C0FFFFE1,$00000000,
$FFFFFF03,$1F0000FC,$F8FFFFFF,$00000000,$FFFFFF0F,$030000C0,$FFFFFFFF,
$0F0000F8,$87FFFFFF,$030F78E0,$E0FFFFF0,$00024000,$FFFFFF07,$1F0000FC,
$F8FFFFFF,$00000000,$FFFFFF1F,$070000E0,$FFFFFFFF,$070000F0,$87FFFFFF,
$030F78E0,$F0FFFFF0,$0003C000,$FFFFFF0F,$3F0000FE,$F8FFFFFF,$00000000,
$FFFFFF1F,$070000E0,$FFFFFFFF,$030000E0,$8FFFFFFF,$831FFCE0,$FCFFFFF8,
$C003C003,$FFFFFF3F,$7F0000FF,$F8FFFFFF,$00000000,$FFFFFF3F,$0F0000F0,
$FFFFFFFF,$00000080,$8FFFFFFF,$831FFCE0,$FFFFFFF8,$FF03C0FF,$FFFFFFFF,
$FF0080FF,$F8FFFFFF,$00000000,$FFFFFF7F,$1F0000F8,$FFFFFFFF,$00000000,
$1FFFFF7F,$831FFCE0,$FFFF7FFC,$FF07E0FF,$FFFFFFFF,$FF01C0FF,$FCFFFFFF,
$00000000,$FFFFFFFF,$3F0000FC,$FEFFFFFF,$00000000,$3FFEFF3F,$C31FFCE1,
$FFFF3FFE,$FF07E0FF,$FFFFFFFF,$FF07F0FF,$FCFFFFFF,$01000000,$FFFFFFFF,
$7F0000FE,$FCFFFFFF,$00000000,$1FFCFF1F,$C31FFCE1,$FFFF1FFC,$FF0FF0FF,
$FFFFFFFF,$FF03E0FF,$FEFFFFFF,$03000000,$FFFFFFFF,$FF0180FF,$FCFFFFFF,
$00000000,$0FF8FF1F,$C33FFEE1,$FFFF0FF8,$FF1FF8FF,$FFFFFFFF,$FF03E0FF,
$FFFFFFFF,$07000000,$FFFFFFFF,$FF03C0FF,$FCFFFFFF,$00000000,$0FF8FF1F,
$E33FFEE3,$FFFF0FF8,$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,$0F000080,
$FFFFFFFF,$FF07E0FF,$FCFFFFFF,$00000000,$0FF8FF1F,$E33FFEE3,$FFFF0FF8,
$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,$3F000080,$FFFFFFFF,$FF1FF8FF,
$FCFFFFFF,$00000000,$1FFCFF1F,$E33FFEE3,$FFFF1FFC,$FF7FFEFF,$FFFFFFFF,
$FF01C0FF,$FFFFFFFF,$FF000080,$FFFFFFFF,$FF1FF8FF,$FCFFFFFF,$00000000,
$3FFEFF1F,$C13FFEC1,$FFFF3FFE,$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,
$FF030080,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$00000000,$FFFFFF1F,$803FFE80,
$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FF03E0FF,$FFFFFFFF,$FF7F0000,$FFFFFFFF,
$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,$801FFC80,$FFFFFFFF,$FF07E0FF,
$FFFFFFFF,$FF03E0FF,$FFFFFFFF,$FFFFC000,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,
$C007F001,$FFFFFF1F,$800FF880,$FFFFFFFF,$FF07E0FF,$FFFFFFFF,$FF07F0FF,
$FFFFFFFF,$FFFFC003,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,
$C10FF8C1,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FF1FFCFF,$FFFFFFFF,$FFFFE107,
$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,$E30FF8E3,$FFFFFFFF,
$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFE10F,$FFFFFFFF,$FF1FF8FF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF1FFCFF,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFF3BF,$FFFFFFFF,$FF3FFCFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FF3FFEFF,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$3FFFFFFF,$FFFFFFFF,$FFFFFFFC,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FF03C0FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$03FF3F00,$07FC3FE0,
$FFFFFFC0,$FFFFFFFF,$FFFFFFFF,$FF0080FF,$FFFFFFFF,$1F0000F8,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$000000F0,$00FE3F00,$00E00700,$E0FF7F00,$00000000,
$FFFFFF03,$030000E0,$FFFFFFFF,$030000C0,$FEFFFFFF,$00000000,$FFFFFF3F,
$000000F0,$00FE3F1C,$00C00300,$E0FF7F00,$00000000,$FEFFFF03,$00000000,
$FEFFFF3F,$00000000,$FCFFFF7F,$00000000,$FFFFFF1F,$000000F0,$00FC3F1C,
$00800100,$E0FF3F00,$00000000,$F8FFFF03,$00000000,$F8FFFF0F,$00000000,
$FCFFFF1F,$00000000,$FFFFFF1F,$000000F0,$00FC3F1C,$00000000,$E0FF3F00,
$00000000,$FCFFFF03,$07000070,$FCFFFF1F,$3E00007C,$FCFFFF3F,$00000000,
$FFFFFF1F,$000000F8,$00FE3F1C,$00000000,$E0FF7F00,$00000000,$FEFFFF03,
$FE01C03F,$FCFFFF3F,$FC07E03F,$FCFFFF3F,$00000000,$FFFFFF1F,$000000F8,
$80FF3F1C,$00000000,$E0FFFF01,$FEFFFF3F,$FEFFFF03,$FCFFFF1F,$FEFFFF3F,
$E0FFFF07,$FCFFFF7F,$00000000,$FFFFFF1F,$000000FC,$F8FF3F1C,$FC01803F,
$E0FFFF1F,$FEFFFF3F,$FFFFFF03,$00FF7F00,$FFFFFF7F,$00FC3F00,$FCFFFFFF,
$00000000,$FFFFFF1F,$000000FC,$FFFF3F1C,$FF07E0FF,$FFFFFFFF,$00000000,
$FFFFFF7F,$00000000,$FFFFFF7F,$1C000038,$FCFFFFFF,$00000000,$FFFFFF1F,
$000000FE,$FFFF3F1C,$FF07E0FF,$FFFFFFFF,$00000000,$FFFFFF7F,$0C000098,
$FFFFFFFF,$FC03803F,$FCFFFFFF,$00000000,$FFFFFF1F,$000000FF,$FFFF3F1C,
$FF0000FF,$FFFFFFFF,$00000000,$FFFFFF7F,$FC00801F,$FFFFFF7F,$F0FFFF0F,
$FEFFFFFF,$00000000,$FCFFFF3F,$0000807F,$FFFF3F1C,$1F0000F8,$FFFFFFFF,
$00000000,$FFFFFF7F,$F8FFFF0F,$FEFFFF7F,$00FE7F00,$FEFFFF7F,$00000000,
$C0FFFF3F,$0000803F,$FFFF3F1C,$0F0000F0,$FFFFFFFF,$FEFFFF3F,$FFFFFF7F,
$80FFFF00,$FEFFFF7F,$00000000,$FEFFFF7F,$00000000,$00FFFF3F,$0000C01F,
$FFFF3F1C,$8F0000F1,$FFFFFFFF,$FCFFFF1F,$FFFFFF7F,$00000000,$FEFFFF7F,
$00000000,$FFFFFF7F,$00000000,$0EFEFF7F,$0000E01F,$FFFF3F18,$9F0FF0F9,
$FFFFFFFF,$00000080,$FEFFFFFF,$06000030,$FCFFFF3F,$1E000078,$FFFFFF3F,
$00000000,$07FEFF7F,$0000F00F,$FFFF7F38,$1FFFFFF8,$FFFFFFFF,$010000C0,
$FEFFFFFF,$7E00003F,$F8FFFF3F,$FF03C0FF,$FFFFFF1F,$00000080,$03FCFFFF,
$0000F807,$FFFF7F38,$1FF81FF8,$FFFFFFFF,$030000E0,$FCFFFFFF,$FCFFFF1F,
$F1FFFF1F,$F37FFECF,$FFFFFF8F,$00000080,$23FCFFFF,$0000FC03,$FFFF7F38,
$3F0000FC,$FFFFFFFF,$C7FFFFF1,$F8FFFFFF,$C0FFFF01,$E3FFFF0F,$017E7E80,
$FFFFFFC7,$010000C0,$70FCFFFF,$0000FE01,$FFFF7F38,$3F0000FC,$FFFFFFFF,
$CFFFFFF9,$F8FFFFFF,$00000000,$C7FFFF0F,$00600600,$FFFFFFE3,$030000E0,
$70FCFFFF,$00007E00,$FFFF7F38,$7F0E70FE,$FFFFFFFF,$0F0000F8,$F0FFFFFF,
$00000000,$8EFFFF07,$00600600,$FFFFFF71,$070000F0,$30FCFFFF,$00001F00,
$FFFF7F30,$3FFE7FFC,$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$1CFFFF07,
$00600600,$FFFFFF38,$0F0000F8,$00FCFFFF,$00000700,$FFFFFF70,$1FF81FF8,
$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$38FFFF07,$00600600,$FFFFFF1C,
$3F0000FE,$00FEFFFF,$00800100,$FFFFFF70,$0F0000F0,$FFFFFFFF,$0F0000F8,
$F0FFFFFF,$00000000,$30FEFF07,$00300C00,$FFFF7F0C,$FF0080FF,$00FEFFFF,
$00800000,$FFFFFF70,$0F0000F0,$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,
$70FEFF07,$00300C00,$FFFF7F0E,$7F0000FF,$00FFFFFF,$00000000,$FFFFFF60,
$070000E0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,$02000020,$60FEFF03,$00300C00,
$FFFF7F06,$3F0000FE,$00FFFFFF,$00000000,$FFFFFFE1,$078001E0,$FFFFFFFF,
$0F0000F8,$E0FFFFFF,$02000020,$60FEFF03,$00181800,$FFFF7F06,$1F0000FC,
$80FFFFFF,$00E00000,$FFFFFFE1,$038001C0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,
$03041060,$60FEFF03,$00981900,$FFFF7F06,$1F0000FC,$C0FFFFFF,$00E00000,
$FFFFFFE1,$038001C0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,$03041060,$70FEFF03,
$008C3100,$FFFF7F0E,$0F0000F8,$C0FFFFFF,$00600000,$FFFFFFC3,$038001C0,
$FFFFFFFF,$0F0000F8,$E0FFFFFF,$030410E0,$30FEFF83,$00CE7300,$FFFF7F0C,
$0F0000F8,$E0FFFFFF,$01600000,$FFFFFFC3,$038001C0,$FFFFFFFF,$0F0000F8,
$C0FFFFFF,$030630E0,$38FFFF81,$00C7E300,$FFFFFF1C,$0F0000F8,$E0FFFFFF,
$01700000,$FFFFFF87,$03FC3FC0,$FFFFFFFF,$0F0000F8,$C1FFFFFF,$030E38E0,
$1CFFFFC1,$C063C603,$FFFFFF38,$0F0000F8,$F0FFFFFF,$03300000,$FFFFFF87,
$03FC3FC0,$FFFFFFFF,$0F0000F8,$C3FFFFFF,$030E38E0,$8FFFFFC1,$F061860F,
$FFFFFFF1,$0F0000F8,$F0FFFFFF,$033000C0,$FFFFFF0F,$038001C0,$FFFFFFFF,
$0F0000F8,$C3FFFFFF,$030F78E0,$C7FFFFE1,$7F6006FE,$FFFFFFE3,$1F0000FC,
$F8FFFFFF,$032000F0,$FFFFFF0F,$038001C0,$FFFFFFFF,$0F0000F8,$87FFFFFF,
$030F78E0,$E1FFFFF0,$1F324CF8,$FFFFFF87,$1F0000FC,$F8FFFFFF,$060000F8,
$FFFFFF1F,$078001E0,$FFFFFFFF,$C7FFFFF1,$87FFFFFF,$030F78E0,$F0FFFFF0,
$0033CC00,$FFFFFF0F,$3F0000FE,$F8FFFFFF,$0C000078,$FFFFFF1F,$078001E0,
$FFFFFFFF,$C3FFFFE1,$8FFFFFFF,$831FFCE0,$FCFFFFF8,$C033CC03,$FFFFFF3F,
$7F0000FF,$F8FFFFFF,$1C00007C,$FFFFFF3F,$0F8001F0,$FFFFFFFF,$00000080,
$8FFFFFFF,$831FFCE0,$FFFFFFF8,$FF63C6FF,$FFFFFFFF,$FF0080FF,$F8FFFFFF,
$1800001C,$FFFFFF7F,$1F0000F8,$FFFFFFFF,$00000000,$1FFFFF7F,$831FFCE0,
$FFFF7FFC,$FFE7E7FF,$FFFFFFFF,$FF01C0FF,$FCFFFFFF,$30000000,$FFFFFFFF,
$3F0000FC,$FEFFFFFF,$00000000,$3FFEFF3F,$C31FFCE1,$FFFF3FFE,$FF87E1FF,
$FFFFFFFF,$FF07F0FF,$FCFFFFFF,$61000000,$FFFFFFFF,$7F0000FE,$FCFFFFFF,
$FEFFFF3F,$1FFCFF1F,$C31FFCE1,$FFFF1FFC,$FF0FF0FF,$FFFFFFFF,$FF03E0FF,
$FEFFFFFF,$C3010000,$FFFFFFFF,$FF0180FF,$FCFFFFFF,$FFFFFF7F,$0FF8FF1F,
$C33FFEE1,$FFFF0FF8,$FF1FF8FF,$FFFFFFFF,$FF03E0FF,$FFFFFFFF,$87030000,
$FFFFFFFF,$FF03C0FF,$FCFFFFFF,$00000000,$0FF8FF1F,$E33FFEE3,$FFFF0FF8,
$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,$0F1E8087,$FFFFFFFF,$FF07E0FF,
$FCFFFFFF,$00000000,$0FF8FF1F,$E33FFEE3,$FFFF0FF8,$FF7FFEFF,$FFFFFFFF,
$FF01C0FF,$FFFFFFFF,$3F700080,$FFFFFFFF,$FF1FF8FF,$FCFFFFFF,$00000000,
$1FFCFF1F,$E33FFEE3,$FFFF1FFC,$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,
$FF000080,$FFFFFFFF,$FF1FF8FF,$FCFFFFFF,$00000000,$3FFEFF1F,$C13FFEC1,
$FFFF3FFE,$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,$FF030080,$FFFFFFFF,
$FF0FF0FF,$FCFFFFFF,$00000000,$FFFFFF1F,$803FFE80,$FFFFFFFF,$FF7FFEFF,
$FFFFFFFF,$FF03E0FF,$FFFFFFFF,$FF7F0000,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,
$C007F001,$FFFFFF1F,$801FFC80,$FFFFFFFF,$FF07E0FF,$FFFFFFFF,$FF03E0FF,
$FFFFFFFF,$FFFFC000,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,
$800FF880,$FFFFFFFF,$FF07E0FF,$FFFFFFFF,$FF07F0FF,$FFFFFFFF,$FFFFC003,
$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,$C10FF8C1,$FFFFFFFF,
$FF7FFEFF,$FFFFFFFF,$FF1FFCFF,$FFFFFFFF,$FFFFE107,$FFFFFFFF,$FF0FF0FF,
$FCFFFFFF,$C007F001,$FFFFFF1F,$E30FF8E3,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFE10F,$FFFFFFFF,$FF1FF8FF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FF1FFCFF,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFF3BF,$FFFFFFFF,$FF3FFCFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF3FFEFF,
$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FFFFFFF,$FFFFFFFF,
$FFFFFFFC,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF03C0FF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$000000F0,$03FF3F00,$07FC3FE0,$FFFFFFC0,$FFFFFFFF,
$FFFFFFFF,$FF0080FF,$FFFFFFFF,$1F0000F8,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$000000F0,$00FE3F00,$00E00700,$E0FF7F00,$00000000,$FFFFFF03,$030000E0,
$FFFFFFFF,$030000C0,$FEFFFFFF,$00000000,$FFFFFF3F,$000000F0,$00FE3F00,
$00C00300,$E0FF7F00,$00000000,$FEFFFF03,$00000000,$FEFFFF3F,$00000000,
$FCFFFF7F,$00000000,$FFFFFF1F,$000000F0,$00FC3F00,$00800100,$E0FF3F00,
$00000000,$F8FFFF03,$00000000,$F8FFFF0F,$00000000,$FCFFFF1F,$00000000,
$FFFFFF1F,$000000F0,$00FC3F00,$00000000,$E0FF3F00,$00000000,$FCFFFF03,
$00000000,$FCFFFF1F,$00000000,$FCFFFF3F,$00000000,$FFFFFF1F,$000000F8,
$00FE3F00,$00000000,$E0FF7F00,$00000000,$FEFFFF03,$00000000,$FCFFFF3F,
$00000000,$FCFFFF3F,$00000000,$FFFFFF1F,$000000F8,$80FF3F00,$00000000,
$E0FFFF01,$00000000,$FEFFFF03,$00000000,$FEFFFF3F,$00000000,$FCFFFF7F,
$00000000,$FFFFFF1F,$000000FC,$F8FF3F00,$FC01803F,$E0FFFF1F,$00000000,
$FFFFFF03,$00000000,$FFFFFF7F,$00000000,$FCFFFFFF,$00000000,$FFFFFF1F,
$000000FC,$FFFF3F00,$FF07E0FF,$FFFFFFFF,$00000000,$FFFFFF7F,$00000000,
$FFFFFF7F,$00000000,$FCFFFFFF,$00000000,$FFFFFF1F,$000000FE,$FFFF3F00,
$FF07E0FF,$FFFFFFFF,$00000000,$FFFFFF7F,$00000080,$FFFFFFFF,$00000000,
$FCFFFFFF,$00000000,$FFFFFF1F,$000000FF,$FFFF3F00,$FF0000FF,$FFFFFFFF,
$00000000,$FFFFFF7F,$00000000,$FFFFFF7F,$00000000,$FEFFFFFF,$00000000,
$FCFFFF3F,$0000807F,$FFFF3F00,$1F0000F8,$FFFFFFFF,$00000000,$FFFFFF7F,
$00000000,$FEFFFF7F,$00000000,$FEFFFF7F,$00000000,$C0FFFF3F,$0000803F,
$FFFF3F00,$0F0000F0,$FFFFFFFF,$00000000,$FFFFFF7F,$00000000,$FEFFFF7F,
$00000000,$FEFFFF7F,$00000000,$00FFFF3F,$0000C01F,$FFFF3F00,$0F0000F0,
$FFFFFFFF,$00000000,$FFFFFF7F,$00000000,$FEFFFF7F,$00000000,$FFFFFF7F,
$00000000,$00FEFF7F,$0000E01F,$FFFF3F00,$1F0000F8,$FFFFFFFF,$00000080,
$FEFFFFFF,$00000000,$FCFFFF3F,$00000000,$FFFFFF3F,$00000000,$00FEFF7F,
$0000F00F,$FFFF7F00,$1F0000F8,$FFFFFFFF,$010000C0,$FEFFFFFF,$00000000,
$F8FFFF3F,$00000000,$FFFFFF1F,$00000080,$00FCFFFF,$0000F807,$FFFF7F00,
$1F0000F8,$FFFFFFFF,$030000E0,$FCFFFFFF,$00000000,$F0FFFF1F,$00000000,
$FFFFFF0F,$00000080,$00FCFFFF,$0000FC03,$FFFF7F00,$3F0000FC,$FFFFFFFF,
$070000F0,$F8FFFFFF,$00000000,$E0FFFF0F,$00000000,$FFFFFF07,$010000C0,
$00FCFFFF,$0000FE01,$FFFF7F00,$3F0000FC,$FFFFFFFF,$0F0000F8,$F8FFFFFF,
$00000000,$C0FFFF0F,$00000000,$FFFFFF03,$030000E0,$00FCFFFF,$00007E00,
$FFFF7F00,$7F0000FE,$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$80FFFF07,
$00000000,$FFFFFF01,$070000F0,$00FCFFFF,$00001F00,$FFFF7F00,$3F0000FC,
$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$00FFFF07,$00000000,$FFFFFF00,
$0F0000F8,$00FCFFFF,$00000700,$FFFFFF00,$1F0000F8,$FFFFFFFF,$0F0000F8,
$F0FFFFFF,$00000000,$00FFFF07,$00000000,$FFFFFF00,$3F0000FE,$00FEFFFF,
$00800100,$FFFFFF00,$0F0000F0,$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,
$00FEFF07,$00000000,$FFFF7F00,$FF0080FF,$00FEFFFF,$00800000,$FFFFFF00,
$0F0000F0,$FFFFFFFF,$0F0000F8,$F0FFFFFF,$00000000,$00FEFF07,$00000000,
$FFFF7F00,$7F0000FF,$00FFFFFF,$00000000,$FFFFFF00,$070000E0,$FFFFFFFF,
$0F0000F8,$E0FFFFFF,$02000020,$00FEFF03,$00000000,$FFFF7F00,$3F0000FE,
$00FFFFFF,$00000000,$FFFFFF01,$070000E0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,
$02000020,$00FEFF03,$00000000,$FFFF7F00,$1F0000FC,$80FFFFFF,$00000000,
$FFFFFF01,$030000C0,$FFFFFFFF,$0F0000F8,$E0FFFFFF,$03041060,$00FEFF03,
$00000000,$FFFF7F00,$1F0000FC,$C0FFFFFF,$00000000,$FFFFFF01,$030000C0,
$FFFFFFFF,$0F0000F8,$E0FFFFFF,$03041060,$00FEFF03,$00000000,$FFFF7F00,
$0F0000F8,$C0FFFFFF,$00000000,$FFFFFF03,$030000C0,$FFFFFFFF,$0F0000F8,
$E0FFFFFF,$030410E0,$00FEFF83,$00000000,$FFFF7F00,$0F0000F8,$E0FFFFFF,
$00000000,$FFFFFF03,$030000C0,$FFFFFFFF,$0F0000F8,$C0FFFFFF,$030630E0,
$00FFFF81,$00000000,$FFFFFF00,$0F0000F8,$E0FFFFFF,$00000000,$FFFFFF07,
$030000C0,$FFFFFFFF,$0F0000F8,$C1FFFFFF,$030E38E0,$00FFFFC1,$00000000,
$FFFFFF00,$0F0000F8,$F0FFFFFF,$00000000,$FFFFFF07,$030000C0,$FFFFFFFF,
$0F0000F8,$C3FFFFFF,$030E38E0,$80FFFFC1,$00000000,$FFFFFF01,$0F0000F8,
$F0FFFFFF,$00000000,$FFFFFF0F,$030000C0,$FFFFFFFF,$0F0000F8,$C3FFFFFF,
$030F78E0,$C0FFFFE1,$00000000,$FFFFFF03,$1F0000FC,$F8FFFFFF,$00000000,
$FFFFFF0F,$030000C0,$FFFFFFFF,$0F0000F8,$87FFFFFF,$030F78E0,$E0FFFFF0,
$00024000,$FFFFFF07,$1F0000FC,$F8FFFFFF,$00000000,$FFFFFF1F,$070000E0,
$FFFFFFFF,$070000F0,$87FFFFFF,$030F78E0,$F0FFFFF0,$0003C000,$FFFFFF0F,
$3F0000FE,$F8FFFFFF,$00000000,$FFFFFF1F,$070000E0,$FFFFFFFF,$030000E0,
$8FFFFFFF,$831FFCE0,$FCFFFFF8,$C003C003,$FFFFFF3F,$7F0000FF,$F8FFFFFF,
$00000000,$FFFFFF3F,$0F0000F0,$FFFFFFFF,$00000080,$8FFFFFFF,$831FFCE0,
$FFFFFFF8,$FF03C0FF,$FFFFFFFF,$FF0080FF,$F8FFFFFF,$00000000,$FFFFFF7F,
$1F0000F8,$FFFFFFFF,$00000000,$1FFFFF7F,$831FFCE0,$FFFF7FFC,$FF07E0FF,
$FFFFFFFF,$FF01C0FF,$FCFFFFFF,$00000000,$FFFFFFFF,$3F0000FC,$FEFFFFFF,
$00000000,$3FFEFF3F,$C31FFCE1,$FFFF3FFE,$FF07E0FF,$FFFFFFFF,$FF07F0FF,
$FCFFFFFF,$01000000,$FFFFFFFF,$7F0000FE,$FCFFFFFF,$00000000,$1FFCFF1F,
$C31FFCE1,$FFFF1FFC,$FF0FF0FF,$FFFFFFFF,$FF03E0FF,$FEFFFFFF,$03000000,
$FFFFFFFF,$FF0180FF,$FCFFFFFF,$00000000,$0FF8FF1F,$C33FFEE1,$FFFF0FF8,
$FF1FF8FF,$FFFFFFFF,$FF03E0FF,$FFFFFFFF,$07000000,$FFFFFFFF,$FF03C0FF,
$FCFFFFFF,$00000000,$0FF8FF1F,$E33FFEE3,$FFFF0FF8,$FF7FFEFF,$FFFFFFFF,
$FF01C0FF,$FFFFFFFF,$0F000080,$FFFFFFFF,$FF07E0FF,$FCFFFFFF,$00000000,
$0FF8FF1F,$E33FFEE3,$FFFF0FF8,$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,
$3F000080,$FFFFFFFF,$FF1FF8FF,$FCFFFFFF,$00000000,$1FFCFF1F,$E33FFEE3,
$FFFF1FFC,$FF7FFEFF,$FFFFFFFF,$FF01C0FF,$FFFFFFFF,$FF000080,$FFFFFFFF,
$FF1FF8FF,$FCFFFFFF,$00000000,$3FFEFF1F,$C13FFEC1,$FFFF3FFE,$FF7FFEFF,
$FFFFFFFF,$FF01C0FF,$FFFFFFFF,$FF030080,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,
$00000000,$FFFFFF1F,$803FFE80,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FF03E0FF,
$FFFFFFFF,$FF7F0000,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,
$801FFC80,$FFFFFFFF,$FF07E0FF,$FFFFFFFF,$FF03E0FF,$FFFFFFFF,$FFFFC000,
$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,$FFFFFF1F,$800FF880,$FFFFFFFF,
$FF07E0FF,$FFFFFFFF,$FF07F0FF,$FFFFFFFF,$FFFFC003,$FFFFFFFF,$FF0FF0FF,
$FCFFFFFF,$C007F001,$FFFFFF1F,$C10FF8C1,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,
$FF1FFCFF,$FFFFFFFF,$FFFFE107,$FFFFFFFF,$FF0FF0FF,$FCFFFFFF,$C007F001,
$FFFFFF1F,$E30FF8E3,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFE10F,$FFFFFFFF,$FF1FF8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF1FFCFF,
$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFF3BF,$FFFFFFFF,
$FF3FFCFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF3FFEFF,$FFFFFFFF,$FF7FFEFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FFFFFFF,$FFFFFFFF,$FFFFFFFC,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF03C0FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$000000F0,$03FF3F00,$07FC3FE0,$FFFFFFC0,$FFFFFFFF,$FFFFFFFF,$FF0080FF,
$FFFFFFFF,$1F0000F8,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000F0,$40FE3F00,
$00E00700,$E0FF7F02,$00000000,$FFFFFF03,$030000E0,$FFFFFFFF,$03F81FC0,
$FEFFFFFF,$00000000,$FFFFFF3F,$FFFFFFF3,$78FE3FFC,$F0C1830F,$E0FF7F1E,
$00000000,$FEFFFF03,$00FE3F00,$FEFFFF3F,$C0FFFF03,$FCFFFF7F,$00000000,
$FFFFFF1F,$FFFFFFF3,$FFFC3FFC,$FF8FF1FF,$E7FF3FFF,$FFFFFFFF,$F8FFFFF3,
$FCFFFF1F,$F8FFFF0F,$FCFFFF3F,$FCFFFF1F,$FFFFFF7F,$FFFFFF1F,$FFFFFFF1,
$3FFC3FFC,$FF1FF8FF,$E7FF3FFC,$FFFFFFFF,$FCFFFFF3,$FCFFFF1F,$FCFFFF1F,
$C0FFFF03,$FCFFFF3F,$FFFFFFFF,$FFFFFF9F,$FFFFFFF9,$03FE3FFC,$013C3C80,
$E7FF7FC0,$FFFFFFFF,$FEFFFFF3,$00FE3F00,$FCFFFF3F,$02F81F40,$FCFFFF3F,
$FFFFFFFF,$FFFFFF9F,$FFFFFFF8,$80FF3FFC,$00300C00,$E0FFFF01,$00000000,
$FEFFFF03,$02000020,$FEFFFF3F,$1C000038,$FCFFFF7F,$FFFFFFFF,$FFFFFF9F,
$FFFFFFFC,$F8FF3FFC,$FC61863F,$E0FFFF1F,$00000000,$FFFFFF03,$FE00803F,
$FFFFFF7F,$FC03C03F,$FCFFFFFF,$FFFFFFFF,$FFFFFF9F,$FFFF7FFC,$FFFF3FFC,
$FF67E6FF,$FFFFFFFF,$FEFFFF3F,$FFFFFF7F,$FCFFFF1F,$FFFFFF7F,$E0FFFF07,
$FCFFFFFF,$FFFFFFFF,$FFFFFF9F,$FFFF3FFE,$FFFF3FFC,$FF07E0FF,$FFFFFFFF,
$FEFFFF3F,$FFFFFF7F,$F0FFFF87,$FFFFFFFF,$00FC3F00,$FCFFFFFF,$FFFFFF7F,
$FFFFFF1F,$FFFF1FFF,$FFFF3FFC,$FF0000FF,$FFFFFFFF,$FEFFFF3F,$FFFFFF7F,
$00FF7F00,$FFFFFF7F,$0C000030,$FEFFFFFF,$FFFFFF7F,$FCFFFF3F,$FFFF9F7F,
$FFFF3FFC,$1FF00FF8,$FFFFFFFF,$FEFFFF3F,$FFFFFF7F,$06000030,$FEFFFF7F,
$FC01803F,$FEFFFF7F,$FFFFFF7F,$C0FFFF3F,$FFFF8F3F,$FFFF3FFC,$0FFFFFF0,
$FFFFFFFF,$00000000,$FFFFFF7F,$7E00003F,$FEFFFF7F,$FEFFFF7F,$FEFFFF7F,
$FEFFFF3F,$01FFFF3F,$FFFFC71F,$FFFF3FFC,$0FFFFFF0,$FFFFFFFF,$00000000,
$FFFFFF7F,$FEFFFF3F,$FEFFFF7F,$FEFFFF7F,$FFFFFF7F,$FEFFFF3F,$11FEFF7F,
$FFFFE39F,$FFFF3FF8,$1FF00FF8,$FFFFFFFF,$F8FFFF8F,$FEFFFFFF,$F8FFFF0F,
$FCFFFF3F,$E0FFFF07,$FFFFFF3F,$FCFFFF1F,$38FEFF7F,$FFFFF18F,$FFFF7FF8,
$9F0000F9,$FFFFFFFF,$F1FFFFC7,$FEFFFFFF,$80FFFF00,$F8FFFF3F,$00FC3F00,
$FFFFFF1F,$FCFFFF9F,$7CFCFFFF,$FFFFF8C7,$FFFF7FF8,$1F07E0F8,$FFFFFFFF,
$E3FFFFE3,$FCFFFFFF,$03000060,$F1FFFF1F,$0F0000F0,$FFFFFF8F,$F8FFFF8F,
$DCFCFFFF,$FF7FFCE3,$FFFF7FF8,$3FFEFFFC,$FFFFFFFF,$070000F0,$F8FFFFFF,
$3F0000FE,$E3FFFF8F,$FF0180FF,$FFFFFFC7,$F1FFFFC7,$8FFCFFFF,$FF7FFEF1,
$FFFF7FF8,$3FFE7FFC,$FFFFFFFF,$0F0000F8,$F9FFFFFF,$FCFFFF9F,$C7FFFFCF,
$FF7FFEFF,$FFFFFFE3,$E3FFFFE3,$8FFCFFFF,$FF3F7EF8,$FFFF7FF8,$7FF00FFE,
$FFFFFFFF,$CFFFFFF9,$F0FFFFFF,$E0FFFF03,$8FFFFF07,$FF7FFEFF,$FFFFFFF1,
$87FFFFF0,$CFFCFFFF,$FF3F1FFE,$FFFF7FF0,$3F0000FC,$FFFFFFFF,$CFFFFFF9,
$F0FFFFFF,$C0E1C301,$1FFFFF07,$FF7FFEFF,$FFFFFFF8,$0FFE3FF8,$7FFCFFFF,
$FF1F07FF,$FFFFFFF0,$1F07E0F8,$FFFFFFFF,$CFFFFFF9,$F3FFFFFF,$00800080,
$3FFFFFE7,$FF7FFEFF,$FFFFFFFC,$3FF80FFE,$7FFEFFFF,$FF9FC1FF,$FFFFFFF0,
$8FFFFFF1,$FFFFFFFF,$CFFFFFF9,$F3FFFFFF,$10100484,$3FFEFFE7,$FF3FFCFF,
$FFFF7FFC,$FFF88FFF,$3FFEFFFF,$FF9FF0FF,$FFFFFFF0,$CFFFFFF3,$FFFFFFFF,
$CFFFFFF9,$F3FFFFFF,$30314606,$7FFEFF67,$FF3FFCFF,$FFFF7FFE,$7FFC1FFF,
$3FFFFFFF,$FF1FF8FF,$FFFFFFE0,$C7FFFFE3,$FFFFFFFF,$CFFFFFF9,$E3FFFFFF,
$F261C327,$7FFEFF63,$FF3FFCFF,$FFFF7FFE,$3FFE3FFE,$1FFFFFFF,$FF3FFEFF,
$FFFFFFE1,$E77FFEE7,$FFFFFFFF,$CFFFFFF9,$E6FFFFFF,$F2E08327,$7FFEFF33,
$FF1FF8FF,$FFFF7FFE,$1FFF7FFC,$8FFFFFFF,$FF1FFFFF,$FFFFFFE1,$E37FFEC7,
$FFFFFFFF,$CFFFFFF9,$E6FFFFFF,$F3E49367,$7FFEFF33,$FF9FF9FF,$FFFF7FFE,
$9FFFFFFC,$CFFFFFFF,$FF1FFFFF,$FFFFFFE1,$F37FFECF,$FFFFFFFF,$CFFFFFF9,
$E4FFFFFF,$F3E49367,$7FFEFF13,$FF8FF1FF,$FFFF7FFE,$8FFFFFF8,$C7FFFFFF,
$FF9FFFFF,$FFFFFFC3,$F37FFECF,$FFFFFFFF,$CFFFFFF9,$E4FFFFFF,$73E413E7,
$3FFEFF93,$FFCFF3FF,$FFFF7FFC,$CFFFFFF9,$E7FFFFFF,$FF9FFFFF,$FFFFFFC3,
$F37FFECF,$FFFFFFFF,$CFFFFFF9,$C0FFFFFF,$73C631E7,$3FFFFF81,$FFC7E3FF,
$FFFFFFFC,$CFFFFFF9,$E3FFFFFF,$FF8FFFFF,$FFFFFF87,$F303C0CF,$FFFFFFFF,
$CFFFFFF9,$C1FFFFFF,$73CE39E7,$1FFFFFC1,$FFE3C7FF,$FFFFFFF8,$CFFFFFF9,
$F3FFFFFF,$FFCFFFFF,$FFFFFF87,$F303C0CF,$FFFFFFFF,$CFFFFFF9,$C3FFFFFF,
$33CE39E6,$8FFFFFC1,$FFE187FF,$FFFFFFF1,$8FFFFFF8,$F1FFFFFF,$FFCFFF3F,
$FFFFFF0F,$F37FFECF,$FFFFFFFF,$CFFFFFF9,$C3FFFFFF,$33CF79E6,$C7FFFFE1,
$7FE007FE,$FFFFFFE3,$9FFFFFFC,$F9FFFFFF,$FFDFFF0F,$FFFFFF0F,$E37FFEC7,
$FFFFFFFF,$CFFFFFF9,$87FFFFFF,$33CF79E6,$E1FFFFF0,$1FF24FF8,$FFFFFF87,
$1FFF7FFC,$F9FFFFFF,$FEFFFF07,$FFFFFF1F,$E77FFEE7,$FFFFFFFF,$070000F0,
$87FFFFFF,$138F78E4,$F0FFFFF0,$00F3CF00,$FFFFFF0F,$3FFE3FFE,$F9FFFFFF,
$FCFFFF87,$FFFFFF1F,$C77FFEE3,$FFFFFFFF,$030000E0,$8FFFFFFF,$939FFCE4,
$FCFFFFF8,$C0F3CF03,$FFFFFF3F,$7FFC1FFF,$F9FFFFFF,$FCFFFF83,$FFFFFF3F,
$8F7FFEF1,$FFFFFFFF,$E0FFFF83,$8FFFFFFF,$939FFCE4,$FFFFFFF8,$FFE3C7FF,
$FFFFFFFF,$FFF087FF,$F8FFFFFF,$F8FFFFE3,$FFFFFF7F,$1FFFFFF8,$FFFFFFFF,
$F8FFFF0F,$1FFFFF7F,$839FFCE0,$FFFF7FFC,$FFE7E7FF,$FFFFFFFF,$FFC1C1FF,
$FCFFFFFF,$F0FFFFFF,$FFFFFFFF,$3FFC3FFC,$FEFFFFFF,$FCFFFF1F,$3FFEFF3F,
$C39FFCE1,$FFFF3FFE,$FF87E1FF,$FFFFFFFF,$FFC7F1FF,$FCFFFFFF,$E1FFFF7F,
$FFFFFFFF,$7FF81FFE,$FCFFFFFF,$00000000,$1FFCFF1F,$C31FFCE1,$FFFF1FFC,
$FF0FF0FF,$FFFFFFFF,$FFE3E3FF,$FEFFFFFF,$C3FFFF33,$FFFFFFFF,$FFF18FFF,
$FCFFFFFF,$00000000,$8FF8FF1F,$C33FFEE1,$FFFF8FF8,$FF1FF8FF,$FFFFFFFF,
$FFF3E7FF,$FFFFFFFF,$877FFE11,$FFFFFFFF,$FFE3C7FF,$FCFFFFFF,$FFFFFFFF,
$CFF9FF9F,$E33FFEE3,$FFFFCFF9,$FF7FFEFF,$FFFFFFFF,$FFF1C7FF,$FFFFFFFF,
$0F7E9E90,$FFFFFFFF,$FF87E1FF,$FCFFFFFF,$FFFFFFFF,$8FF8FF9F,$E33FFEE3,
$FFFF8FF8,$FF7FFEFF,$FFFFFFFF,$FFF9CFFF,$FFFFFFFF,$3F701E98,$FFFFFFFF,
$FF1FF8FF,$FCFFFFFF,$FFFFFFFF,$1FFCFF9F,$E33FFEE3,$FFFF1FFC,$FF7FFEFF,
$FFFFFFFF,$FFF9CFFF,$FFFFFFFF,$FF001E9E,$FFFFFFFF,$FF1FF8FF,$FCFFFFFF,
$0FE003F8,$3FFEFF9F,$C13FFEC1,$FFFF3FFE,$FF7FFEFF,$FFFFFFFF,$FFF1C7FF,
$FFFFFFFF,$FF030E9C,$FFFFFFFF,$FF8FF1FF,$FCFFFFFF,$0FE003F8,$FFFFFF9F,
$883FFE88,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFF3E7FF,$FFFFFFFF,$FF7F0C18,
$FFFFFFFF,$FFCFF3FF,$FCFFFFFF,$CFE7F3F9,$FFFFFF9F,$9C1FFC9C,$FFFFFFFF,
$FF07E0FF,$FFFFFFFF,$FFC3E1FF,$FFFFFFFF,$FFFFCC30,$FFFFFFFF,$FFCFF3FF,
$FCFFFFFF,$CFE7F3F9,$FFFFFF9F,$888FF888,$FFFFFFFF,$FF07E0FF,$FFFFFFFF,
$FF07F0FF,$FFFFFFFF,$FFFFC823,$FFFFFFFF,$FFCFF3FF,$FCFFFFFF,$C007F001,
$FFFFFF1F,$C1CFF9C1,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FF1FFCFF,$FFFFFFFF,
$FFFFE107,$FFFFFFFF,$FF8FF1FF,$FCFFFFFF,$C007F001,$FFFFFF1F,$E38FF8E3,
$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFE10F,$FFFFFFFF,
$FF1FF8FF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FF1FFCFF,$FFFFFFFF,$FF7FFEFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFF3BF,$FFFFFFFF,$FF3FFCFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FF3FFEFF,$FFFFFFFF,$FF7FFEFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0000FFFF);

set80: Array [0..4815] of Integer=(
$4B3E4D42,$00000000,$003E0000,$00280000,$01E00000,$01400000,$00010000,
$00000001,$4B000000,$00000000,$00000000,$00020000,$00020000,$00000000,
$FFFF0000,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,$00000000,$DFFF7F00,$FFFFFFFF,
$FFFBFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FFFFFF3F,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,$00000000,$87FF7F00,
$FC3FC0FF,$FFE1FF03,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000FF,$FFFFFF01,
$FFFFFFFF,$FFFF03C0,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,$00000000,
$80FF7F00,$F00F0070,$FF010E00,$0000F8FF,$00000000,$FFFFFF3F,$000000F0,
$FFFF1F00,$F0FFFFFF,$FF0F0000,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,
$00000000,$00FF7F00,$C0030000,$FF000000,$0000F8FF,$00000000,$FFFFFF3F,
$000000C0,$FFFF0700,$00FFFFFF,$FF000000,$FFFFFFFF,$00000000,$FFFF0000,
$C0FFFFFF,$00000000,$00FF7F00,$80010000,$FF000000,$0000F8FF,$00000000,
$FFFFFF3F,$00000000,$FFFF0100,$00F8FFFF,$1F000000,$FFFFFFFF,$00000000,
$FFFF0000,$C0FFFFFF,$00000000,$00FE7F00,$80010000,$7F000000,$0000F8FF,
$00000000,$FFFFFF3F,$00000000,$FFFF0100,$00E0FFFF,$07000000,$FFFFFFFF,
$00000000,$FFFF0000,$C0FFFFFF,$00000000,$00FE7F00,$00000000,$7F000000,
$0000F8FF,$00000000,$FFFFFF3F,$00000080,$FFFF0300,$00C0FFFF,$03000000,
$FFFFFFFF,$00000000,$FFFF0000,$E0FFFFFF,$00000000,$00FC7F00,$00000000,
$3F000000,$0000F8FF,$00000000,$FFFFFF3F,$00000080,$FFFF0300,$0080FFFF,
$01000000,$FFFFFFFF,$00000000,$FFFF0000,$E0FFFFFF,$00000000,$00FE7F00,
$00000000,$7F000000,$0000F8FF,$00000000,$FFFFFF3F,$000000C0,$FFFF0700,
$00C0FFFF,$03000000,$FFFFFFFF,$00000000,$FFFF0000,$E0FFFFFF,$00000000,
$00FF7F00,$00000000,$FF000000,$0000F8FF,$00000000,$FFFFFF3F,$000000C0,
$FFFF0700,$00C0FFFF,$03000000,$FFFFFFFF,$00000000,$FFFF0000,$F0FFFFFF,
$00000000,$C0FF7F00,$0000FC00,$FF03003F,$00C0FFFF,$07000000,$FFFFFFFF,
$000000E0,$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,$00000000,$FFFF0000,
$F0FFFFFF,$00000000,$F8FF7F00,$0000FF1F,$FF1FF8FF,$00C0FFFF,$07000000,
$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,$00000080,
$FFFF0100,$F8FFFFFF,$00000000,$FFFF7F00,$0180FFFF,$FFFFFFFF,$00C0FFFF,
$07000000,$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,
$00000080,$FFFF0100,$FC9FFFFF,$00000000,$FFFFFF00,$0180FFFF,$FFFFFFFF,
$00C0FFFF,$07000000,$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,$07000000,
$FFFFFFFF,$00000080,$FFFF0100,$FE07FCFF,$00000000,$FFFFFF00,$0000F0FF,
$FFFFFF0F,$00C0FFFF,$07000000,$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,
$07000000,$FFFFFFFF,$000000C0,$FFFF0300,$FF03F0FF,$00000000,$FFFFFF00,
$0000C0FF,$FFFFFF03,$00C0FFFF,$07000000,$FFFFFFFF,$000000C0,$FFFF0700,
$00E0FFFF,$07000000,$FFFFFFFF,$000000C0,$FFFF0300,$FF01E0FF,$00000080,
$FFFFFF00,$000080FF,$FFFFFF01,$00C0FFFF,$07000000,$FFFFFFFF,$000000C0,
$FFFF0700,$00C0FFFF,$03000000,$FFFFFFFF,$000000E0,$FFFF0700,$FF00C0FF,
$000000C0,$FFFFFF00,$000080FF,$FFFFFF01,$00E0FFFF,$0F000000,$FFFFFFFF,
$00000080,$FFFF0300,$00C0FFFF,$03000000,$FFFFFFFF,$000000E0,$FFFF0700,
$7F0080FF,$000000E0,$FFFFFF00,$000080FF,$FFFFFF01,$00F8FFFF,$3F000000,
$FFFFFFFF,$00000000,$FFFF0100,$00C0FFFF,$03000000,$FFFFFFFF,$000000F0,
$FFFF0F00,$7F0000FF,$000000F0,$FFFFFF00,$0000C0FF,$FFFFFF03,$00FCFFFF,
$7F000000,$FEFFFFFF,$00000000,$FFFF0000,$0080FFFF,$01000000,$FFFFFFFF,
$000000F0,$FFFF0F00,$3F0000FF,$000000F8,$FFFFFF00,$0000C0FF,$FFFFFF03,
$00FEFFFF,$FF000000,$FEFFFFFF,$00000000,$FFFF0000,$0000FFFF,$00000000,
$FFFFFFFF,$000000F8,$FFFF1F00,$1F0000FE,$000000FC,$FFFFFF00,$0000C0FF,
$FFFFFF03,$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,$FF7F0000,$0000FEFF,
$00000000,$FFFFFF7F,$000000FC,$FFFF3F00,$0F0000FE,$000000FC,$FFFFFF01,
$0000C0FF,$FFFFFF03,$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,$FF7F0000,
$0000FCFF,$00000000,$FFFFFF3F,$000000FE,$FFFF7F00,$070000FE,$000000FE,
$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,
$FF7F0000,$0000F8FF,$00000000,$FFFFFF1F,$000000FF,$FFFFFF00,$030000FE,
$000000FE,$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,$FCFFFFFF,
$00000000,$FF7F0000,$0000F0FF,$00000000,$FFFFFF0F,$0000C0FF,$FFFFFF03,
$010000FE,$000000FF,$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,
$F8FFFFFF,$00000000,$FF3F0000,$0000E0FF,$00000000,$FFFFFF07,$0000F0FF,
$FFFFFF0F,$000000FE,$0000001F,$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,
$FF010000,$F8FFFFFF,$00000000,$FF3F0000,$0000C0FF,$00000000,$FFFFFF03,
$0000FCFF,$FFFFFF3F,$000000FE,$00008007,$FFFFFF03,$0000C0FF,$FFFFFF03,
$00FFFFFF,$FF010000,$F8FFFFFF,$00000000,$FF3F0000,$0000C0FF,$00000000,
$FFFFFF03,$0000FFFF,$FFFFFFFF,$000000FF,$00008003,$FFFFFF03,$000080FF,
$FFFFFF01,$00FFFFFF,$FF010000,$F8FFFFFF,$00000000,$FF3F0000,$000080FF,
$00000000,$FFFFFF01,$0000FCFF,$FFFFFF3F,$000000FF,$00008001,$FFFFFF03,
$000000FF,$FFFFFF00,$00FFFFFF,$FF010000,$F0FFFFFF,$00000000,$FF1F0000,
$000080FF,$00000000,$FFFFFF01,$0000F8FF,$FFFFFF1F,$000080FF,$00008000,
$FFFFFF03,$000000FF,$FFFFFF00,$00FFFFFF,$FF010000,$F0FFFFFF,$00000004,
$FF1F4000,$000080FF,$00000000,$FFFFFF01,$0000F0FF,$FFFFFF0F,$000080FF,
$00000000,$FFFFFF07,$000000FE,$FFFF7F00,$00FFFFFF,$FF010000,$F0FFFFFF,
$00000004,$FF1F4000,$000000FF,$00000000,$FFFFFF00,$0000E0FF,$FFFFFF07,
$0000C0FF,$00000000,$FFFFFF07,$000000FE,$FFFF7F00,$00FFFFFF,$FF010000,
$F0FFFFFF,$0000000C,$FF1F6000,$000000FF,$00000000,$FFFFFF00,$0000E0FF,
$FFFFFF07,$0000E0FF,$00000000,$FFFFFF07,$000000FC,$FFFF3F00,$00FFFFFF,
$FF010000,$F0FFFFFF,$06C0000C,$FF1F6000,$000000FF,$00000000,$FFFFFF00,
$0000C0FF,$FFFFFF03,$0000E0FF,$00000000,$FFFFFF0F,$000000FC,$FFFF3F00,
$00FFFFFF,$FF010000,$E0FFFFFF,$06C0001C,$FF0F7000,$000000FF,$00000000,
$FFFFFF00,$0000C0FF,$FFFFFF03,$0000F0FF,$00000000,$FFFFFF0F,$000000FC,
$FFFF3F00,$00FFFFFF,$FF010000,$E0FFFFFF,$06C0001C,$FF0F7000,$000000FF,
$00000000,$FFFFFF00,$000080FF,$FFFFFF01,$0000F0FF,$00000000,$FFFFFF0F,
$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,$E0FFFFFF,$07C0013C,$FF0F7800,
$000000FF,$00000000,$FFFFFF00,$000080FF,$FFFFFF01,$0000F8FF,$00000000,
$FFFFFF1F,$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,$E0FFFFFF,$07C0013C,
$FF0F7800,$000000FF,$00000000,$FFFFFF00,$000080FF,$FFFFFF01,$0000FCFF,
$00000000,$FFFFFF1F,$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,$E0FFFFFF,
$0FE0017C,$FF0F7C00,$000000FF,$00000000,$FFFFFF00,$000080FF,$FFFFFF01,
$0000FCFF,$00000000,$FFFFFF3F,$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,
$C0FFFFFF,$0FE0037C,$FF077C80,$000080FF,$00000000,$FFFFFF01,$000080FF,
$FFFFFF01,$0000FEFF,$00000000,$FFFFFF3F,$000000F8,$FFFF1F00,$00FFFFFF,
$FF010000,$C0FFFFFF,$0FE003FC,$FF077E80,$000080FF,$00000000,$FFFFFF01,
$000080FF,$FFFFFF01,$0000FEFF,$00000000,$FFFFFF7F,$000000F8,$FFFF1F00,
$00FFFFFF,$FF010000,$C0FFFFFF,$0FE003FC,$FF077E80,$000080FF,$00000420,
$FFFFFF01,$0000C0FF,$FFFFFF03,$0000FFFF,$00000000,$FFFFFF7F,$000000FC,
$FFFF3F00,$00FFFFFF,$FF010000,$C1FFFFFF,$0FE007FC,$FF077FC0,$0000C0FF,
$00000660,$FFFFFF03,$0000C0FF,$FFFFFF03,$0000FFFF,$00000000,$FFFFFFFF,
$000000FC,$FFFF3F00,$00FFFFFF,$FF010000,$C1FFFFFF,$1FF007FC,$FF077FC0,
$0000E0FF,$000003C0,$FFFFFF07,$0000E0FF,$FFFFFF07,$0000FFFF,$00000000,
$FFFFFFFF,$000000FC,$FFFF3F00,$00FFFFFF,$FF010000,$83FFFFFF,$1FF007FC,
$FF837FC0,$0100F0FF,$008003C0,$FFFFFF0F,$0000E0FF,$FFFFFF07,$0000FFFF,
$01000000,$FFFFFFFF,$000000FE,$FFFF7F00,$00FFFFFF,$FF010000,$83FFFFFF,
$1FF00FFC,$FF837FE0,$0300F8FF,$00C00180,$FFFFFF1F,$0000F0FF,$FFFFFF0F,
$0000FFFF,$01000000,$FFFFFFFF,$000000FE,$FFFF7F00,$00FEFFFF,$FF000000,
$87FFFFFF,$1FF00FFC,$FFC37FE0,$0F00FEFF,$00F00180,$FFFFFF7F,$0000F8FF,
$FFFFFF1F,$0000FFFF,$03000000,$FFFFFFFF,$000000FF,$FFFFFF00,$00FCFFFF,
$7F000000,$87FFFFFF,$1FF00FFC,$FFC37FE0,$7FC0FFFF,$03FE0180,$FFFFFFFF,
$0000FCFF,$FFFFFF3F,$0080FFFF,$07000000,$FFFFFFFF,$000080FF,$FFFFFF01,
$00F8FFFF,$3F000000,$0FFFFFFF,$3FF81FFC,$FFE17FF0,$FFFFFFFF,$FFFF0180,
$FFFFFFFF,$0000FFFF,$FFFFFFFF,$0080FFFF,$0F000000,$FFFFFFFF,$0000C0FF,
$FFFFFF03,$00F0FFFF,$1F000000,$1FFCFFFF,$3FF81FFC,$7FF07FF0,$FFFFFFFF,
$FFFF0180,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$00C0FFFF,$1F000000,$FFFFFFFF,
$0000E0FF,$FFFFFF07,$00E0FFFF,$0F000000,$0FF8FFFF,$3FF81FFC,$3FE07FF0,
$FFFFFFFF,$FFFF03C0,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$00C0FFFF,$3F000000,
$FFFFFFFF,$0000F8FF,$FFFFFF1F,$00C0FFFF,$07000000,$07F0FFFF,$3FF83FFC,
$1FC07FF8,$FFFFFFFF,$FFFF03C0,$FFFFFFFF,$0180FFFF,$FFFFFFFF,$00E0FFFF,
$7F000000,$FFFFFFFF,$0000FEFF,$FFFFFF7F,$00C0FFFF,$07000000,$07F0FFFF,
$3FF83FFC,$1FC07FF8,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,$0180FFFF,$FFFFFFFF,
$00E0FFFF,$FF000000,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$00C0FFFF,$07000000,
$07F0FFFF,$7FFC3FFC,$1FC07FF8,$FFFFFFFF,$FFFF0FF0,$FFFFFFFF,$0000FFFF,
$FFFFFFFF,$00F0FFFF,$FF030000,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$00C0FFFF,
$07000000,$07F0FFFF,$7FFC7FFC,$1FC07FFC,$FFFFFFFF,$FFFF3FFC,$FFFFFFFF,
$0000FFFF,$FFFFFFFF,$00F8FFFF,$FF070000,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,
$00C0FFFF,$07000000,$07F0FFFF,$7FFC7FFC,$1FC07FFC,$FFFFFFFF,$FFFF7FFE,
$FFFFFFFF,$0000FFFF,$FFFFFFFF,$00F8FFFF,$FF1F0000,$FFFFFFFF,$0FF0FFFF,
$FFFFFFFF,$00C0FFFF,$07000000,$0FF8FFFF,$7FFC3FF8,$3FE03FF8,$FFFFFFFF,
$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$00F8FFFF,$FFFF0000,$FFFFFFFF,
$3FFCFFFF,$FFFFFFFF,$00C0FFFF,$07000000,$1FFCFFFF,$7FFC1FF0,$7FF01FF0,
$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$07F8FFFF,$FFFF7F00,
$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$FFFFFFFF,$3FF80FE0,
$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$0FF0FFFF,
$FFFFFF00,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,
$1FF00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0180FFFF,$FFFFFFFF,
$1FF0FFFF,$FFFFFF80,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,
$FFFFFFFF,$0FE00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,$0180FFFF,
$FFFFFFFF,$3FF0FFFF,$FFFFFF81,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$0FC0FFFF,
$07E007C0,$FFFFFFFF,$0FE00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,
$03C0FFFF,$FFFFFFFF,$7FF0FFFF,$FFFFFFC1,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,
$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,
$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$FFF8FFFF,$FFFFFFC3,$FFFFFFFF,$03C0FFFF,
$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE01FF0,$FFFF1FF0,$FFFFFFFF,
$FFFF7FFE,$FFFFFFFF,$1FF8FFFF,$FFFFFFFF,$FFF9FFFF,$FFFFFFC3,$FFFFFFFF,
$07E0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE03FF8,$FFFF3FF8,
$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFE7,
$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$1FF0FFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFE7,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$3FF8FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FFCFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,
$00000000,$DFFF7F00,$FFFFFFFF,$FFFBFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$0000F8FF,$FFFFFF3F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$C0FFFFFF,$00000000,$87FF7F00,$FC3FC0FF,$FFE1FF03,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$000000FF,$FFFFFF01,$FFFFFFFF,$FFFF03C0,$FFFFFFFF,$00000000,
$FFFF0000,$C0FFFFFF,$00000000,$80FF7F00,$F00F0070,$FF010E00,$0000F8FF,
$00000000,$FFFFFF3F,$000000F0,$FFFF1F00,$F0FFFFFF,$FF0F0000,$FFFFFFFF,
$00000000,$FFFF0000,$C0FFFFFF,$00000000,$00FF7F38,$C0030000,$FF000000,
$0000F8FF,$00000000,$FFFFFF3F,$000000C0,$FFFF0700,$00FFFFFF,$FF000000,
$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,$00000000,$00FF7F38,$80010000,
$FF000000,$0000F8FF,$00000000,$FFFFFF3F,$00000000,$FFFF0100,$00F8FFFF,
$1F000000,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,$00000000,$00FE7F38,
$80010000,$7F000000,$0000F8FF,$00000000,$FFFFFF3F,$0000001C,$FFFF7100,
$00E0FFFF,$07000000,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,$00000000,
$00FE7F38,$00000000,$7F000000,$0000F8FF,$00000000,$FFFFFF3F,$0000809F,
$FFFFF303,$00C6FFFF,$63000000,$FFFFFFFF,$00000000,$FFFF0000,$E0FFFFFF,
$00000000,$00FC7F38,$00000000,$3F000000,$0000F8FF,$00000000,$FFFFFF3F,
$0000FE8F,$FFFFE3FF,$E08FFFFF,$F1070000,$FFFFFFFF,$00000000,$FFFF0000,
$E0FFFFFF,$00000000,$00FE7F38,$00000000,$7F000000,$FF1FF8FF,$F0FFFFFF,
$FFFFFF3F,$FFFFFFC7,$FFFFC7FF,$FFC7FFFF,$E3FF03C0,$FFFFFFFF,$00000000,
$FFFF0000,$E0FFFFFF,$00000000,$00FF7F38,$00000000,$FF000000,$FF1FF8FF,
$F0FFFFFF,$FFFFFF3F,$FFFFFFC0,$FFFF07FE,$FFC1FFFF,$83FFFFFF,$FFFFFFFF,
$00000000,$FFFF0000,$F0FFFFFF,$00000000,$C0FF7F38,$781EFC00,$FF03003F,
$FFCFFFFF,$E7FFFFFF,$FFFFFFFF,$FFFF03E0,$FFFF0F80,$1FE0FFFF,$07F8FFFF,
$FFFFFFFF,$00000000,$FFFF0000,$F0FFFFFF,$00000000,$F8FF7F38,$781EFF1F,
$FF1FF8FF,$00C0FFFF,$07000000,$FFFFFFFF,$000000E0,$FFFF0F00,$00E6FFFF,
$6700FC3F,$FFFFFFFF,$00000080,$FFFF0100,$F8FFFFFF,$00000000,$FFFF7F78,
$718EFFFF,$FFFFFFFF,$00C0FFFF,$07000000,$FFFFFFFF,$000000E6,$FFFFCF00,
$E0E7FFFF,$E7070000,$FFFFFFFF,$00000080,$FFFF0100,$FC9FFFFF,$00000000,
$FFFFFF70,$0180FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$FFFFFFFF,$0000E0E7,
$FFFFCF0F,$FFE7FFFF,$E7FF0180,$FFFFFFFF,$00000080,$FFFF0100,$FE07FCFF,
$00000000,$FFFFFF70,$0000F0FF,$FFFFFF0F,$FFCFFFFF,$E7FFFFFF,$FFFFFFFF,
$0380FFE7,$FFFFCFFF,$FFE1FFFF,$87FFFFFF,$FFFFFFFF,$000000C0,$FFFF0300,
$FF03F0FF,$00000000,$FFFFFF70,$0000C0FF,$FFFFFF03,$FFCFFFFF,$E7FFFFFF,
$FFFFFFFF,$FFFFFFC3,$FFFF87FF,$1FE0FFFF,$07F8FFFF,$FFFFFFFF,$000000C0,
$FFFF0300,$FF41E0FF,$00000080,$FFFFFF70,$000080FF,$FFFFFF01,$FFCFFFFF,
$E7FFFFFF,$FFFFFFFF,$FFFF3FC0,$FFFF07F8,$00C0FFFF,$0300FE7F,$FFFFFFFF,
$000000E0,$FFFF0700,$FFC0C0FF,$000000C0,$FFFFFF70,$00009CFF,$FFFFFF39,
$00E0FFFF,$0F000000,$FFFFFFFF,$FEFF0080,$FFFF0300,$00C0FFFF,$03000000,
$FFFFFFFF,$000000E0,$FFFF0700,$7FE080FF,$000000E0,$FFFFFF70,$0FF09FFF,
$FFFFFFF9,$00F8FFFF,$3F000000,$FFFFFFFF,$00000018,$FFFF3100,$00C0FFFF,
$03000000,$FFFFFFFF,$000000F0,$FFFF0F00,$7F7000FF,$000000F0,$FFFFFF70,
$FFFFCFFF,$FFFFFFF3,$00FCFFFF,$7F000000,$FEFFFFFF,$0000803F,$FFFFF803,
$008EFFFF,$71000000,$FFFFFFFF,$000000F0,$FFFF0F00,$3F3000FF,$000000F8,
$FFFFFF70,$FFFFC7FF,$FFFFFFE3,$7FFEFFFF,$FFFCFFFF,$FEFFFFFF,$0100FF1F,
$FFFFF0FF,$E01FFFFF,$F8070000,$FFFFFFFF,$000000F8,$FFFF1F00,$1F2000FE,
$000000FC,$FFFFFFF0,$F81FC0FF,$FFFFFF03,$3FFFFFFF,$FFF9FFFF,$FCFFFFFF,
$FFFFFF0F,$FF7FE0FF,$FF3FFEFF,$FCFF0180,$FFFFFF7F,$000000FC,$FFFF3F00,
$0F0018FE,$000000FC,$FFFFFFE1,$0000C0FF,$FFFFFF03,$3FFFFFFF,$FFF9FFFF,
$FCFFFFFF,$FFFFFF00,$FF7F00FE,$FF7BFCFF,$DEFF7FFE,$FFFFFF3F,$000000FE,
$FFFF7F00,$07003CFE,$000000FE,$FFFFFFE1,$0000E4FF,$FFFFFF27,$00FFFFFF,
$FF010000,$FCFFFFFF,$FFFF0100,$FF7F0000,$3FF0F8FF,$0FFC7FFE,$FFFFFF1F,
$000000FF,$FFFFFF00,$03001CFE,$000000FE,$FFFFFFE1,$07E0E7FF,$FFFFFFE7,
$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,$FF7F0000,$00E0F1FF,$07007FFE,
$FFFFFF8F,$0000C0FF,$FFFFFF03,$01001EFE,$000000FF,$FFFFFFE1,$FFFFE7FF,
$FFFFFFE7,$00FFFFFF,$FF010000,$F8FFFFFF,$00000000,$FF3F0000,$00C0E3FF,
$0300700E,$FFFFFFC7,$0000F0FF,$FFFFFF0F,$00000EFE,$0100001F,$FFFFFFE1,
$FFFFE3FF,$FFFFFFC7,$00FFFFFF,$FF010000,$F8FFFFFF,$00000000,$FF3F0000,
$0080C7FF,$0100700E,$FFFFFFE3,$0000FCFF,$FFFFFF3F,$000006FE,$01008007,
$FFFFFFC3,$FC3FC0FF,$FFFFFF03,$00FFFFFF,$FF010000,$F8FFFFFF,$00000000,
$FF3F0000,$0000CFFF,$0000700E,$FFFFFFF3,$0000FFFF,$FFFFFFFF,$000000FF,
$01008003,$FFFFFFC3,$000080FF,$FFFFFF01,$00FFFFFF,$FF010000,$F8FFFFFF,
$00000000,$FF3F0000,$00008EFF,$0000781E,$FFFFFF71,$0000FCFF,$FFFFFF3F,
$000000FF,$01008001,$FFFFFFC3,$000000FF,$FFFFFF00,$00FFFFFF,$FF010000,
$F0FFFFFF,$00000000,$FF1F0000,$00009EFF,$0000381C,$FFFFFF79,$0000F8FF,
$FFFFFF1F,$000080FF,$03008000,$FFFFFFC3,$000000FF,$FFFFFF00,$00FFFFFF,
$FF010000,$F0FFFFFF,$00000004,$FF1F4000,$00009CFF,$0000381C,$FFFFFF39,
$0000F0FF,$FFFFFF0F,$000080FF,$03000000,$FFFFFF87,$000000FE,$FFFF7F00,
$00FFFFFF,$FF010000,$F0FFFFFF,$00000004,$FF1F4000,$00001CFF,$0000381C,
$FFFFFF38,$0000E0FF,$FFFFFF07,$0000C0FF,$03000000,$FFFFFF87,$000000FE,
$FFFF7F00,$00FFFFFF,$FF010000,$F0FFFFFF,$0000000C,$FF1F6000,$00003CFF,
$00003C3C,$FFFFFF3C,$0000E0FF,$FFFFFF07,$0000E0FF,$0700E000,$FFFFFF87,
$000000FC,$FFFF3F00,$00FFFFFF,$FF010000,$F0FFFFFF,$06C0000C,$FF1F6000,
$000038FF,$00001C38,$FFFFFF1C,$0000C0FF,$FFFFFF03,$0000E0FF,$0700E000,
$FFFFFF0F,$C00100FC,$FFFF3F00,$00FFFFFF,$FF010000,$E0FFFFFF,$06C0001C,
$FF0F7000,$000038FF,$00009E79,$FFFFFF1C,$0000C0FF,$FFFFFF03,$0000F0FF,
$07006000,$FFFFFF0F,$C00100FC,$FFFF3F00,$00FFFFFF,$FF010000,$E0FFFFFF,
$06C0001C,$FF0F7000,$000038FF,$00008E71,$FFFFFF1C,$000080FF,$FFFFFF01,
$0000F0FF,$0F007000,$FFFFFF0F,$C00100F8,$FFFF1F00,$00FFFFFF,$FF010000,
$E0FFFFFF,$07C0013C,$FF0F7800,$000038FF,$00008FF1,$FFFFFF1C,$000080FF,
$FFFFFF01,$0000F8FF,$0E007000,$FFFFFF1F,$C00100F8,$FFFF1F00,$00FFFFFF,
$FF010000,$E0FFFFFF,$07C0013C,$FF0F7800,$01003CFF,$0080C7E3,$FFFFFF3C,
$000080FF,$FFFFFF01,$0000FCFF,$1E003000,$FFFFFF1F,$C00100F8,$FFFF1F00,
$00FFFFFF,$FF010000,$E0FFFFFF,$0FE0017C,$FF0F7C00,$03001CFF,$00C0C3C3,
$FFFFFF38,$000080FF,$FFFFFF01,$0000FCFF,$1C003000,$FFFFFF3F,$C00100F8,
$FFFF1F00,$00FFFFFF,$FF010000,$C0FFFFFF,$0FE0037C,$FF077C80,$07009CFF,
$00E0C183,$FFFFFF39,$000080FF,$FFFFFF01,$0000FEFF,$3C003000,$FFFFFF3F,
$FF7F00F8,$FFFF1F00,$00FFFFFF,$FF010000,$C0FFFFFF,$0FE003FC,$FF077E80,
$0F009EFF,$00F0E007,$FFFFFF79,$000080FF,$FFFFFF01,$0008FEFF,$38003000,
$FFFFFF7F,$FF7F00F8,$FFFF1F00,$00FFFFFF,$FF010000,$C0FFFFFF,$0FE003FC,
$FF077E80,$3E808FFF,$017C6426,$FFFFFFF1,$0000C0FF,$FFFFFF03,$C00CFFFF,
$78000000,$FFFFFF7F,$FF7F00FC,$FFFF3F00,$00FFFFFF,$FF010000,$C1FFFFFF,
$0FE007FC,$FF077FC0,$FCF1C7FF,$8F3F6666,$FFFFFFE3,$0000C0FF,$FFFFFF03,
$C00EFFFF,$70000000,$FFFFFFFF,$C00100FC,$FFFF3F00,$00FFFFFF,$FF010000,
$C1FFFFFF,$1FF007FC,$FF077FC0,$F8FFE3FF,$FF1F73CE,$FFFFFFC7,$0000E0FF,
$FFFFFF07,$E007FFFF,$F0000000,$FFFFFFFF,$C00100FC,$FFFF3F00,$3FFFFFFF,
$FFF9FFFF,$83FFFFFF,$1FF007FC,$FF837FC0,$E1FFF0FF,$FF8733CC,$FFFFFF0F,
$0000E0FF,$FFFFFF07,$C003FFFF,$E1010000,$FFFFFFFF,$C00100FE,$FFFF7F00,
$3FFFFFFF,$FFF9FFFF,$83FFFFFF,$1FF00FFC,$FF837FE0,$031FF8FF,$F8C0318C,
$FFFFFF1F,$0000F0FF,$FFFFFF0F,$E001FFFF,$E1010000,$FFFFFFFF,$C00100FE,
$FFFF7F00,$7FFEFFFF,$FFFCFFFF,$87FFFFFF,$1FF00FFC,$FFC37FE0,$0F00FEFF,
$00F0399C,$FFFFFF7F,$0000F8FF,$FFFFFF1F,$F000FFFF,$C3030000,$FFFFFFFF,
$C00100FF,$FFFFFF00,$00FCFFFF,$7F000000,$87FFFFFF,$1FF00FFC,$FFC37FE0,
$7FC0FFFF,$03FE1998,$FFFFFFFF,$0000FCFF,$FFFFFF3F,$3080FFFF,$87070000,
$FFFFFFFF,$C00180FF,$FFFFFF01,$00F8FFFF,$3F000000,$0FFFFFFF,$3FF81FFC,
$FFE17FF0,$FFFFFFFF,$FFFF399C,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$0080FFFF,
$0F0F0000,$FFFFFFFF,$0000C0FF,$FFFFFF03,$00F0FFFF,$1F000000,$1FFCFFFF,
$3FF81FFC,$7FF07FF0,$FFFFFFFF,$FFFF718E,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,
$00C0FFFF,$1F1E0000,$FFFFFFFF,$0000E0FF,$FFFFFF07,$00E0FFFF,$0F000000,
$0FF8FFFF,$3FF81FFC,$3FE07FF0,$FFFFFFFF,$FFFFF3CF,$FFFFFFFF,$03C0FFFF,
$FFFFFFFF,$00C0FFFF,$3F3C0000,$FFFFFFFF,$0000F8FF,$FFFFFF1F,$FFCFFFFF,
$E7FFFFFF,$07F0FFFF,$3FF83FFC,$1FC07FF8,$FFFFFFFF,$FFFFE3C7,$FFFFFFFF,
$0180FFFF,$FFFFFFFF,$C0E1FFFF,$7FF80000,$FFFFFFFF,$0000FEFF,$FFFFFF7F,
$FFCFFFFF,$E7FFFFFF,$07F0FFFF,$3FF83FFC,$1FC07FF8,$FFFFFFFF,$FFFF07E0,
$FFFFFFFF,$0180FFFF,$FFFFFFFF,$F8E1FFFF,$FFE00700,$FFFFFFFF,$0000FFFF,
$FFFFFFFF,$FFCFFFFF,$E7FFFFFF,$07F0FFFF,$7FFC3FFC,$1FC07FF8,$FFFFFFFF,
$FFFF0FF0,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$3EF0FFFF,$FF837F00,$FFFFFFFF,
$03C0FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$07F0FFFF,$7FFC7FFC,$1FC07FFC,
$FFFFFFFF,$FFFF3FFC,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$07F8FFFF,$FF077C00,
$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$07F0FFFF,$7FFC7FFC,
$1FC07FFC,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$00F8FFFF,
$FF1F0000,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$0FF8FFFF,
$7FFC3FF8,$3FE03FF8,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,
$00F8FFFF,$FFFF0000,$FFFFFFFF,$3FFCFFFF,$FFFFFFFF,$00C0FFFF,$07000000,
$1FFCFFFF,$7FFC1FF0,$7FF01FF0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0000FFFF,
$FFFFFFFF,$07F8FFFF,$FFFF7F00,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$00C0FFFF,
$07000000,$FFFFFFFF,$3FF80FE0,$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,
$0000FFFF,$FFFFFFFF,$0FF0FFFF,$FFFFFF00,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,
$0FC0FFFF,$07E007C0,$FFFFFFFF,$1FF00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,
$FFFFFFFF,$0180FFFF,$FFFFFFFF,$1FF0FFFF,$FFFFFF80,$FFFFFFFF,$07E0FFFF,
$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE00FE0,$FFFF0FE0,$FFFFFFFF,
$FFFF07E0,$FFFFFFFF,$0180FFFF,$FFFFFFFF,$3FF0FFFF,$FFFFFF81,$FFFFFFFF,
$03C0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE00FE0,$FFFF0FE0,
$FFFFFFFF,$FFFF07E0,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$7FF0FFFF,$FFFFFFC1,
$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE00FE0,
$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$FFF8FFFF,
$FFFFFFC3,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,
$0FE01FF0,$FFFF1FF0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$1FF8FFFF,$FFFFFFFF,
$FFF9FFFF,$FFFFFFC3,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,
$FFFFFFFF,$0FE03FF8,$FFFF3FF8,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFE7,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$1FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFE7,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FF8FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FFCFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,$00000000,$DFFF7F00,$FFFFFFFF,$FFFBFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,$FFFFFF3F,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,$00000000,$87FF7F00,$FC3FC0FF,
$FFE1FF03,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$000000FF,$FFFFFF01,$FFFFFFFF,
$FFFF03C0,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,$00000000,$80FF7F00,
$F00F0070,$FF010E00,$0000F8FF,$00000000,$FFFFFF3F,$000000F0,$FFFF1F00,
$F0FFFFFF,$FF0F0000,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,$00000000,
$00FF7F00,$C0030000,$FF000000,$0000F8FF,$00000000,$FFFFFF3F,$000000C0,
$FFFF0700,$00FFFFFF,$FF000000,$FFFFFFFF,$00000000,$FFFF0000,$C0FFFFFF,
$00000000,$00FF7F00,$80010000,$FF000000,$0000F8FF,$00000000,$FFFFFF3F,
$00000000,$FFFF0100,$00F8FFFF,$1F000000,$FFFFFFFF,$00000000,$FFFF0000,
$C0FFFFFF,$00000000,$00FE7F00,$80010000,$7F000000,$0000F8FF,$00000000,
$FFFFFF3F,$00000000,$FFFF0100,$00E0FFFF,$07000000,$FFFFFFFF,$00000000,
$FFFF0000,$C0FFFFFF,$00000000,$00FE7F00,$00000000,$7F000000,$0000F8FF,
$00000000,$FFFFFF3F,$00000080,$FFFF0300,$00C0FFFF,$03000000,$FFFFFFFF,
$00000000,$FFFF0000,$E0FFFFFF,$00000000,$00FC7F00,$00000000,$3F000000,
$0000F8FF,$00000000,$FFFFFF3F,$00000080,$FFFF0300,$0080FFFF,$01000000,
$FFFFFFFF,$00000000,$FFFF0000,$E0FFFFFF,$00000000,$00FE7F00,$00000000,
$7F000000,$0000F8FF,$00000000,$FFFFFF3F,$000000C0,$FFFF0700,$00C0FFFF,
$03000000,$FFFFFFFF,$00000000,$FFFF0000,$E0FFFFFF,$00000000,$00FF7F00,
$00000000,$FF000000,$0000F8FF,$00000000,$FFFFFF3F,$000000C0,$FFFF0700,
$00C0FFFF,$03000000,$FFFFFFFF,$00000000,$FFFF0000,$F0FFFFFF,$00000000,
$C0FF7F00,$0000FC00,$FF03003F,$00C0FFFF,$07000000,$FFFFFFFF,$000000E0,
$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,$00000000,$FFFF0000,$F0FFFFFF,
$00000000,$F8FF7F00,$0000FF1F,$FF1FF8FF,$00C0FFFF,$07000000,$FFFFFFFF,
$000000E0,$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,$00000080,$FFFF0100,
$F8FFFFFF,$00000000,$FFFF7F00,$0180FFFF,$FFFFFFFF,$00C0FFFF,$07000000,
$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,$00000080,
$FFFF0100,$FC9FFFFF,$00000000,$FFFFFF00,$0180FFFF,$FFFFFFFF,$00C0FFFF,
$07000000,$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,$07000000,$FFFFFFFF,
$00000080,$FFFF0100,$FE07FCFF,$00000000,$FFFFFF00,$0000F0FF,$FFFFFF0F,
$00C0FFFF,$07000000,$FFFFFFFF,$000000E0,$FFFF0F00,$00E0FFFF,$07000000,
$FFFFFFFF,$000000C0,$FFFF0300,$FF03F0FF,$00000000,$FFFFFF00,$0000C0FF,
$FFFFFF03,$00C0FFFF,$07000000,$FFFFFFFF,$000000C0,$FFFF0700,$00E0FFFF,
$07000000,$FFFFFFFF,$000000C0,$FFFF0300,$FF01E0FF,$00000080,$FFFFFF00,
$000080FF,$FFFFFF01,$00C0FFFF,$07000000,$FFFFFFFF,$000000C0,$FFFF0700,
$00C0FFFF,$03000000,$FFFFFFFF,$000000E0,$FFFF0700,$FF00C0FF,$000000C0,
$FFFFFF00,$000080FF,$FFFFFF01,$00E0FFFF,$0F000000,$FFFFFFFF,$00000080,
$FFFF0300,$00C0FFFF,$03000000,$FFFFFFFF,$000000E0,$FFFF0700,$7F0080FF,
$000000E0,$FFFFFF00,$000080FF,$FFFFFF01,$00F8FFFF,$3F000000,$FFFFFFFF,
$00000000,$FFFF0100,$00C0FFFF,$03000000,$FFFFFFFF,$000000F0,$FFFF0F00,
$7F0000FF,$000000F0,$FFFFFF00,$0000C0FF,$FFFFFF03,$00FCFFFF,$7F000000,
$FEFFFFFF,$00000000,$FFFF0000,$0080FFFF,$01000000,$FFFFFFFF,$000000F0,
$FFFF0F00,$3F0000FF,$000000F8,$FFFFFF00,$0000C0FF,$FFFFFF03,$00FEFFFF,
$FF000000,$FEFFFFFF,$00000000,$FFFF0000,$0000FFFF,$00000000,$FFFFFFFF,
$000000F8,$FFFF1F00,$1F0000FE,$000000FC,$FFFFFF00,$0000C0FF,$FFFFFF03,
$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,$FF7F0000,$0000FEFF,$00000000,
$FFFFFF7F,$000000FC,$FFFF3F00,$0F0000FE,$000000FC,$FFFFFF01,$0000C0FF,
$FFFFFF03,$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,$FF7F0000,$0000FCFF,
$00000000,$FFFFFF3F,$000000FE,$FFFF7F00,$070000FE,$000000FE,$FFFFFF01,
$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,$FF7F0000,
$0000F8FF,$00000000,$FFFFFF1F,$000000FF,$FFFFFF00,$030000FE,$000000FE,
$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,$FCFFFFFF,$00000000,
$FF7F0000,$0000F0FF,$00000000,$FFFFFF0F,$0000C0FF,$FFFFFF03,$010000FE,
$000000FF,$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,$F8FFFFFF,
$00000000,$FF3F0000,$0000E0FF,$00000000,$FFFFFF07,$0000F0FF,$FFFFFF0F,
$000000FE,$0000001F,$FFFFFF01,$0000E0FF,$FFFFFF07,$00FFFFFF,$FF010000,
$F8FFFFFF,$00000000,$FF3F0000,$0000C0FF,$00000000,$FFFFFF03,$0000FCFF,
$FFFFFF3F,$000000FE,$00008007,$FFFFFF03,$0000C0FF,$FFFFFF03,$00FFFFFF,
$FF010000,$F8FFFFFF,$00000000,$FF3F0000,$0000C0FF,$00000000,$FFFFFF03,
$0000FFFF,$FFFFFFFF,$000000FF,$00008003,$FFFFFF03,$000080FF,$FFFFFF01,
$00FFFFFF,$FF010000,$F8FFFFFF,$00000000,$FF3F0000,$000080FF,$00000000,
$FFFFFF01,$0000FCFF,$FFFFFF3F,$000000FF,$00008001,$FFFFFF03,$000000FF,
$FFFFFF00,$00FFFFFF,$FF010000,$F0FFFFFF,$00000000,$FF1F0000,$000080FF,
$00000000,$FFFFFF01,$0000F8FF,$FFFFFF1F,$000080FF,$00008000,$FFFFFF03,
$000000FF,$FFFFFF00,$00FFFFFF,$FF010000,$F0FFFFFF,$00000004,$FF1F4000,
$000080FF,$00000000,$FFFFFF01,$0000F0FF,$FFFFFF0F,$000080FF,$00000000,
$FFFFFF07,$000000FE,$FFFF7F00,$00FFFFFF,$FF010000,$F0FFFFFF,$00000004,
$FF1F4000,$000000FF,$00000000,$FFFFFF00,$0000E0FF,$FFFFFF07,$0000C0FF,
$00000000,$FFFFFF07,$000000FE,$FFFF7F00,$00FFFFFF,$FF010000,$F0FFFFFF,
$0000000C,$FF1F6000,$000000FF,$00000000,$FFFFFF00,$0000E0FF,$FFFFFF07,
$0000E0FF,$00000000,$FFFFFF07,$000000FC,$FFFF3F00,$00FFFFFF,$FF010000,
$F0FFFFFF,$06C0000C,$FF1F6000,$000000FF,$00000000,$FFFFFF00,$0000C0FF,
$FFFFFF03,$0000E0FF,$00000000,$FFFFFF0F,$000000FC,$FFFF3F00,$00FFFFFF,
$FF010000,$E0FFFFFF,$06C0001C,$FF0F7000,$000000FF,$00000000,$FFFFFF00,
$0000C0FF,$FFFFFF03,$0000F0FF,$00000000,$FFFFFF0F,$000000FC,$FFFF3F00,
$00FFFFFF,$FF010000,$E0FFFFFF,$06C0001C,$FF0F7000,$000000FF,$00000000,
$FFFFFF00,$000080FF,$FFFFFF01,$0000F0FF,$00000000,$FFFFFF0F,$000000F8,
$FFFF1F00,$00FFFFFF,$FF010000,$E0FFFFFF,$07C0013C,$FF0F7800,$000000FF,
$00000000,$FFFFFF00,$000080FF,$FFFFFF01,$0000F8FF,$00000000,$FFFFFF1F,
$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,$E0FFFFFF,$07C0013C,$FF0F7800,
$000000FF,$00000000,$FFFFFF00,$000080FF,$FFFFFF01,$0000FCFF,$00000000,
$FFFFFF1F,$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,$E0FFFFFF,$0FE0017C,
$FF0F7C00,$000000FF,$00000000,$FFFFFF00,$000080FF,$FFFFFF01,$0000FCFF,
$00000000,$FFFFFF3F,$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,$C0FFFFFF,
$0FE0037C,$FF077C80,$000080FF,$00000000,$FFFFFF01,$000080FF,$FFFFFF01,
$0000FEFF,$00000000,$FFFFFF3F,$000000F8,$FFFF1F00,$00FFFFFF,$FF010000,
$C0FFFFFF,$0FE003FC,$FF077E80,$000080FF,$00000000,$FFFFFF01,$000080FF,
$FFFFFF01,$0000FEFF,$00000000,$FFFFFF7F,$000000F8,$FFFF1F00,$00FFFFFF,
$FF010000,$C0FFFFFF,$0FE003FC,$FF077E80,$000080FF,$00000420,$FFFFFF01,
$0000C0FF,$FFFFFF03,$0000FFFF,$00000000,$FFFFFF7F,$000000FC,$FFFF3F00,
$00FFFFFF,$FF010000,$C1FFFFFF,$0FE007FC,$FF077FC0,$0000C0FF,$00000660,
$FFFFFF03,$0000C0FF,$FFFFFF03,$0000FFFF,$00000000,$FFFFFFFF,$000000FC,
$FFFF3F00,$00FFFFFF,$FF010000,$C1FFFFFF,$1FF007FC,$FF077FC0,$0000E0FF,
$000003C0,$FFFFFF07,$0000E0FF,$FFFFFF07,$0000FFFF,$00000000,$FFFFFFFF,
$000000FC,$FFFF3F00,$00FFFFFF,$FF010000,$83FFFFFF,$1FF007FC,$FF837FC0,
$0100F0FF,$008003C0,$FFFFFF0F,$0000E0FF,$FFFFFF07,$0000FFFF,$01000000,
$FFFFFFFF,$000000FE,$FFFF7F00,$00FFFFFF,$FF010000,$83FFFFFF,$1FF00FFC,
$FF837FE0,$0300F8FF,$00C00180,$FFFFFF1F,$0000F0FF,$FFFFFF0F,$0000FFFF,
$01000000,$FFFFFFFF,$000000FE,$FFFF7F00,$00FEFFFF,$FF000000,$87FFFFFF,
$1FF00FFC,$FFC37FE0,$0F00FEFF,$00F00180,$FFFFFF7F,$0000F8FF,$FFFFFF1F,
$0000FFFF,$03000000,$FFFFFFFF,$000000FF,$FFFFFF00,$00FCFFFF,$7F000000,
$87FFFFFF,$1FF00FFC,$FFC37FE0,$7FC0FFFF,$03FE0180,$FFFFFFFF,$0000FCFF,
$FFFFFF3F,$0080FFFF,$07000000,$FFFFFFFF,$000080FF,$FFFFFF01,$00F8FFFF,
$3F000000,$0FFFFFFF,$3FF81FFC,$FFE17FF0,$FFFFFFFF,$FFFF0180,$FFFFFFFF,
$0000FFFF,$FFFFFFFF,$0080FFFF,$0F000000,$FFFFFFFF,$0000C0FF,$FFFFFF03,
$00F0FFFF,$1F000000,$1FFCFFFF,$3FF81FFC,$7FF07FF0,$FFFFFFFF,$FFFF0180,
$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$00C0FFFF,$1F000000,$FFFFFFFF,$0000E0FF,
$FFFFFF07,$00E0FFFF,$0F000000,$0FF8FFFF,$3FF81FFC,$3FE07FF0,$FFFFFFFF,
$FFFF03C0,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$00C0FFFF,$3F000000,$FFFFFFFF,
$0000F8FF,$FFFFFF1F,$00C0FFFF,$07000000,$07F0FFFF,$3FF83FFC,$1FC07FF8,
$FFFFFFFF,$FFFF03C0,$FFFFFFFF,$0180FFFF,$FFFFFFFF,$00E0FFFF,$7F000000,
$FFFFFFFF,$0000FEFF,$FFFFFF7F,$00C0FFFF,$07000000,$07F0FFFF,$3FF83FFC,
$1FC07FF8,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,$0180FFFF,$FFFFFFFF,$00E0FFFF,
$FF000000,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$07F0FFFF,
$7FFC3FFC,$1FC07FF8,$FFFFFFFF,$FFFF0FF0,$FFFFFFFF,$0000FFFF,$FFFFFFFF,
$00F0FFFF,$FF030000,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$00C0FFFF,$07000000,
$07F0FFFF,$7FFC7FFC,$1FC07FFC,$FFFFFFFF,$FFFF3FFC,$FFFFFFFF,$0000FFFF,
$FFFFFFFF,$00F8FFFF,$FF070000,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$00C0FFFF,
$07000000,$07F0FFFF,$7FFC7FFC,$1FC07FFC,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,
$0000FFFF,$FFFFFFFF,$00F8FFFF,$FF1F0000,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,
$00C0FFFF,$07000000,$0FF8FFFF,$7FFC3FF8,$3FE03FF8,$FFFFFFFF,$FFFF7FFE,
$FFFFFFFF,$0000FFFF,$FFFFFFFF,$00F8FFFF,$FFFF0000,$FFFFFFFF,$3FFCFFFF,
$FFFFFFFF,$00C0FFFF,$07000000,$1FFCFFFF,$7FFC1FF0,$7FF01FF0,$FFFFFFFF,
$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$07F8FFFF,$FFFF7F00,$FFFFFFFF,
$0FF0FFFF,$FFFFFFFF,$00C0FFFF,$07000000,$FFFFFFFF,$3FF80FE0,$FFFF0FE0,
$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0000FFFF,$FFFFFFFF,$0FF0FFFF,$FFFFFF00,
$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$1FF00FE0,
$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$0180FFFF,$FFFFFFFF,$1FF0FFFF,
$FFFFFF80,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,
$0FE00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,$0180FFFF,$FFFFFFFF,
$3FF0FFFF,$FFFFFF81,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,
$FFFFFFFF,$0FE00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,$03C0FFFF,
$FFFFFFFF,$7FF0FFFF,$FFFFFFC1,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,$0FC0FFFF,
$07E007C0,$FFFFFFFF,$0FE00FE0,$FFFF0FE0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,
$07E0FFFF,$FFFFFFFF,$FFF8FFFF,$FFFFFFC3,$FFFFFFFF,$03C0FFFF,$FFFFFFFF,
$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE01FF0,$FFFF1FF0,$FFFFFFFF,$FFFF7FFE,
$FFFFFFFF,$1FF8FFFF,$FFFFFFFF,$FFF9FFFF,$FFFFFFC3,$FFFFFFFF,$07E0FFFF,
$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$0FE03FF8,$FFFF3FF8,$FFFFFFFF,
$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFE7,$FFFFFFFF,
$07E0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$1FF0FFFF,$FFFFFFFF,
$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFE7,
$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FF8FFFF,
$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$3FFCFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,$00000000,
$DFFF7F00,$FFFFFFFF,$FFFBFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000F8FF,
$FFFFFF3F,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$C0FFFFFF,
$00000000,$87FF7F00,$FC3FC0FF,$FFE1FF03,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$000000FF,$FFFFFF01,$FFFFFFFF,$FFFF03C0,$FFFFFFFF,$00000000,$FFFF0000,
$CFFFFFFF,$FFFFFFFF,$80FF7FF8,$F00F0070,$FF010E00,$0000F8FF,$00000000,
$FFFFFF3F,$FFFF03F0,$FFFF1F80,$F0FFFFFF,$FF0F0000,$FFFFFFFF,$00000000,
$FFFF0000,$CFFFFFFF,$FFFFFFFF,$30FF7FF8,$C1831F00,$FF0C00F8,$0000F8FF,
$00000000,$FFFFFF3F,$FFFF7FC0,$FFFF07FC,$00FFFFFF,$FF00F81F,$FFFFFFFF,
$FFFFFF3F,$FFFFFCFF,$CFFFFFFF,$FFFFFFFF,$3FFF7FF8,$87E1FF07,$FFFCE0FF,
$FFFFF9FF,$FFFFFFFF,$FFFFFF3F,$FFFFFF03,$FFFF81FF,$07F8FFFF,$1FE0FFFF,
$FFFFFFFF,$FFFFFF3F,$FFFFFCFF,$CFFFFFFF,$FFFFFFFF,$7FFE7FF8,$8FF1FFFF,
$7FFEFFFF,$FFFFF9FF,$FFFFFFFF,$FFFFFF3F,$FFFFFF03,$FFFF81FF,$7FE0FFFF,
$07FEFFFF,$FFFFFFFF,$FFFFFF3F,$FFFFFCFF,$CFFFFFFF,$FFFFFFFF,$7FFE7FF8,
$3FFCFFFF,$7FFEFFFF,$FFFFF9FF,$FFFFFFFF,$FFFFFF3F,$FFFF7F80,$FFFF03FC,
$FFC0FFFF,$03FFFFFF,$FFFFFFFF,$FFFFFF3F,$FFFFFCFF,$E7FFFFFF,$FFFFFFFF,
$7FFC7FF8,$7FFEFFFF,$3FFEFFFF,$FFFFF9FF,$FFFFFFFF,$FFFFFF3F,$FFFF0198,
$FFFF3300,$1F80FFFF,$01F8FFFF,$FFFFFFFF,$FFFFFF3F,$FFFFFCFF,$E7FFFFFF,
$FFFFFFFF,$1FFE7FF8,$7FFE01FE,$7FF87F80,$0000F8FF,$00000000,$FFFFFF3F,
$000000CF,$FFFFE701,$00CEFFFF,$7300FC3F,$FFFFFFFF,$FFFFFF3F,$FFFFFCFF,
$E7FFFFFF,$FFFFFFFF,$03FF7FF8,$7E7E00C0,$FFC00300,$0000F8FF,$00000000,
$FFFFFF3F,$0000FCCF,$FFFFE77F,$E0C7FFFF,$E3070000,$FFFFFFFF,$FFFFFF3F,
$FFFFFCFF,$F3FFFFFF,$FFFFFFFF,$C0FF7FF8,$7C3EFC00,$FF03003F,$FFCFFFFF,
$E7FFFFFF,$FFFFFFFF,$FFFFFFE7,$FFFFCFFF,$FFE7FFFF,$E7FF03C0,$FFFFFFFF,
$FFFFFF1F,$FFFFF8FF,$F1FFFFFF,$FFFFFFFF,$F8FF7FF8,$781EFF1F,$FF1FF8FF,
$FFCFFFFF,$E7FFFFFF,$FFFFFFFF,$FFFFFFE7,$FFFFCFFF,$FFE1FFFF,$87FFFFFF,
$FFFFFFFF,$FFFFFF9F,$FFFFF9FF,$F8FFFFFF,$FFFFFFFF,$FFFF7FF8,$799EFFFF,
$FFFFFFFF,$FFCFFFFF,$E7FFFFFF,$FFFFFFFF,$FFFFFFE1,$FFFF0FFF,$1FE0FFFF,
$07F8FFFF,$FFFFFFFF,$FFFFFF9F,$FFFFF9FF,$FC9FFFFF,$FFFFFF7F,$FFFFFFF0,
$0180FFFF,$FFFFFFFF,$FFCFFFFF,$E7FFFFFF,$FFFFFFFF,$FFFF1FE0,$FFFF0FF0,
$00E6FFFF,$6700FE7F,$FFFFFFFF,$FFFFFF8F,$FFFFF1FF,$FE07FCFF,$FFFFFF3F,
$FFFFFFF0,$0000F0FF,$FFFFFF0F,$FFCFFFFF,$E7FFFFFF,$FFFFFFFF,$FC7F00E4,
$FFFF4F00,$E0E7FFFF,$E7070000,$FFFFFFFF,$FFFFFFCF,$FFFFF3FF,$FF03F0FF,
$FFFFFF1F,$FFFFFFF0,$FC3FC0FF,$FFFFFF03,$00C0FFFF,$07000000,$FFFFFFFF,
$0000C0CF,$FFFFE707,$FFE7FFFF,$E7FF0180,$FFFFFFFF,$FFFFFFC7,$FFFFE3FF,
$FF31E3FF,$FFFFFF8F,$FFFFFFF0,$FFFF8FFF,$FFFFFFF1,$00C0FFFF,$07000000,
$FFFFFFFF,$0100FFCF,$FFFFE7FF,$FFC7FFFF,$E3FFFFFF,$FFFFFFFF,$FFFFFFE7,
$FFFFE7FF,$FF38CFFF,$FFFFFFC7,$FFFFFFF0,$FFFF83FF,$FFFFFFC1,$FFE3FFFF,
$8FFFFFFF,$FFFFFFFF,$FFFFFF8F,$FFFFE3FF,$FFCFFFFF,$F3FFFFFF,$FFFFFFFF,
$FFFFFFE3,$FFFFC7FF,$7F1C9FFF,$FFFFFFE3,$FFFFFFF0,$F00F80FF,$FFFFFF01,
$FFF8FFFF,$3FFEFFFF,$FFFFFFFF,$FFFFFF07,$FFFFC1FF,$FFC7FFFF,$E3FFFFFF,
$FFFFFFFF,$FFFFFFF1,$FFFF8FFF,$7F8E3FFF,$FFFFFFF1,$FFFFFFF0,$0000C8FF,
$FFFFFF13,$7FFCFFFF,$7FFCFFFF,$FEFFFFFF,$FFFF7F20,$FFFF08FC,$FF81FFFF,
$81FFFFFF,$FFFFFFFF,$FFFFFFF1,$FFFF8FFF,$3FCE3FFF,$FFFFFFF8,$FFFFFFF0,
$07E0CFFF,$FFFFFFF3,$00FEFFFF,$FF000000,$FEFFFFFF,$FEFF0070,$FFFF1C00,
$1F18FFFF,$18F8FFFF,$FFFFFFFF,$FFFFFFF8,$FFFF1FFF,$1FDF7FFE,$FFFFFFFC,
$FFFFFFF0,$FFFFCFFF,$FFFFFFF3,$00FFFFFF,$FF010000,$FCFFFFFF,$000000FF,
$FF7FFE01,$003EFEFF,$7C00FE7F,$FFFFFF7F,$FFFF7FFC,$FFFF3FFE,$8FFF67FE,
$FFFFFFFC,$FFFFFFE1,$FFFFCFFF,$FFFFFFF3,$3FFFFFFF,$FFF9FFFF,$FCFFFFFF,
$0000FEFF,$FF7FFEFF,$E07FFCFF,$FE070000,$FFFFFF3F,$FFFF1FFE,$FFFF7FF8,
$C7FF43FE,$FFFF7FFE,$FFFFFFE1,$FFFFE7FF,$FFFFFFE7,$3FFFFFFF,$FFF9FFFF,
$FCFFFFFF,$FFFFFFFF,$FF7FFEFF,$FFFFF8FF,$FFFF0180,$FFFFFF1F,$FFFF07FF,
$FFFFFFE0,$E3FF63FE,$FFFF7FFE,$FFFFFFE1,$FFFFE3FF,$FFFFFFC7,$3FFFFFFF,
$FFF9FFFF,$FCFFFFFF,$FFFFFFCF,$FF7FE6FF,$FFFFF1FF,$FFFF7FFE,$FFFFFF8F,
$FFFFC1FF,$FFFFFF83,$F1FF61FE,$FFFF3FFF,$FFFFFFE1,$F81FE0FF,$FFFFFF07,
$3FFFFFFF,$FFF9FFFF,$F8FFFFFF,$FFFFFF07,$FF3FC0FF,$FFFFE3FF,$FFFF7FFE,
$FFFFFFC7,$FE7FF0FF,$FFFFFF0F,$F8FF71FE,$FFFF3F1F,$FFFFFFE1,$0000E4FF,
$FFFFFF27,$3FFFFFFF,$FFF9FFFF,$F8FFFFFF,$FFFFFF30,$FF3F18FE,$FFFFC7FF,
$FFFF7FFE,$FFFFFFE3,$FC3FFCFF,$FFFFFF3F,$FEFF79FE,$FFFF9F07,$FFFFFFC3,
$03C0C7FF,$FFFFFFE3,$3FFFFFFF,$FFF9FFFF,$F9FFFFFF,$F01F7CF8,$FF3F3F7C,
$FFFFCFFF,$FFFF7FFE,$FFFFFFF3,$F81FFFFF,$FFFFFFFF,$FFFF3FFF,$FFFF9FC3,
$FFFFFFC3,$FFFF8FFF,$FFFFFFF1,$3FFFFFFF,$FFF9FFFF,$F9FFFFFF,$C00720F0,
$FF3F1F08,$FFFF8FFF,$FFFF7FFE,$FFFFFFF1,$FC3FFCFF,$FFFFFF3F,$FFFF3FFF,
$FFFF9FF1,$FFFFFFC3,$FFFF1FFF,$FFFFFFF8,$3FFFFFFF,$FFF9FFFF,$F3FFFFFF,
$010183F0,$FF9F1F82,$FFFF9FFF,$FFFF3FFC,$FFFFFFF9,$FFFFF8FF,$FFFFFF1F,
$FFFF9FFF,$FFFF9FF8,$FFFFFFC3,$FFFF3FFF,$FFFFFFFC,$3FFFFFFF,$FFF9FFFF,
$F3FFFFFF,$3018DEE4,$FF9F4FF6,$FFFF9FFF,$FFFF3FFC,$FFFFFFF9,$FFFFF1FF,
$FFFFFF8F,$FFFF9FFF,$FFFF1FFC,$FFFFFF87,$FFFF7FFE,$FFFF7FFE,$3FFFFFFF,
$FFF9FFFF,$F3FFFFFF,$F01EFEE4,$FF9F4FFE,$FFFF1FFF,$FFFF3FFC,$FFFFFFF8,
$FFFFE3FF,$FFFFFFC7,$FFFFCFFF,$FFFF1FFE,$FFFFFF87,$FFFF7FFE,$FFFF7FFE,
$3FFFFFFF,$FFF9FFFF,$F3FFFFFF,$F01FFECC,$FF9F67FE,$FFFF3FFF,$FFFF3FFC,
$FFFFFFFC,$FFFFE7FF,$FFFFFFE7,$FFFFE7FF,$FFFF1FFF,$FFFFFF87,$FFFFFFFC,
$FFFF3FFF,$3FFFFFFF,$FFF9FFFF,$F3FFFFFF,$E6CFFCCC,$FF9F677E,$FFFF3FFF,
$FFFF1FF8,$FFFFFFFC,$FFFFC7FF,$FFFFFFE3,$FFFFE7FF,$FFFF1FFF,$FFFFFF0F,
$7FFEFFFC,$FFFF3FFF,$3FFFFFFF,$FFF9FFFF,$E7FFFFFF,$E6CFFC9C,$FFCF737E,
$FFFF3FFF,$FFFF9FF9,$FFFFFFFC,$FFFFCFFF,$FFFFFFF3,$FFFFF3FF,$FFFF9FFF,
$FFFFFF0F,$7FFEFFFC,$FFFF3FFF,$3FFFFFFF,$FFF9FFFF,$E7FFFFFF,$E6CFFC9C,
$FFCF737E,$FFFF3FFF,$FFFF8FF1,$FFFFFFFC,$FFFF8FFF,$FFFFFFF1,$FFFFF3FF,
$FFFFCFFF,$FFFFFF0F,$7FFEFFF9,$FFFF9FFF,$3FFFFFFF,$FFF9FFFF,$E7FFFFFF,
$E7CFF93C,$FFCF793E,$FFFF3FFF,$FFFF8FF1,$FFFFFFFC,$FFFF9FFF,$FFFFFFF9,
$FFFFF9FF,$FEFFCFFF,$FFFFFF1F,$7FFEFFF9,$FFFF9FFF,$3FFFFFFF,$FFF9FFFF,
$E7FFFFFF,$E7CFF93C,$FFCF793E,$FFFF3FFF,$FFFFC7E3,$FFFFFFFC,$FFFF9FFF,
$FFFFFFF9,$FFFFFCFF,$FEFFCFFF,$FFFFFF1F,$7FFEFFF9,$FFFF9FFF,$3FFFFFFF,
$FFF9FFFF,$E6FFFFFF,$CFE7F97C,$FFCF7C3E,$FFFF1FFF,$FFFFC3C3,$FFFFFFF8,
$FFFF9FFF,$FFFFFFF9,$FFFFFCFF,$FCFFCFFF,$FFFFFF3F,$7FFEFFF9,$FFFF9FFF,
$3FFFFFFF,$FFF9FFFF,$CEFFFFFF,$CFE7F37C,$FFE77C9E,$FFFF9FFF,$FFFFC183,
$FFFFFFF9,$FFFF9FFF,$FFFFFFF9,$FF7FFEFF,$FCFFCFFF,$FFFFFF3F,$0180FFF9,
$FFFF9FFF,$3FFFFFFF,$FFF9FFFF,$CCFFFFFF,$CFE7F3FC,$FF677E9E,$FFFF9FFF,
$FFFFE007,$FFFFFFF9,$FFFF8FFF,$FFFFFFF1,$FF77FEFF,$F8FFCFFF,$FFFFFF7F,
$0180FFF9,$FFFF9FFF,$3FFFFFFF,$FFF9FFFF,$CCFFFFFF,$CFE7F3FC,$FF677E9E,
$FEFF8FFF,$FF7FE427,$FFFFFFF1,$FFFFCFFF,$FFFFFFF3,$3F33FFFF,$F8FFFFFF,
$FFFFFF7F,$7FFEFFFC,$FFFF3FFF,$3FFFFFFF,$FFF9FFFF,$C9FFFFFF,$CFE7E7FC,
$FF277FCE,$FCFFC7FF,$FF3FE667,$FFFFFFE3,$FFFFC7FF,$FFFFFFE3,$3F31FFFF,
$F0FFFFFF,$FFFFFFFF,$7FFEFFFC,$FFFF3FFF,$3FFFFFFF,$FFF9FFFF,$C9FFFFFF,
$9FF3E7FC,$FF277FCE,$F8FFE3FF,$FF1FF3CF,$FFFFFFC7,$FFFFE7FF,$FFFFFFE7,
$1F38FFFF,$F0FFFFFF,$FFFFFFFF,$7FFEFFFC,$FFFF3FFF,$3FFFFFFF,$FFF9FFFF,
$93FFFFFF,$9FF3E7FC,$FF937FCE,$E1FFF0FF,$FF87F3CF,$FFFFFF0F,$FFFFE3FF,
$FFFFFFC7,$3F3CFFFF,$E1FFFFFF,$FFFFFFFF,$7FFE7FFE,$FFFF7FFE,$00FFFFFF,
$FF010000,$93FFFFFF,$9FF3CFFC,$FF937FE6,$031FF8FF,$F8C0F18F,$FFFFFF1F,
$FFFFF1FF,$FFFFFF8F,$1F3EFFFF,$E1FFFFFF,$FFFFFFFF,$7FFE3FFE,$FFFF7FFC,
$00FEFFFF,$FF000000,$87FFFFFF,$9FF3CFFC,$FFC37FE6,$0F00FEFF,$00F0F99F,
$FFFFFF7F,$FE7FF8FF,$FFFFFF1F,$0F1FFFFF,$C3FFFFFF,$FFFFFFFF,$7FFE1FFF,
$FFFFFFF8,$7FFCFFFF,$7FFCFFFF,$87FFFFFF,$9FF3CFFC,$FFC37FE6,$7FC0FFFF,
$03FEF99F,$FFFFFFFF,$F81FFCFF,$FFFFFF3F,$CF9FFFFF,$87FFFFFF,$FFFFFFFF,
$FFFF8FFF,$FFFFFFF1,$FFF8FFFF,$3FFEFFFF,$0FFFFFFF,$3FF99FFC,$FFE17FF2,
$FFFFFFFF,$FFFFF99F,$FFFFFFFF,$E007FFFF,$FFFFFFFF,$FF8FFFFF,$0FFFFFFF,
$FFFFFFFF,$FFFFC3FF,$FFFFFFC3,$FFF1FFFF,$1FFFFFFF,$1FFCFFFF,$3FF99FFC,
$7FF07FF2,$FFFFFFFF,$FFFFF18F,$FFFFFFFF,$E7E7FFFF,$FFFFFFFF,$3FCFFFFF,
$1FFEFFFF,$FFFFFFFF,$FFFFE0FF,$FFFFFF07,$FFE3FFFF,$8FFFFFFF,$0FF8FFFF,
$3FF99FFC,$3FE07FF2,$FFFFFFFF,$FFFFF3CF,$FFFFFFFF,$E3C7FFFF,$FFFFFFFF,
$3FCFFFFF,$3FFCFFFF,$FFFFFFFF,$FE7FF8FF,$FFFFFF1F,$00C0FFFF,$07000000,
$C7F1FFFF,$3FF93FFC,$1FC77FF8,$FFFFFFFF,$FFFFE3C7,$FFFFFFFF,$F18FFFFF,
$FFFFFFFF,$1FE7FFFF,$7FF8FFFC,$FFFFFFFF,$F81FFEFF,$FFFFFF7F,$00C0FFFF,
$07000000,$E7F3FFFF,$3FF93FFC,$9FCF7FF8,$FFFFFFFF,$FFFF07E0,$FFFFFFFF,
$F99FFFFF,$FFFFFFFF,$9FE7FFFF,$FFE0FF3C,$FFFFFFFF,$F00FFFFF,$FFFFFFFF,
$FFCFFFFF,$E7FFFFFF,$E7F3FFFF,$7FFC3FFC,$9FCF7FF8,$FFFFFFFF,$FFFF0FF0,
$FFFFFFFF,$F81FFFFF,$FFFFFFFF,$C7F1FFFF,$FF837F3E,$FFFFFFFF,$E3C7FFFF,
$FFFFFFFF,$FFCFFFFF,$E7FFFFFF,$E7F3FFFF,$7FFC7FFC,$9FCF7FFC,$FFFFFFFF,
$FFFF3FFC,$FFFFFFFF,$FC3FFFFF,$FFFFFFFF,$E0F8FFFF,$FF077C3E,$FFFFFFFF,
$C7E3FFFF,$FFFFFFFF,$FFCFFFFF,$E7FFFFFF,$C7F1FFFF,$7FFC7FFC,$1FC77FFC,
$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FC3FFFFF,$FFFFFFFF,$E0F9FFFF,$FF1F003E,
$FFFFFFFF,$8FF1FFFF,$FFFFFFFF,$FFCFFFFF,$E7FFFFFF,$0FF8FFFF,$7FFC3FF8,
$3FE03FF8,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FC3FFFFF,$FFFFFFFF,$C0F9FFFF,
$FFFF003E,$FFFFFFFF,$3FFCFFFF,$FFFFFFFF,$C0CFFFFF,$E707E00F,$1FFCFFFF,
$7FFC1FF0,$7FF01FF0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FC3FFFFF,$FFFFFFFF,
$C7F9FFFF,$FFFF7F3C,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$C0CFFFFF,$E707E00F,
$FFFFFFFF,$3FF88FE3,$FFFF8FE3,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$F81FFFFF,
$FFFFFFFF,$8FF3FFFF,$FFFFFF1C,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$CFCFFFFF,
$E7E7E7CF,$FFFFFFFF,$1FF0CFE7,$FFFFCFE7,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,
$F99FFFFF,$FFFFFFFF,$1FF3FFFF,$FFFFFF9C,$FFFFFFFF,$C7E3FFFF,$FFFFFFFF,
$CFCFFFFF,$E7E7E7CF,$FFFFFFFF,$8FE3CFE7,$FFFFCFE7,$FFFFFFFF,$FFFF07E0,
$FFFFFFFF,$F18FFFFF,$FFFFFFFF,$3FF2FFFF,$FFFFFF89,$FFFFFFFF,$E3C7FFFF,
$FFFFFFFF,$CFCFFFFF,$E7E7E7CF,$FFFFFFFF,$CFE7CFE7,$FFFFCFE7,$FFFFFFFF,
$FFFF07E0,$FFFFFFFF,$C3C3FFFF,$FFFFFFFF,$7FF0FFFF,$FFFFFFC9,$FFFFFFFF,
$E3C7FFFF,$FFFFFFFF,$CFCFFFFF,$E7E7E7CF,$FFFFFFFF,$CFE78FE3,$FFFF8FE3,
$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$FFF8FFFF,$FFFFFFC3,
$FFFFFFFF,$E3C7FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,$CFE71FF0,
$FFFF1FF0,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$1FF8FFFF,$FFFFFFFF,$FFF9FFFF,
$FFFFFFC3,$FFFFFFFF,$C7E3FFFF,$FFFFFFFF,$0FC0FFFF,$07E007C0,$FFFFFFFF,
$8FE33FF8,$FFFF3FF8,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFE7,$FFFFFFFF,$07E0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$1FF0FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFE7,$FFFFFFFF,$0FF0FFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$3FF8FFFF,$FFFFFFFF,$FFFFFFFF,$FFFF7FFE,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$3FFCFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,
$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$FFFFFFFF,$0000FFFF);

type
    IntP=^LongInt;
var
    p: Pointer;
    str: TMemoryStream;
    size: Integer;

    Label Out;
begin
    size:=0;
    p:=nil;
    case _standardSize of
        Size32: begin size:=12406; p:=@set32[0]; goto Out;end;
        Size40: begin size:=19318; p:=@set40[0]; goto Out;end;
        Size64: begin size:=12350; p:=@set64[0]; goto Out;end;
        Size80: begin size:=19262; p:=@set80[0]; goto Out;end
    end;

    Out:

    str:=TMemoryStream.Create;

    str.SetSize(size);
    str.WriteBuffer(p^,size);
    str.Seek(0,soFromBeginning);
    Default.LoadFromStream(str);

    str.Free;
end;

end.
